/* $Id: Checkers.java,v 1.10 2006/08/30 08:52:09 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa._checkers;

import java.util.*;
import java.awt.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public class Checkers extends BoardGameServiceClient
{
  public Checkers() { super(new CheckersGame()); }
  
  public void setDefaultOptions() 
  {
    new CheckersGame().setDefaultOptions();
  }

}


class CheckersGame extends MailBoxBoardGame
{
  public String get_name()    { return "Checkers"; }  
  public String get_base_name()    { return "cks"; }
  public BoardGame new_BoardGame() { return new CheckersGame(); }

  public void setDefaultOptions() {

    String sn = get_base_name() + ".";
    
    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color
    Global.options.put2i(sn+"gc",  BGS+"gc");  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));         // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));         // show last move (boolean)

    // regular empty square color
    Global.options.put(sn+"es0", new Integer(205*65536 + 205*256 + 0));
    Global.options.put(sn+"es1", new Integer(139*65536 + 139*256 + 0));
    
    // anti empty square color (not used)
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));

    Global.options.put(sn+"cm", "");
  }


  // square content
  static final byte BLACK_KING = 4;
  static final byte WHITE_KING = 5;

  // gfx conent
  public static final int GFX_BLACK_CHECKER = 4;
  public static final int GFX_WHITE_CHECKER = 8;
  public static final int GFX_BLACK_KING    = 16;
  public static final int GFX_WHITE_KING    = 32;
  
  public static final int GFX_BLACK_MOVE    = 64;
  public static final int GFX_WHITE_MOVE    = 128;
  public static final int GFX_BLACK_GHOST1  = 256;
  public static final int GFX_WHITE_GHOST1  = 512;
  public static final int GFX_BLACK_GHOST2  = 1024;
  public static final int GFX_WHITE_GHOST2  = 2048;

  public static final int GFX_CHECKER_MASK=GFX_BLACK_CHECKER+GFX_WHITE_CHECKER;
  public static final int GFX_KING_MASK   =GFX_BLACK_KING+GFX_WHITE_KING;  

  protected MailBoxGameState new_MailBoxGameState()
  {
    return new MailBoxGameState();
  }

  public boolean has_synchro()     { return true;  }
  public boolean has_komi()        { return true;  }
  public boolean has_rand()        { return true;  }
  public boolean has_rand_type()   { return false; }
  public boolean has_anti()        { return false; }  
  public boolean has_pref_color()  { return true;  }    
  public boolean has_pass()        { return false; } 

  public boolean never_turn()        { return false; }  
  public boolean turned_if_black()   { return false; }
  protected boolean y_flip()         { return true; }
  protected boolean pieces_on_grid() { return false; }
  protected boolean only_squares()   { return true;  }

  protected int border_width()     { return 1; }

  public int board_type_num()      { return 3; }
  public int default_board_type()  { return 8; }

  public int board_type(int index)
  {
    switch (index) {
    case 0:  return 6;
    case 1:  return 8;
    case 2:  return 10;
    default: return -1;
    }
  }

  public int board_type_index(int type) {
    switch (type) {
    case 6:  return 0;
    case 8:  return 1;
    case 10: return 2;
    default: return -1;
    }
  }
  
  public String board_type_string(int index) {
    switch (index) {
    case 0:  return "6x6";
    case 1:  return "8x8";
    case 2:  return "10x10";
    default: return "?";
    }
  }

  public int width(int type) { return type; }
  
  public int height(int type) { return width(type); }
  
  public byte char2cont(char c)
  {
    switch (c) {
    case 'b': return BLACK;
    case 'B': return BLACK_KING;
    case 'w': return WHITE;
    case 'W': return WHITE_KING;
    case '-': return EMPTY;
    default:
      throw new Error("illegal char in Checkers.char2cont");
    }
  }

  public char cont2char(byte co)
  {
    switch (co) {
    case BLACK:      return 'b';
    case BLACK_KING: return 'B';
    case WHITE:      return 'w';
    case WHITE_KING: return 'W';
    case EMPTY:      return '-'; 
    default:
      throw new Error("illegal cont in Checkers.cont2char");
    }
  }


  private boolean on_same_line(int ind1, int ind2)
  {
    int x1, y1, x2, y2;
    
    x1 = ind2x(ind1);
    y1 = ind2y(ind1);
    x2 = ind2x(ind2);
    y2 = ind2y(ind2);
    
    return x1 == x2 || y1 == y2 || Math.abs(x1-x2) == Math.abs(y1-y2);
  }


  private static int sign(int d)
  {
    if (d > 0) return 1;
    if (d < 0) return -1;
    return 0;
  }

  private int cont_num(MailBoxGameState state, int cont)
  {
    int n = 0;

    for (int i=0; i < BOARD_SIZE; ++i)
      if (state.sq[i] == cont) ++n;

    return n;
  }
                       
  private boolean empty_line(MailBoxGameState st, int ind1, int ind2)
  {
    if (ind1 == ind2) return false;
    if (!on_same_line(ind1, ind2)) return false;
    int x1, y1, x2, y2;

    x1 = ind2x(ind1);
    y1 = ind2y(ind1);
    x2 = ind2x(ind2);
    y2 = ind2y(ind2);

    int dx = sign(x2-x1);
    int dy = sign(y2-y1);
    int d = dx + dy * DX;
    int p = ind1 + d;

    while (p != ind2) {
      if (st.sq[p] != EMPTY) return false;
      p += d;
    }
  
    return true;
  }

  private boolean moves_available(MailBoxGameState st, boolean only_jumps)
  {
    int own_checker, own_king;
    int opp_checker, opp_king;
    int dir;
  
    if (st.to_move == BLACK) {
      own_checker = BLACK; own_king = BLACK_KING; dir = DX;
      opp_checker = WHITE; opp_king = WHITE_KING;
    } else {
      own_checker = WHITE; own_king = WHITE_KING; dir = -DX;
      opp_checker = BLACK; opp_king = BLACK_KING;
    }

    for (int i=0; i < BOARD_SIZE; i++) {
      if (st.sq[i] == own_checker || st.sq[i] == own_king) {
      
	// checker jump possible?

	if ((st.sq[i+dir+1] == opp_checker || st.sq[i+dir+1] == opp_king) &&
	    st.sq[i+2*(dir+1)] == EMPTY) return true;

	if ((st.sq[i+dir-1] == opp_checker || st.sq[i+dir-1] == opp_king) &&
	    st.sq[i+2*(dir-1)] == EMPTY) return true;

	// checker move possible?
      
	if (!only_jumps && (st.sq[i+dir+1] == EMPTY || st.sq[i+dir-1] == EMPTY))
	  return true;
      
	if (st.sq[i] == own_king) {

	  // king jump possible?
        
	  if ((st.sq[i-dir+1] == opp_checker || st.sq[i-dir+1] == opp_king) &&
	      st.sq[i+2*(-dir+1)] == EMPTY) return true;
        
	  if ((st.sq[i-dir-1] == opp_checker || st.sq[i-dir-1] == opp_king) &&
	      st.sq[i+2*(-dir-1)] == EMPTY) return true;

	  // king move possible?
      
	  if (!only_jumps && (st.sq[i-dir+1] == EMPTY || st.sq[i-dir-1] == EMPTY))
	    return true;
	}
      }
    }
    return false;
  }


  private synchronized boolean is_finished() {
    return !moves_available((MailBoxGameState)current_state(), false);
  }

  
  // st changed!

  private boolean do_move(MailBoxGameState st, int ms[], boolean prefix)
  {
    int dist[] = { DX+1, -(DX+1), DX-1, -(DX-1) };

//     Global.dbgmsg("DO_MOVE: " + ms.length);
//     for (int i=0; i < ms.length; ++i) {
//       Global.dbgmsg(">>> " + ms[i]);
    //    }
    
    int own_checker, own_king;
    int opp_checker, opp_king;
    int dir;
    
    if (st.to_move == BLACK) {
      own_checker = BLACK; own_king = BLACK_KING; dir = DX;
      opp_checker = WHITE; opp_king = WHITE_KING;
    } else {
      own_checker = WHITE; own_king = WHITE_KING; dir = -DX;
      opp_checker = BLACK; opp_king = BLACK_KING;
    }
    
    if (st.sq[ms[0]] != own_checker &&
	st.sq[ms[0]] != own_king) return false;
    
    if (!moves_available(st, false)) return false;
    
    boolean must_jump = moves_available(st, true);

    if (must_jump) {

      // move must be sequence of jumps

      int[] capt_pos    = new int[ms.length];
      byte[] capt_piece = new byte[ms.length];      
      int cn = 0;
      int i;
      
      for (i=1; i < ms.length; i++) {

	if (st.sq[ms[i]] != EMPTY) return false;

	if (st.sq[ms[i-1]] == own_king) {

	  int j;
        
	  for (j=0; j < 4; j++)
	    if (ms[i] == ms[i-1] + 2*dist[j]) break;

	  if (j >= 4) return false;
        
	  if (st.sq[ms[i-1]+dist[j]] != opp_checker &&
	      st.sq[ms[i-1]+dist[j]] != opp_king) return false; 

	} else {

	  int e = 0;
        
	  if      (ms[i] == ms[i-1] + 2*(dir+1)) e = dir+1;
	  else if (ms[i] == ms[i-1] + 2*(dir-1)) e = dir-1;

	  if (e == 0) return false;
        
	  if (st.sq[ms[i-1]+e] != opp_checker &&
	      st.sq[ms[i-1]+e] != opp_king) return false;
	}

	capt_pos[cn] = (ms[i]+ms[i-1])/2;
	capt_piece[cn] = st.sq[capt_pos[cn]];
	st.sq[capt_pos[cn]] = EMPTY;
	st.sq[ms[i]] = st.sq[ms[i-1]];
	st.sq[ms[i-1]] = EMPTY;
	cn++;
      }

      // jumps OK

      if (ms.length == 1) {

	// is jump possible at all from this stating location?

	if (!prefix) return false;

	int last_square = ms[0];
	
	if (st.sq[last_square] == own_king) {

	  for (int i1 = 0; i1 < 4; i1++) {
	    if (st.sq[last_square + dist[i1]] == opp_checker ||
		st.sq[last_square + dist[i1]] == opp_king)
	      if (st.sq[last_square + 2*dist[i1]] == EMPTY) return true;
	  }

	} else {

	  if (((st.sq[last_square + dir+1] == opp_checker ||
		st.sq[last_square + dir+1] == opp_king) && 
	       st.sq[last_square + 2*(dir+1)] == EMPTY) ||
	      ((st.sq[last_square + dir-1] == opp_checker ||
		st.sq[last_square + dir-1] == opp_king) && 
	       st.sq[last_square + 2*(dir-1)] == EMPTY)) return true;
	}

	return false;
      }

      if (!prefix) {

	// jumps OK; is there a jump missing?

	int last_square = ms[ms.length-1];
	
	if (st.sq[last_square] == own_king) {

	  for (int i1 = 0; i1 < 4; i1++) {
	    if (st.sq[last_square + dist[i1]] == opp_checker ||
		st.sq[last_square + dist[i1]] == opp_king)
	      if (st.sq[last_square + 2*dist[i1]] == EMPTY) return false;
	  }

	} else {

	  if (((st.sq[last_square + dir+1] == opp_checker ||
		st.sq[last_square + dir+1] == opp_king) && 
	       st.sq[last_square + 2*(dir+1)] == EMPTY) ||
	      ((st.sq[last_square + dir-1] == opp_checker ||
		st.sq[last_square + dir-1] == opp_king) && 
	       st.sq[last_square + 2*(dir-1)] == EMPTY)) return false;
	}
      
      }

    } else {

      // no jump

      if (ms.length < 2 && prefix) {

	// possible move extension?
	
	if (st.sq[ms[0]] == own_king) {

	  for (int i=0; i < 4; i++)
	    if (st.sq[ms[0] + dist[i]] == EMPTY) return true;

	} else {
	  
	  if (st.sq[ms[0]+dir+1] == EMPTY ||
	      st.sq[ms[0]+dir-1] == EMPTY) return true;
	}

	return false;
      }

      if (ms.length < 2) return false;
      
      if (st.sq[ms[1]] != EMPTY) return false;
    
      if (st.sq[ms[0]] == own_king) {

	if (ms[1] != ms[0]+dir+1 && ms[1] != ms[0]+dir-1 &&
	    ms[1] != ms[0]-dir+1 && ms[1] != ms[0]-dir-1)
	  return false;

      } else {
    
	if (ms[1] != ms[0]+dir+1 && ms[1] != ms[0]+dir-1)
	  return false;
      }

      if (!prefix) {
	st.sq[ms[1]] = st.sq[ms[0]];
	st.sq[ms[0]] = EMPTY;
      }
    }

    if (!prefix) {

      int last_sq = ms[ms.length-1];
      int w = WIDTH;

      if (st.to_move == PlayerColor.BLACK && ind2y(last_sq) == w-1)
	st.sq[last_sq] = BLACK_KING;

      if (st.to_move == PlayerColor.WHITE && ind2y(last_sq) == 0)
	st.sq[last_sq] = WHITE_KING;

      st.to_move = PlayerColor.opponent(st.to_move);
    }

    return true;
  }
  
  private synchronized boolean seq_ok(MailBoxGameState st,
				      int[] ms,
				      boolean released)
  {
    // when button is released new start-square must be previous end-square

    int n = ms.length;
    if (!released && (n > 1)) { return ms[n-1] == ms[n-2]; }

    if (n > 1 && ms[n-1] == ms[n-2]) return false; // click on one square

    MailBoxGameState new_state = st.createCopy();
    int conv[] = convert(ms);
    boolean ok = do_move(new_state, conv, true);
    return ok;
  }

  
  public synchronized MailBoxGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;
    if (v.size() < 2) return null;

    int coords[] = new int[v.size()];

    for (int i=0; i < v.size(); ++i) {
      MovePart mp = v.elementAt(i);
      if (!mp.rest.equals("")) return null;
      coords[i] = mp.square;
    }
      
    // check move

    MailBoxGameState new_state = ((MailBoxGameState)current_state()).createCopy();

    if (do_move(new_state, coords, false)) return new_state;
    return null;
  }

  public void input(String s) { }

  protected synchronized void process_mouse_events()
  {
    synchronized (mev) {
      
      int own_checker, own_king;
      
      MailBoxGameState st = (MailBoxGameState) current_state();
      
      if (st.to_move == PlayerColor.BLACK) {
	own_checker = BLACK; own_king = BLACK_KING;
      } else {
	own_checker = WHITE; own_king = WHITE_KING;
      }
      
      MouseEvent me = (MouseEvent)mev.lastElement();

      int coords[] = new int[mev.size()];

      for (int i=0; i < mev.size(); ++i)
	coords[i] = ((MouseEvent)(mev.elementAt(i))).square;
      
//       Global.dbgmsg("PROCESS: " + coords.length);
//       for (int i=0; i < coords.length; ++i) {
// 	Global.dbgmsg(">>> " + ind2string(coords[i]));
//       }
    
      int conv[] = convert(coords);

      if (me.type == PRESSED) {
	
	// first press => must be on own mobile piece

	if (!seq_ok(st, coords, false)) { reset_mev(); return; }

      } else if (mev.size() > 0 && me.type == RELEASED) {
	
	if (!seq_ok(st, coords, true)) { reset_mev(); return; }

	MailBoxGameState new_st = st.createCopy();

	
// 	Global.dbgmsg("CHECK: " + coords.length);
// 	for (int i=0; i < conv.length; ++i) {
// 	  Global.dbgmsg(">>> " + ind2string(conv[i]));
// 	}
 	
	if (do_move(new_st, conv, false)) {

	  // move ok

	  String ms = ind2string(conv[0]);
	  for (int i=1; i < conv.length; ++i) ms += "-" + ind2string(conv[i]);
	  handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID, ms));
	  reset_mev();
	}

      } else {

	// remove last event

	mev.removeElementAt(mev.size()-1);
      }
    }
  }

  // convert mouse event square sequence into a move
  
  private int[] convert(int[] mco)
  {
    if (mco == null || mco.length < 1) return null;
    
    // count different neighbors
        
    int k = 1;
    for (int j=1; j < mco.length; ++j) if (mco[j] != mco[j-1]) k++;

    int[] co = new int[k];
    co[0] = mco[0];
    k = 1;
    for (int j=1; j < mco.length; ++j) {
      if (mco[j] != mco[j-1]) co[k++] = mco[j];
    }
    
    return co;
  }

  protected boolean entire_gfx_cont(MailBoxGameState st0, String last_move) {

    MailBoxGameState st = (MailBoxGameState) st0;
      
    for (int i=0; i < BOARD_SIZE; ++i) {

      int c;
      
      switch (st.sq[i]) {

      case BORDER:     c = GFX_BORDER;        break;
      case BLACK:      c = GFX_BLACK_CHECKER; break;
      case WHITE:      c = GFX_WHITE_CHECKER; break;
      case BLACK_KING: c = GFX_BLACK_KING;    break;
      case WHITE_KING: c = GFX_WHITE_KING;    break;       

      default: c = GFX_EMPTY;

      }

      gfx_sq[i] = c;
    }

    if (show_last_move && last_move_index(last_move) >= 0)
      gfx_sq[last_move_index(last_move)] |= GFX_LAST_MOVE;

    int move_flag = ((st.to_move == PlayerColor.BLACK) ?
		     GFX_BLACK_MOVE : GFX_WHITE_MOVE);
    int opp_move_flag = ((st.to_move == PlayerColor.BLACK) ?
			 GFX_WHITE_MOVE : GFX_BLACK_MOVE);

    int ghost1_flag = ((st.to_move == PlayerColor.BLACK) ?
		       GFX_BLACK_GHOST1 : GFX_WHITE_GHOST1);
    int ghost2_flag = ((st.to_move == PlayerColor.BLACK) ?
		       GFX_BLACK_GHOST2 : GFX_WHITE_GHOST2);

    int own_checker = ((st.to_move == PlayerColor.BLACK) ? BLACK : WHITE);
    int own_king = ((st.to_move == PlayerColor.BLACK) ? BLACK_KING : WHITE_KING);

    int n = mev.size();
    int[] co1 = new int[1];
    int[] co2 = new int[2];
    int[] con1 = new int[n+1];

    for (int i=0; i < n; ++i) con1[i] = ((MouseEvent)mev.elementAt(i)).square;
    
    MouseEvent me = null;

    if (n > 0) me = (MouseEvent)mev.elementAt(n-1);

    if (show_moves) {

      if (n == 0) {
        
        // no click: mark checkers to move
        
        for (int i=0; i < BOARD_SIZE; i++) {
          
          if (st.sq[i] == own_checker || st.sq[i] == own_king) {
            
            // move possible?

            co1[0] = i;
	    MailBoxGameState new_state = st.createCopy();
            if (do_move(new_state, co1, true)) gfx_sq[i] |= opp_move_flag;
          }
        }

      } else {

	int dist[] = { DX+1, -(DX+1), DX-1, -(DX-1) };

	if (me.type.equals(PRESSED)) {

          // mark moves if button is pressed

          int last_pos = me.square;
          
          for (int j=0; j < 4; j++) {
            int new_pos = last_pos + dist[j];
            
            if (st.sq[new_pos] == EMPTY) {
              
              con1[n] = new_pos;
	      MailBoxGameState new_state = st.createCopy();
              if (do_move(new_state, convert(con1), true))
		gfx_sq[new_pos] |= move_flag;
              
            } else if (st.sq[new_pos] != BORDER) {
              
              new_pos += dist[j];
              con1[n] = new_pos;
	      MailBoxGameState new_state = st.createCopy();
              if (do_move(new_state, convert(con1), true))
		gfx_sq[new_pos] |= move_flag;
            }
          }
        }
      }
    }

    if (n > 0) {
      
      for (int i=0; i < n-1; i++) {
        gfx_sq[con1[i]] &= ~(GFX_CHECKER_MASK | GFX_KING_MASK);
      }

      if (n == 1) {

        gfx_sq[con1[0]] &= ~(GFX_CHECKER_MASK | GFX_KING_MASK);
        gfx_sq[con1[0]] |= ghost1_flag; 

      } else {

        if (me.type.equals(PRESSED)) {

          gfx_sq[con1[n-1]] |= ghost2_flag; 
          
        } else {

          gfx_sq[con1[n-1]] |= ghost1_flag;
          
        }
      }
    }
    
    return true;
  }

  protected int last_move_index(String move)
  {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null) return -1;
    return v.elementAt(v.size()-1).square;
  }

  public int gfx_cont(MailBoxGameState st, int i) { return GFX_EMPTY; }

  protected void paint_gfx_cont(Graphics g, int x, int y, int parity, int gfx_cont)
  {
    if ((gfx_cont & GFX_BORDER) != 0)
      return;
  
    if (parity == 0) g.setColor(col_empty0); else g.setColor(col_empty1);

    g.fillRect(x+1,y+1,gfx_sq_w-1, gfx_sq_h-1);
    int disc_diam = (int)Math.round(gfx_sq_w * 0.85);
    disc_diam &= ~1;

    if ((gfx_cont & GFX_BLACK_GHOST1) + (gfx_cont & GFX_WHITE_GHOST1) > 0) {

      Color col;
      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_w - disc_diam)/2;
      
      if ((gfx_cont & GFX_BLACK_GHOST1) != 0) col = col_black; else col = col_white;

      g.setColor(col);
      Bresenham.drawCircle(g, xl, yt, disc_diam);

    } else if ((gfx_cont & GFX_BLACK_GHOST2) + (gfx_cont & GFX_WHITE_GHOST2) > 0) {

      Color col;
      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_w - disc_diam)/2;
      
      if ((gfx_cont & GFX_BLACK_GHOST2) != 0) col = col_black; else col = col_white;

      g.setColor(col);
      Bresenham.drawCircle(g, xl, yt, disc_diam);

      int diam2 = (disc_diam*3)/4;
      diam2 &= ~1;
      
      xl = x + (gfx_sq_w - diam2)/2;
      yt = y + (gfx_sq_w - diam2)/2;
      Bresenham.drawCircle(g, xl, yt, diam2);
    }

    if ((gfx_cont & GFX_CHECKER_MASK) != 0) {

      // disc

      Color d_col;
      Color b_col;

      if ((gfx_cont & GFX_BLACK_CHECKER) != 0) {

	d_col = col_black;
	b_col = col_white;

      } else {

	b_col = col_black;
	d_col = col_white;

      }

      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_w - disc_diam)/2;

      g.setColor(d_col);
      Bresenham.fillCircle(g, xl, yt, disc_diam);

      g.setColor(b_col);
      Bresenham.drawCircle(g, xl, yt, disc_diam);
    }

    if ((gfx_cont & GFX_KING_MASK) != 0) {

      // disc

      Color d_col;
      Color b_col;

      if ((gfx_cont & GFX_BLACK_KING) != 0) {

	d_col = col_black;
	b_col = col_white;

      } else {

	b_col = col_black;
	d_col = col_white;

      }

      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_w - disc_diam)/2;

      g.setColor(d_col);
      Bresenham.fillCircle(g, xl, yt, disc_diam);

      g.setColor(b_col);
      Bresenham.drawCircle(g, xl, yt, disc_diam);

      int diam2 = (disc_diam*3)/4;
      diam2 &= ~1;
      
      xl = x + (gfx_sq_w - diam2)/2;
      yt = y + (gfx_sq_w - diam2)/2;
      Bresenham.drawCircle(g, xl, yt, diam2);
    }

    if ((gfx_cont & GFX_BLACK_MOVE) != 0 ||
	(gfx_cont & GFX_WHITE_MOVE) != 0) {

      // legal move: draw X
      
      int cx = x + gfx_sq_w/2;
      int cy = y + gfx_sq_w/2;
      int l = gfx_sq_w/16;
      
      if ((gfx_cont & GFX_BLACK_MOVE) != 0)
	g.setColor(col_black);
      else 
	g.setColor(col_white);
      
      g.drawLine(cx-l,   cy-l, cx+l,   cy+l);
      g.drawLine(cx-l-1, cy-l, cx+l-1, cy+l);
      g.drawLine(cx-l,   cy+l, cx+l,   cy-l);
      g.drawLine(cx-l-1, cy+l, cx+l-1, cy-l);
      
    }

    if ((gfx_cont & GFX_LAST_MOVE) != 0) {

      // last move: draw inner square
      
      Color b_col;

      if ((gfx_cont & GFX_BLACK_CHECKER) != 0 || (gfx_cont & GFX_BLACK_KING) != 0) {

	b_col = col_white;

      } else {

	b_col = col_black;

      }

      g.setColor(b_col);

      int dot_diam = disc_diam/8;
      dot_diam &= ~1;
      
      int xl = x + (gfx_sq_w - dot_diam)/2;
      int yt = y + (gfx_sq_w - dot_diam)/2;
      g.drawRect(xl, yt, dot_diam, dot_diam);
    }
  }


  protected void init_special_border(MailBoxGameState state) { }

  public String info_string() {
    
    int e = cont_num((MailBoxGameState)current_state(), EMPTY);
    int a = BOARD_SIZE;
    
    return "(" + (a-e) + " " + (e) + ")";
  }
  
  protected String read_extra_info(MailBoxGameState st, StringStream is) {
    return null;
  }
  
  protected String extra_info_to_string(MailBoxGameState st) {
    return "";
  }

  public boolean paint_all(Graphics g, boolean repaint,
			   int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    return false;
  }  
}

