/* $Id: Chess.java,v 1.11 2006/08/30 08:52:09 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa._chess;

import java.util.*;
import java.awt.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;


public class Chess extends BoardGameServiceClient
{
  public Chess() { super(new ChessGame()); }
  
  public void setDefaultOptions() 
  {
    new ChessGame().setDefaultOptions();
  }

}


class ChessGame extends MailBoxBoardGame
{
  public String get_name()    { return "Chess"; }  
  public String get_base_name()    { return "cs"; }
  public BoardGame new_BoardGame() { return new ChessGame(); }

  public void setDefaultOptions()
  {
    String sn = get_base_name() + ".";
    
    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color
    Global.options.put2i(sn+"gc",  BGS+"gc");  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));         // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));         // show last move (boolean)

    // regular empty square color
    Global.options.put(sn+"es0", new Integer(0*65536  + 205*256 + 0));
    Global.options.put(sn+"es1", new Integer(0*65536   + 165*256 + 0));
    
    // move square color
    Global.options.put(sn+"ms",  new Integer(255*65536 + 160*256 + 122));

    // last move square color
    Global.options.put(sn+"lms",  new Integer(160*65536 + 160*256 + 160));

    // default piece image path template: #=size, %=piece prefix
    Global.options.put(sn+"pt", "/data/images/chess/#/%d#.gif");

    // default piece sizes
    Global.options.put(sn+"ps", "21,25,29,33,37,45,58,80,108,129");

    // anti empty square color (not used)
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));    

    Global.options.put(sn+"cm", "");
  }


  // square content
  protected static final byte PAWN   = 4;
  protected static final byte KNIGHT = 5;
  protected static final byte BISHOP = 6;
  protected static final byte ROOK   = 7;
  protected static final byte QUEEN  = 8;
  protected static final byte KING   = 9;  

  protected static final byte W_PAWN   = PAWN;
  protected static final byte W_KNIGHT = KNIGHT;
  protected static final byte W_BISHOP = BISHOP;
  protected static final byte W_ROOK   = ROOK;
  protected static final byte W_QUEEN  = QUEEN;
  protected static final byte W_KING   = KING;   

  protected static final byte B_PAWN   = -PAWN;  
  protected static final byte B_KNIGHT = -KNIGHT;
  protected static final byte B_BISHOP = -BISHOP;
  protected static final byte B_ROOK   = -ROOK;  
  protected static final byte B_QUEEN  = -QUEEN; 
  protected static final byte B_KING   = -KING;    

  protected static final int  NO_PROM        = 0;
  protected static final int  PROM_TO_KNIGHT = 1;
  protected static final int  PROM_TO_BISHOP = 2;
  protected static final int  PROM_TO_ROOK   = 3;
  protected static final int  PROM_TO_QUEEN  = 4;


  // gfx content
  protected static final int GFX_W_PAWN   = 1*4;
  protected static final int GFX_W_KNIGHT = 2*4;
  protected static final int GFX_W_BISHOP = 3*4;
  protected static final int GFX_W_ROOK   = 4*4;
  protected static final int GFX_W_QUEEN  = 5*4;
  protected static final int GFX_W_KING   = 6*4;

  protected static final int GFX_B_PAWN   = 7*4;
  protected static final int GFX_B_KNIGHT = 8*4;
  protected static final int GFX_B_BISHOP = 9*4;
  protected static final int GFX_B_ROOK   = 10*4;
  protected static final int GFX_B_QUEEN  = 11*4;
  protected static final int GFX_B_KING   = 12*4;

  protected static final int GFX_MOVE       = 16*4;    
  protected static final int GFX_PIECE_MASK = 15*4;


  // image data
  private static String[] piece_strings = {
    "pl", "nl", "bl", "rl", "ql", "kl",
    "pd", "nd", "bd", "rd", "qd", "kd"
  };

  protected static final int IMGW_PAWN   = 0;
  protected static final int IMGW_KNIGHT = 1;
  protected static final int IMGW_BISHOP = 2;
  protected static final int IMGW_ROOK   = 3;
  protected static final int IMGW_QUEEN  = 4;
  protected static final int IMGW_KING   = 5;
					   
  protected static final int IMGB_PAWN   = 6;
  protected static final int IMGB_KNIGHT = 7;
  protected static final int IMGB_BISHOP = 8;
  protected static final int IMGB_ROOK   = 9;
  protected static final int IMGB_QUEEN  = 10;
  protected static final int IMGB_KING   = 11;
    
  private static PieceImages[] ci;
  private static MediaTracker mt;
  private static boolean gfx_initialized = false;

  int move_from =-1, move_to = -1;  // stored here for promotion protocol
  
  protected MailBoxGameState new_MailBoxGameState()
  {
    return new ChessGameState();
  }

  public boolean has_synchro()     { return true;  }
  public boolean has_komi()        { return true;  }
  public boolean has_rand()        { return true;  }
  public boolean has_rand_type()   { return false; }
  public boolean has_anti()        { return false; }  
  public boolean has_pref_color()  { return true;  }    
  public boolean has_pass()        { return false; } 

  public boolean never_turn()        { return false; }  
  public boolean turned_if_black()   { return true; }
  protected boolean y_flip()         { return true; }
  protected boolean pieces_on_grid() { return false; }
  protected boolean only_squares()   { return true;  }

  protected int border_width()     { return 2; }

  public int board_type_num()      { return 4; }
  public int default_board_type()  { return 8; }

  public int board_type(int index)
  {
    switch (index) {
    case 0:  return 6;
    case 1:  return 8;
    case 2:  return 10;
    case 3:  return 12;      
    default: return -1;
    }
  }

  public int board_type_index(int type) {
    switch (type) {
    case 6:  return 0;
    case 8:  return 1;
    case 10: return 2;
    case 12: return 3;      
    default: return -1;
    }
  }
  
  public String board_type_string(int index) {
    switch (index) {
    case 0:  return "6x6";
    case 1:  return "8x8";
    case 2:  return "10x10";
    case 3:  return "12x12";      
    default: return "?";
    }
  }

  public int width(int type) { return type; }
  
  public int height(int type) { return width(type); }
  
  public byte char2cont(char c)
  {
    switch (c) {
    case '-': return EMPTY;   
    case 'p': return B_PAWN;  
    case 'n': return B_KNIGHT;
    case 'b': return B_BISHOP;
    case 'r': return B_ROOK;  
    case 'q': return B_QUEEN; 
    case 'k': return B_KING;  
                              
    case 'P': return W_PAWN;  
    case 'N': return W_KNIGHT;
    case 'B': return W_BISHOP;
    case 'R': return W_ROOK;  
    case 'Q': return W_QUEEN; 
    case 'K': return W_KING;  
      
    default: throw new Error("illegal char in char2cont");

    }
  }
  
  public char cont2char(byte cont)
  {
    switch (cont) {
    case EMPTY:    return '-';

    case B_PAWN:   return 'p';
    case B_KNIGHT: return 'n';
    case B_BISHOP: return 'b';
    case B_ROOK:   return 'r';
    case B_QUEEN:  return 'q';
    case B_KING:   return 'k';
                              
    case W_PAWN:   return 'P';
    case W_KNIGHT: return 'N';
    case W_BISHOP: return 'B';
    case W_ROOK:   return 'R';
    case W_QUEEN:  return 'Q';
    case W_KING:   return 'K';

    default: throw new Error("illegal cont in cont2char");
    }
  }

  public synchronized void input(String s) {

    int i = s.indexOf('?');
    if (i < 0) return;
    
    StringStream is = new StringStream(s.substring(i+1));
    char c = is.get_skip_ws();

    if (!is.ok()) c = 'Q';
    else {
      
      c = Character.toUpperCase(c);

      switch (c) {
      case 'Q': break;
      case 'R': break;
      case 'B': break;
      case 'N': break;
      default: return;
      }
    }
    
    handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
					  ind2string(move_from) + "-" +
					  ind2string(move_to) + c));
  }

  protected synchronized void process_mouse_events() {

    boolean ok = false, promo = false;
    int from = PASS;
    int to   = PASS;
    
    synchronized (mev) {
      
      int n = mev.size();
      int i;

      for (i=0; i < mev.size(); ++i) {
	MouseEvent m = (MouseEvent)mev.elementAt(i);
	if (m.type == PRESSED) break;
	mev.removeElementAt(i);
	--i;
	continue;
      }

      ++i;
      
      for (;i < mev.size(); ++i) {
	MouseEvent m = (MouseEvent)mev.elementAt(i);
	if (m.type == RELEASED) break;
	mev.removeElementAt(i);
	--i;
	continue;
      }

      if (mev.size() == 0) return;

      ChessGameState state = (ChessGameState)current_state();

      if (mev.size() >= 1) {
	from = ((MouseEvent)mev.elementAt(0)).square;
	if (color_of(state.sq[from]) != state.to_move) { reset_mev(); return; }
      }	

      if (mev.size() >= 2) {
	to = ((MouseEvent)mev.elementAt(1)).square;
	if (from != to) ok = true;
	else reset_mev();

	if (Math.abs(state.sq[from]) == PAWN) {
	  promo = (state.to_move == WHITE && ind2y(to) == HEIGHT-1) ||
	         (state.to_move == BLACK && ind2y(to) == 0);
	}
      }
    }

    if (ok) {

      move_from = from;
      move_to = to;
      
      if (promo)
	handler.handle_event(new EventMsg(GameWindow.INPUT_ID,
					  "promote to Q,R,B,N? Q"));
      else
      
	handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
					  ind2string(from) + "-" +
					  ind2string(to)));
      reset_mev();
    }
  }

  protected synchronized boolean entire_gfx_cont(MailBoxGameState st0, String last_move) {

    ChessGameState st = (ChessGameState)st0;

    for (int i=0; i < BOARD_SIZE; ++i) {

      byte co;
      
      switch (st.sq[i]) {
      case EMPTY:    co = GFX_EMPTY;  break;
      case BORDER:   co = GFX_BORDER; break;
      case W_PAWN:   co = GFX_W_PAWN; break;
      case B_PAWN:   co = GFX_B_PAWN; break;
      case W_KNIGHT: co = GFX_W_KNIGHT; break;
      case B_KNIGHT: co = GFX_B_KNIGHT; break;	
      case W_BISHOP: co = GFX_W_BISHOP; break;
      case B_BISHOP: co = GFX_B_BISHOP; break;	
      case W_ROOK:   co = GFX_W_ROOK; break;
      case B_ROOK:   co = GFX_B_ROOK; break;	
      case W_QUEEN:  co = GFX_W_QUEEN; break;
      case B_QUEEN:  co = GFX_B_QUEEN; break;	
      case W_KING:   co = GFX_W_KING; break;
      case B_KING:   co = GFX_B_KING; break;	
      default: throw new Error("unknown square contents");
      }

      gfx_sq[i] = co;
    }

    if (show_last_move) {
      Vector<MovePart> v = new Vector<MovePart>();
      if (parse_move(last_move, v) == null) {
        gfx_sq[v.elementAt(1).square] |= GFX_LAST_MOVE;
      }
    }

    int n = mev.size();
    MouseEvent me = null;
    if (n > 0) me = mev.elementAt(n-1);

    if (n == 0) {

      if (false && show_moves) {
      
	// no click: mark piece to move if in check
	
	for (int i=0; i < BOARD_SIZE; ++i) {
	  
	  if (st.sq[i] == st.to_move) {
	    
	    // move possible?
	    
	    for (int j=0; j < 8; j++) 
	      if (st.sq[i + D[j]] == EMPTY) {
		gfx_sq[i] |= 0; // opp_move_flag;
		break;
	      }
	  }
	}
      }
      
    } else if (n == 1) {

      gfx_sq[me.square] |= GFX_MOVE;

      if (false && show_moves) {
	
	// one click: mark legal moves
	
	for (int i=0; i < 8; i++) {
	  
	  int pos = me.square + D[i];
	  
	  while (st.sq[pos] == EMPTY) {
	    gfx_sq[pos] |= 0; // move_flag;
	    pos += D[i];
	  }
	}
      }
      
    } else if (n == 2) {

      if (false && show_moves) {
	
	// two clicks: mark arrow locations
      
	for (int i=0; i < 8; i++) {
	  
	  int d = D[i];
	  int pos = me.square + d;
	  
	  while (st.sq[pos] == EMPTY) {
	    st.sq[pos] |= 0; // move_flag;
	    pos += d;
	  }
	}
      }
    }

    return true;
  }

  protected synchronized int last_move_index(String move)
  {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null) return -1;
    return v.elementAt(1).square;
  }

  public int gfx_cont(MailBoxGameState st, int i) { return GFX_EMPTY; }

  
  protected synchronized void paint_gfx_cont(Graphics g, int x, int y, int parity, int gfx_cont)
  {
    if (!gfx_initialized) {

      gfx_initialized = true;
      
      // initialize piece images

      String sn = get_base_name() + ".";
      String template = Global.options.getString(sn+"pt");
      String size_string = Global.options.getString(sn+"ps");    
    
      ci = new PieceImages[piece_strings.length];

      for (int i=0; i < piece_strings.length; ++i) {
	String piece_template = PieceImages.expand(template, '%', piece_strings[i]);
	ci[i] = new PieceImages(piece_template, size_string);
      }
    }
    
    if ((gfx_cont & GFX_BORDER) != 0) return;

    String sn = get_base_name() + ".";

    if ((gfx_cont & GFX_MOVE) != 0) {

      g.setColor(new Color(Global.options.getInteger(sn+"ms").intValue()));
      
    } else if ((gfx_cont & GFX_LAST_MOVE) != 0) {

      g.setColor(new Color(Global.options.getInteger(sn+"lms").intValue()));
      
    } else {
    
      if (parity == 0) g.setColor(col_empty0); else g.setColor(col_empty1);

    }
    
    g.fillRect(x+1,y+1,gfx_sq_w-1, gfx_sq_h-1);

    if ((gfx_cont & GFX_PIECE_MASK) != 0) {

      int co = gfx_cont & GFX_PIECE_MASK;

      // draw piece

      int sz = (int)(gfx_sq_w * 0.95);
      int pref_size = ci[IMGW_PAWN].get_pref_size(sz);
      int index = ci[IMGW_PAWN].get_image_index(sz);
      Image im = null;
      
      switch (co) {
      case GFX_W_PAWN:   im = ci[IMGW_PAWN]  .get(index, mtrack); break;
      case GFX_W_KNIGHT: im = ci[IMGW_KNIGHT].get(index, mtrack); break;
      case GFX_W_BISHOP: im = ci[IMGW_BISHOP].get(index, mtrack); break;
      case GFX_W_ROOK:   im = ci[IMGW_ROOK]  .get(index, mtrack); break; 
      case GFX_W_QUEEN:  im = ci[IMGW_QUEEN] .get(index, mtrack); break;
      case GFX_W_KING:   im = ci[IMGW_KING]  .get(index, mtrack); break; 
           				             
      case GFX_B_PAWN:   im = ci[IMGB_PAWN]  .get(index, mtrack); break; 
      case GFX_B_KNIGHT: im = ci[IMGB_KNIGHT].get(index, mtrack); break;
      case GFX_B_BISHOP: im = ci[IMGB_BISHOP].get(index, mtrack); break;
      case GFX_B_ROOK:   im = ci[IMGB_ROOK]  .get(index, mtrack); break; 
      case GFX_B_QUEEN:  im = ci[IMGB_QUEEN] .get(index, mtrack); break;
      case GFX_B_KING:   im = ci[IMGB_KING]  .get(index, mtrack); break;
      default: throw new Error("illegal square content " + co);
      }
        
      int d = (gfx_sq_w - pref_size)/2;
      g.drawImage(im, x + d, y + d, pref_size, pref_size, component);

    }
  }

  protected void init_special_border(MailBoxGameState state) { }

  public synchronized String info_string() {

    ChessGameState st = (ChessGameState)current_state();

    String eps = "%";
    if (st.ep_square >= 0) eps = ind2string(st.ep_square);
    
    return (index+1) + " " +
      st.rev_move_num + " " +
      eps + 
      " " + (st.CQ[WHITE]?1:0) + (st.CK[WHITE]?1:0) +
      " " + (st.CQ[BLACK]?1:0) + (st.CK[BLACK]?1:0);
  }

  public synchronized String read_extra_info(MailBoxGameState st0, StringStream is)
  {
    ChessGameState st = (ChessGameState) st0;

    st.ep_square = -1;

    String eps = is.get_string();
    Vector<MovePart> v = new Vector<MovePart>();

    if (eps == null || parse_move(eps, v) != null || v.size() != 1 ||
	!v.elementAt(0).rest.equals(""))
      return "illegal ep square";

    st.ep_square = v.elementAt(0).square;

    int b;

    b = is.get_int_skip_ws(); st.CK[WHITE] = b != 0;
    b = is.get_int_skip_ws(); st.CQ[WHITE] = b != 0;
    b = is.get_int_skip_ws(); st.CK[BLACK] = b != 0;
    b = is.get_int_skip_ws(); st.CQ[BLACK] = b != 0;

    if (!is.ok()) return "read_extra_info: error in castling status";
    return null;
  }

  private static int color_of(int cont)
  {
    if (cont >= W_PAWN) return PlayerColor.WHITE;
    if (cont <= B_PAWN) return PlayerColor.BLACK;
    return PlayerColor.UNDEF;
  }
  
  private byte make_cont(byte pc, int col)
  {
    if (pc == EMPTY) return EMPTY;
    int sc = pc;
    if (col == WHITE) return (byte)sc; else return (byte)(-sc);
  }


  // lower left
  private int ll_sq() { return xy2ind(0, 0); }

  // lower right
  private int lr_sq() { return xy2ind(WIDTH-1, 0); }

  // upper left
  private int ul_sq() { return xy2ind(0, WIDTH-1); }

  // upper right
  private int ur_sq() { return xy2ind(WIDTH-1, WIDTH-1); }

  private int l_sq(int col) {
    if (col == WHITE) return ll_sq(); else return ul_sq();
  }
  
  public int r_sq(int col)
  {
    if (col == WHITE) return lr_sq(); else return ur_sq();
  }
  
  // weak test only
  
  public synchronized MailBoxGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;
    if (v.size() != 2) return null;

    int from=PASS, to=PASS, prom=NO_PROM;
    
    MovePart mp = v.elementAt(0);
    if (!mp.rest.equals("")) return null;
    from = mp.square;

    mp = v.elementAt(1);

    if (!mp.rest.equals("")) {
      char c = Character.toUpperCase(mp.rest.charAt(0));
      switch (c) {
      case 'Q': prom = PROM_TO_QUEEN;  break;
      case 'R': prom = PROM_TO_ROOK;   break;
      case 'B': prom = PROM_TO_BISHOP; break;
      case 'N': prom = PROM_TO_KNIGHT; break;
      default: return null;
      }
    }

    to = mp.square;

    // check move

    ChessGameState st = (ChessGameState)(current_state().createCopy());
    
    int moved_piece = st.sq[from];
    int col = color_of(moved_piece);
    int victim = st.sq[to];

    if (color_of(moved_piece) != st.to_move) return null;

    // special case: queen side castling on 6x6 ("capture own rook")
    if (victim != EMPTY && color_of(victim) == st.to_move) {
      if (to != r_sq(col)) return null;
      victim = EMPTY;
    }

    if (victim != EMPTY || Math.abs(st.sq[from]) == PAWN) {

      // irreversible move

      st.rev_move_num = 0;

    } else {

      // reversible move
    
      ++st.rev_move_num;
    }
    
    // check if castable rook is captured: update castling status

    if (Math.abs(victim) == ROOK) {
      if (victim == W_ROOK) {
	if      (to == ll_sq()) st.CQ[WHITE] = false;
	else if (to == lr_sq()) st.CK[WHITE] = false;
      } else {
	if      (to == ul_sq()) st.CQ[BLACK] = false;
	else if (to == ur_sq()) st.CK[BLACK] = false;
      }
    }

    
    if (Math.abs(moved_piece) == KING && Math.abs(from-to) == 2) {

      // castling

      if (to > from) {

	// king side
	
	st.sq[r_sq(col)] = EMPTY;
	st.sq[to-1] = make_cont(ROOK, col);
      } else {

	// queen side
	
	st.sq[l_sq(col)] = EMPTY;
	st.sq[to+1] = make_cont(ROOK, col);
      }
      st.CQ[col] = false;
      st.CK[col] = false;
    }

    if (Math.abs(moved_piece) == PAWN && to == st.ep_square) {

      // ep

      if (st.to_move == BLACK) st.sq[to+DX] = EMPTY;
      else                     st.sq[to-DX] = EMPTY;
    }

    st.sq[to]    = st.sq[from];
    st.sq[from]  = EMPTY;
    st.ep_square = -1;

    switch (prom) {

    case NO_PROM:

      switch (Math.abs(moved_piece)) {

      case PAWN:

	if (Math.abs(to-from) == 2*DX)
	  st.ep_square = (to+from) / 2;

	break;

      case ROOK:

	if (col == WHITE) {

	  if (from == ll_sq()) {
	    st.CQ[col] = false;
	  } else if (from == lr_sq()) {
	    st.CK[col] = false;
	  }

	} else {

	  if (from == ul_sq()) {
	    st.CQ[col] = false;
	  } else if (from == ur_sq()) {
	    st.CK[col] = false;
	  }
	}
	break;

      case KING:

	st.CQ[col] = false;
	st.CK[col] = false;
	break;

      default: ;
      }
      break;

    case PROM_TO_KNIGHT:
      st.sq[to] = make_cont(KNIGHT, col);
      break;

    case PROM_TO_BISHOP:
      st.sq[to] = make_cont(BISHOP, col);
      break;

    case PROM_TO_ROOK:
      st.sq[to] = make_cont(ROOK, col);
      break;

    case PROM_TO_QUEEN:
      st.sq[to] = make_cont(QUEEN, col);
      break;

    default: throw new Error("unknown prom");

    }

    st.to_move = PlayerColor.opponent(st.to_move);
    return st;
  }

  protected String extra_info_to_string(MailBoxGameState st) {
    return "";
  }
  
  public boolean paint_all(Graphics g, boolean repaint,
			   int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    return false;
  }  

}


