/* $Id: Domineering.java,v 1.10 2006/08/30 08:52:09 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa._domineering;

import java.util.*;
import java.awt.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public class Domineering extends BoardGameServiceClient
{
  public Domineering() { super(new DomineeringGame()); }
  
  public void setDefaultOptions() 
  {
    new DomineeringGame().setDefaultOptions();
  }

}


class DomineeringGame extends MailBoxBoardGame
{
  public String get_name()    { return "Domineering"; }  
  public String get_base_name()    { return "dm"; }
  public BoardGame new_BoardGame() { return new DomineeringGame(); }

  public void setDefaultOptions()
  {
    String sn = get_base_name() + ".";

    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color / grey!
    Global.options.put(sn+"gc",  new Integer(128*65536+128*128+128));  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));         // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));         // show last move (boolean)

    // regular empty square color
    Global.options.put(sn+"es0", new Integer(200*65536 + 195 * 256 + 101));
    Global.options.put(sn+"es1", new Integer(200*65536 + 195 * 256 + 101));

    // anti empty square color - not used
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));

    Global.options.put(sn+"cm", "");
  }


  // gfx content

  public static final int GFX_BLACK  = 4;
  public static final int GFX_WHITE  = 8;
  public static final int GFX_BORDER = 16;
  public static final int GFX_EMPTY  = 32;

  public static final int MAX_BOARD_WIDTH = 26;
  public static final int MIN_BOARD_WIDTH = 4;

  protected Color col_white2 = new Color(255,255,204);

  protected MailBoxGameState new_MailBoxGameState()
  {
    return new MailBoxGameState();
  }

  public boolean has_synchro()     { return true; }
  public boolean has_komi()        { return true; }
  public boolean has_rand()        { return true; }
  public boolean has_rand_type()   { return true; }
  public boolean has_anti()        { return true; }
  public boolean has_pref_color()  { return true; }
  public boolean has_pass()        { return false; }

  public boolean never_turn()      { return true; }
  public boolean turned_if_black() { return false; }
  protected boolean y_flip()       { return false; }

  public int board_type_num()      { return 12; }
  public int default_board_type()  { return 8; }

  public int board_type(int index)
  {
    switch (index) {
    case 0:  return 4;
    case 1:  return 6;
    case 2:  return 8;
    case 3:  return 10;
    case 4:  return 12;
    case 5:  return 14;
    case 6:  return 16;
    case 7:  return 18;
    case 8:  return 20;
    case 9:  return 22;
    case 10: return 24;
    case 11: return 26;
    default: return -1;
    }
  }

  public int board_type_index(int type) {
    switch (type) {
    case 4:  return 0;
    case 6:  return 1;
    case 8:  return 2;
    case 10: return 3;
    case 12: return 4;
    case 14: return 5;
    case 16: return 6;
    case 18: return 7;
    case 20: return 8;
    case 22: return 9;
    case 24: return 10;
    case 26: return 11;
    default: return -1;
    }
  }

  public String board_type_string(int index) {
    switch (index) {
    case 0:  return "4x4";
    case 1:  return "6x6";
    case 2:  return "8x8";
    case 3:  return "10x10";
    case 4:  return "12x12";
    case 5:  return "14x14";
    case 6:  return "16x16";
    case 7:  return "18x18";
    case 8:  return "20x20";
    case 9:  return "22x22";
    case 10: return "24x24";
    case 11: return "26x26";
    default: return "?";
    }
  }

  public int width(int type) { return type; }

  public int height(int type) { return width(type); }

  public int border_width() { return 1; }

  public byte char2cont(char c)
  {
    switch (c) {
    case '*': return BLACK;
    case 'O': return WHITE;
    case '-': return EMPTY;
    default:
      throw new Error("illegal char in Domineering.char2cont");
    }
  }

  public char cont2char(byte co)
  {
    switch (co) {
    case BLACK: return '*';
    case WHITE: return 'O';
    case EMPTY: return '-';
    default:
      throw new Error("illegal cont in Domineering.cont2char");
    }
  }

  public synchronized void input(String s) { }

  private boolean is_legal(MailBoxGameState state, int ind)
  {
    if( state.to_move == PlayerColor.BLACK ) // vertical
      return state.sq[ind] == EMPTY && state.sq[ind+S] == EMPTY;
    else // horizontal
      return state.sq[ind] == EMPTY && state.sq[ind+E] == EMPTY;
  }

  private int cont_num(MailBoxGameState state, int cont)
  {
    int n = 0;

    for (int i=0; i < BOARD_SIZE; ++i)
      if (state.sq[i] == cont) ++n;

    return n;
  }

  public MailBoxGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;
    if (v.size() != 1) return null;

    MovePart mp = v.elementAt(0);

    if (!mp.rest.equals("")) return null;

    int ind = mp.square;
    MailBoxGameState new_state = current_state().createCopy();

    if (!is_legal(new_state, ind)) return null;

    if (new_state.to_move == PlayerColor.BLACK) // vertical
      new_state.sq[ind] = new_state.sq[ind+S] = BLACK;
    else
      new_state.sq[ind] = new_state.sq[ind+E] = WHITE;

    new_state.to_move = PlayerColor.opponent(new_state.to_move);
    return new_state;
  }

  protected void process_mouse_events() 
  {
    boolean ok = false;
    int square = -1;

    synchronized (mev) {

      int n = mev.size();

      for (MouseEvent m : mev) {
	if (m.type == CLICKED) {
	  ok = true;
	  square = m.square;
	  reset_mev();
	  break;
	}
      }
    }

    if (ok) {
      handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
					ind2string(square)));
    }
  }

  protected int last_move_index(String move)
  {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null) return -1;
    return v.elementAt(0).square;
  }

  public int gfx_cont(MailBoxGameState st, int i) {

    MailBoxGameState state = st;

    byte cont = state.sq[i];
    int gfx;

    if( cont == BORDER )
      gfx = GFX_BORDER;
    else if( cont == BLACK )
      gfx = GFX_BLACK;
    else if( cont == WHITE )
      gfx = GFX_WHITE;
    else
      gfx = GFX_EMPTY;

    return gfx;
  }

  protected boolean entire_gfx_cont(MailBoxGameState st, String last_move) {
    return false;
  }


  public boolean paint_all(Graphics g, boolean repaint,int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    return false;
  }


  protected void init_special_border(MailBoxGameState state) {}

  public boolean only_squares() { return true; }

  public boolean pieces_on_grid() { return false; }

  public String info_string() {

    int bd = cont_num((MailBoxGameState)current_state(), BLACK)
      + cont_num((MailBoxGameState)current_state(), WHITE);
    return "ply " + (bd/2+1);
  }

  protected String extra_info_to_string(MailBoxGameState st) {
    return "";
  }

  public synchronized String read_extra_info(MailBoxGameState st0, StringStream is)
  {
    return null;
  }

  protected void paint_gfx_cont(Graphics g, int x, int y, int parity, int gfx_cont)
  {
    if( (gfx_cont & GFX_BORDER) != 0) return;

    if (anti) {
      
      if (parity == 0) g.setColor(col_aempty0); else g.setColor(col_aempty1);

    } else {

      if (parity == 0) g.setColor(col_empty0); else g.setColor(col_empty1);

    }
    
    g.fillRect(x+1,y+1,gfx_sq_w-1, gfx_sq_h-1);

    
    if( (gfx_cont & GFX_BLACK) != 0 )
      g.setColor(col_black);
    else if( (gfx_cont & GFX_WHITE) != 0 )
      g.setColor(col_white); // changed to white (mic)
    else {
      if (anti) 
	g.setColor(col_aempty0);
      else
	g.setColor(col_empty0);
    }

    g.fillRect(x+1,y+1,gfx_sq_w-1, gfx_sq_h-1);

    if ((gfx_cont & GFX_LAST_MOVE) != 0) {

      // last move: draw inner square
      
      int dot_diam = gfx_sq_w/4;
      dot_diam &= ~1;
      
      if ((gfx_cont & GFX_WHITE) != 0)
	g.setColor(col_black);
      else 
	g.setColor(col_white);
      
      int xl = x + (gfx_sq_w - dot_diam)/2;
      int yt = y + (gfx_sq_h - dot_diam)/2;
      g.drawRect(xl, yt, dot_diam, dot_diam);
    }
    
  }

}
