/* $Id: DotsAndBoxes.java,v 1.11 2006/09/04 17:47:38 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/


package ggs.ggsa._dotsandboxes;

import java.util.*;
import java.awt.*;
import ggs.ggsa.boardgamesvc.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public class DotsAndBoxes extends BoardGameServiceClient
{
  public DotsAndBoxes() { super(new DotsAndBoxesGame()); }
  
  public void setDefaultOptions() 
  {
    new DotsAndBoxesGame().setDefaultOptions();
  }

}


class DotsAndBoxesGame extends EdgeBoardGame
{
  public String get_name()    { return "DotsAndBoxes"; }  
  public String get_base_name()    { return "db"; }
  public BoardGame new_BoardGame() { return new DotsAndBoxesGame(); }

  public void setDefaultOptions() {

    String sn = get_base_name() + ".";
    
    super.set_default_options();
    
    Global.options.put2i(sn+"ipw", BGS+"ipw"); // info panel
    Global.options.put2i(sn+"tph", BGS+"tph"); // game window vertical split

    Global.options.put2i(sn+"ge1", BGS+"ge1"); // 3 board window geoms
    Global.options.put2i(sn+"ge2", BGS+"ge2");
    Global.options.put2i(sn+"ge3", BGS+"ge3"); // also synchro geometry

    // playfield and grid color
    Global.options.put2i(sn+"gc",  BGS+"gc");  // grid color

    Global.options.put(sn+"mo",  new Boolean(true));  // show moves (boolean)
    Global.options.put(sn+"la",  new Boolean(true));  // show last move (boolean)

    // regular empty square color
    Global.options.put(sn+"es0", new Integer(0*65536  + 178*256 + 238));
    Global.options.put(sn+"es1", new Integer(205*65536  + 133*256 + 63));

    // last move square color
    Global.options.put(sn+"lms",  new Integer(255*65536 + 0*256 + 0));

    // anti empty square color (not used)
    Global.options.put(sn+"aes0", new Integer(255*65536 + 114 * 256 + 86));
    Global.options.put(sn+"aes1", new Integer(255*65536 + 114 * 256 + 86));    

    Global.options.put(sn+"cm", "");
  }



  // square content (also edges!)
  protected static final byte OCCUPIED_B = 4;
  protected static final byte OCCUPIED_W = 5;
  protected static final byte CONT_MASK  = 7;
  
  // gfx content & flags

  protected static final int GFX_BLACK  = 2;  
  protected static final int GFX_WHITE  = 3;
  protected static final int GFX_CONT_MASK = 3; 
  
  protected static final int GFX_NODE   = 4;  
  protected static final int GFX_EDGE_H = 8;
  protected static final int GFX_EDGE_V = 16;    
  protected static final int GFX_BOX    = 32;

  
  protected EdgeGameState new_EdgeGameState()
  {
    return new EdgeGameState();
  }
  
  public boolean has_synchro()     { return true;  }
  public boolean has_komi()        { return true;  }
  public boolean has_rand()        { return true; }
  public boolean has_rand_type()   { return true; }
  public boolean has_anti()        { return false; }  
  public boolean has_pref_color()  { return true;  }    
  public boolean has_pass()        { return false; } 

  public boolean never_turn()        { return false; }
  public boolean turned_if_black()   { return false; }
  protected boolean y_flip()         { return false; }
  protected boolean only_squares()   { return true;  }

  protected int border_width()     { return 1; }

  public int board_type_num()      { return 11; }
  public int default_board_type()  { return 5; }  // 6x6

  public int board_type(int index)
  { 
    return index+3;  // 3..13
  } 

  public int board_type_index(int type) {
    return type-2;
  }
  
  public String board_type_string(int index) {
    return ""+board_type(index);
  }

  public int width(int type) { return type; }
  
  public int height(int type) { return width(type); }

  public char node_char() { return '+'; }
  
  final static private char C_EMPTY_CORNER = '.';
  final static private char C_EMPTY_SQUARE = ':';  
  
    public byte h_char2cont(char c)
  {
    switch (c) {
    case C_EMPTY_CORNER: return EMPTY;
    case '-': return OCCUPIED_B;
      default: throw new Error("illegal char in h_char2cont: " + c);
    }
  }
  
  public char h_cont2char(byte cont)
  {
    switch (cont & CONT_MASK) {
    case EMPTY: return C_EMPTY_CORNER;
    case OCCUPIED_B: return '-';
      default: throw new Error("illegal cont in h_cont2char: " + cont);
    }
  }

  public byte v_char2cont(char c)
  {
    switch (c) {
    case C_EMPTY_CORNER: return EMPTY;
    case '|': return OCCUPIED_B;
      default: throw new Error("illegal char in v_char2cont: " + c);
    }
  }
  
  public char v_cont2char(byte cont)
  {
    switch (cont & CONT_MASK) {
    case EMPTY: return C_EMPTY_CORNER;
    case OCCUPIED_B:  return '|';
      default: throw new Error("illegal cont in v_cont2char: " + cont);
    }
  }

  public byte s_char2cont(char c)
  {
    switch (c) {
    case C_EMPTY_SQUARE: return EMPTY;
    case '*': return OCCUPIED_B;
    case 'O': return OCCUPIED_W;
      default: throw new Error("illegal char in s_char2cont: "+ c);
    }
  }

  public char s_cont2char(byte cont)
  {
    switch (cont & CONT_MASK) {
    case EMPTY:      return C_EMPTY_SQUARE;
    case OCCUPIED_B: return '*';
    case OCCUPIED_W: return 'O';
      default: throw new Error("illegal cont in s_cont2char: " + cont);
    }
  }

  public synchronized void input(String s) { }

  protected synchronized void process_mouse_events() 
  {
    boolean ok = false;
    int square = -1;
    
    synchronized (mev) {
      
      int n = mev.size();

      for (MouseEvent m : mev) {
        if (m.type == CLICKED) {
          ok = true;
          square = m.square;
          reset_mev();
	  break;
        }
      }
    }

    if (ok) {
      handler.handle_event(new EventMsg(GameWindow.USER_MOVE_ID,
                                        ind2string(square)));
    }
  }

  protected synchronized boolean entire_gfx_cont(EdgeGameState st0,
						 String last_move)
  {
    return false;
  }

  protected synchronized int last_move_index(String move)
  {
    Vector<MovePart> v = new Vector<MovePart>();
    if (parse_move(move, v) != null) return -1;
    return v.elementAt(0).square;
  }

  public int gfx_cont(EdgeGameState st, int i)
  {
    int gfxco = GFX_EMPTY;
    byte cont = st.sq[i];

    if      ((cont & CONT_MASK) == OCCUPIED_B) gfxco |= GFX_BLACK;
    else if ((cont & CONT_MASK) == OCCUPIED_W) gfxco |= GFX_WHITE;

    switch (cont & KIND_MASK) {
    case NODE:   gfxco |= GFX_NODE;   break;
    case FACE:   gfxco |= GFX_BOX;    break;
    case EDGE_V: gfxco |= GFX_EDGE_V; break;
    case EDGE_H: gfxco |= GFX_EDGE_H; break;      
    default: throw new Error("unknown square kind");
    }

    return gfxco;
  }
  
  protected synchronized void paint_gfx_cont(Graphics g,
					     int x, int y, int parity, int gfx_cont)
  {
    final double BOX_FRAC  = 0.95;
    final double NODE_FRAC = 0.6;    

    if (gfx_cont == GFX_BORDER) return;

    int gfx_w = gfx_sq_w & ~1;
    int gfx_h = gfx_sq_h & ~1;    
    int box_width = (int)Math.round(gfx_w * BOX_FRAC) & ~1;
    int node_diam = (int)Math.round(gfx_w * NODE_FRAC) & ~1;    
    String sn = get_base_name() + ".";
    Color grid_color = new Color(Global.options.getInteger(sn+"gc").intValue());
    
    // fill square background
    
    g.setColor(col_empty0);
    g.fillRect(x,y,gfx_sq_w+1, gfx_sq_h+1);

    if ((gfx_cont & GFX_NODE) != 0) {

      // draw node

      int xl = x + (gfx_w - node_diam)/2;
      int yt = y + (gfx_h - node_diam)/2;

      g.setColor(grid_color);
      Bresenham.fillCircle(g, xl, yt, node_diam);
    }

    if ((gfx_cont & GFX_BOX) != 0) {

      // draw box

      int xl = x + (gfx_w - box_width)/2;
      int yt = y + (gfx_h - box_width)/2;

      Color f_col, b_col;

      if ((gfx_cont & GFX_CONT_MASK) == GFX_BLACK) {
	f_col = col_black; b_col = col_white;
      } else if ((gfx_cont & GFX_CONT_MASK) == GFX_WHITE) {
	f_col = col_white; b_col = col_black;
      } else {
	f_col = b_col = col_empty0;
      }
      
      g.setColor(f_col);
      g.fillRect(xl, yt, box_width, box_width);

      g.setColor(b_col);
      g.drawRect(xl, yt, box_width, box_width);
    }

    if ((gfx_cont & GFX_EDGE_H) != 0 && (gfx_cont & GFX_CONT_MASK) == GFX_BLACK) {

      if ((gfx_cont & GFX_LAST_MOVE) != 0)
	g.setColor(new Color(Global.options.getInteger(sn+"lms").intValue()));
      else
	g.setColor(grid_color);

      g.drawLine(x, y+gfx_h/2, x+gfx_w, y+gfx_h/2);
      g.drawLine(x, y+gfx_h/2+1, x+gfx_w, y+gfx_h/2+1);	
    }

    if ((gfx_cont & GFX_EDGE_V) != 0 && (gfx_cont & GFX_CONT_MASK) == GFX_BLACK) {

      if ((gfx_cont & GFX_LAST_MOVE) != 0)
	g.setColor(new Color(Global.options.getInteger(sn+"lms").intValue()));
      else
	g.setColor(grid_color);
      
      g.drawLine(x+gfx_w/2, y, x+gfx_w/2, y+gfx_h);
      g.drawLine(x+gfx_w/2+1, y, x+gfx_w/2+1, y+gfx_h);	
    }
  }

  protected void init_special_border(EdgeGameState state) { }

  public synchronized String info_string() {
    return "B: " + box_count(BLACK) + " W: " + box_count(WHITE);
  }

  public synchronized String read_extra_info(EdgeGameState st0, StringStream is) 
  {
    return null;
  }

  private boolean occupied(int sq) {
    return (sq & CONT_MASK) == OCCUPIED_B;
  }
  
  public synchronized EdgeGameState make_move2(String mv)
  {
    // parse move

    Vector<MovePart> v = new Vector<MovePart>();
    String err = parse_move(mv, v);

    if (err != null) return null;
    if (v.size() != 1) return null;

    MovePart mp = v.elementAt(0);
    if (!mp.rest.equals("")) return null;

    int move = mp.square;

    // check move

    int x = ind2x(move);
    int y = ind2y(move);

    if (x < 0 || x >= WIDTH) return null;
    if (y < 0 || y >= WIDTH) return null;

    if (((x+y) & 1) != 1) return null;

    EdgeGameState st = (EdgeGameState)(current_state().createCopy());

    if ((st.sq[move] & CONT_MASK) != EMPTY) return null;

    st.sq[move] = (byte)(OCCUPIED_B | (st.sq[move] & KIND_MASK));

    byte box_cont;
    if (st.to_move == BLACK) box_cont = OCCUPIED_B; else box_cont = OCCUPIED_W;
    box_cont |= FACE;
    
    // box test

    boolean new_box = false;
  
    if ((y & 1) == 0) {

      //  o-o
    
      if (occupied(st.sq[move+DX-1]) &&
	  occupied(st.sq[move+DX+1]) &&
	  occupied(st.sq[move+2*DX])) {

	// o-o
	//  X
      
	st.sq[move+DX] = box_cont;
	new_box = true;
      }

      if (occupied(st.sq[move-DX-1]) && 
	  occupied(st.sq[move-DX+1]) &&
	  occupied(st.sq[move-2*DX])) {

	//  X
	// o-o

	st.sq[move-DX] = box_cont;
	new_box = true;
      }
    
    } else {

      //  o
      //  |
      //  o
    
      if (occupied(st.sq[move-DX+1]) &&
	  occupied(st.sq[move+2]) &&
	  occupied(st.sq[move+DX+1])) {

	//  o
	//  |X
	//  o

	st.sq[move+1] = box_cont;
	new_box = true;
      }

      if (occupied(st.sq[move-DX-1]) && 
	  occupied(st.sq[move-2]) &&
	  occupied(st.sq[move+DX-1])) {

	//   o
	//  X|
	//   o

	st.sq[move-1] = box_cont;
	new_box = true;
      }
    }

    if (!new_box) st.to_move = PlayerColor.opponent(st.to_move);
    return st;
  }

  private int box_count(int col) {

    int n = 0;
    int box = OCCUPIED_B;
    EdgeGameState st = (EdgeGameState)current_state();
    
    if (col == WHITE) box = OCCUPIED_W;

    for (int i=0; i < BOARD_SIZE; ++i) {
      byte cont = st.sq[i];
      if (cont != BORDER &&
	  (cont & KIND_MASK) == FACE &&
	  (cont & CONT_MASK) == box) n++;
    }
    return n;
  }

  protected String extra_info_to_string(EdgeGameState st) { return ""; }
  
  public boolean paint_all(Graphics g, boolean repaint,
			   int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    return false;
  }  
  
}
