// $Id: CardPanel.java,v 1.2 2006/09/04 03:40:42 orts_mburo Exp $

package ggs.ggsa._skat.gui;

import ggs.ggsa.main.Global;

import java.awt.Graphics;
import java.util.Observable;
import java.util.Observer;

import javax.swing.JPanel;

/**
 * CardPanel for showing a card
 * 
 * @author Jan Sch&auml;fer <j@nschaefer.net>
 */
public class CardPanel extends JPanel implements Observer {

  /**
   * Creates a new instance of CardPanel
   * 
   * @param holdingPanel
   *            The CardHoldingPanel that groups together cards on a hand or
   *            in skat or trick
   * @param showBack
   *            TRUE if the card should hide its face
   */
  public CardPanel(CardHoldingPanel holdingPanel,
		   SkatGraphicRepository bitmaps,
		   boolean showBack)
  {
    this.bitmaps = bitmaps;
    this.cardHoldingPanel = holdingPanel;
    this.showBack = showBack;
  }

  /**
   * Overwrites the paintComponent method of JPanel draws the card image on
   * the CardPanel
   * 
   * @param g
   *            The Graphics environment
   */
  protected void paintComponent(Graphics g)
  {
    super.paintComponent(g);

    if (suit >= 0 && value >= 0) {

      if (showBack) {

	g.drawImage(bitmaps.getCardImage(-1, -1), 0, 0, this);

      } else {

	g.drawImage(bitmaps.getCardImage(suit, value), 0, 0, this);
      }
    }
  }

  /**
   * Sets the card suit and value
   * 
   * @param newSuit
   *            ID of the suit
   * @param newValue
   *            value of the card
   */
  public void setCard(int newSuit, int newValue)
  {
    suit = newSuit;
    value = newValue;
  }

  /**
   * Flips the card
   */
  public void flipCard()
  {
    showBack = !showBack;
  }

  /**
   * Gets the suit of the card
   * 
   * @return ID of the suit
   */
  public int getSuit()
  {
    return suit;
  }

  /**
   * Gets the value of the card
   * 
   * @return The value of the card
   */
  public int getValue()
  {
    return value;
  }

  /**
   * Tells the JSkatMaster when the panel was clicked by the user
   */
  public void cardClicked()
  {
    cardHoldingPanel.cardPanelClicked(suit, value);
  }

  /**
   * 
   */
  public void update(Observable observ, Object obj)
  {
    Global.dbgmsg("UPDATE " + observ + ": " + obj + " has changed...");

    if (observ instanceof SkatGraphicRepository) {
      repaint();
    }
  }

  private CardHoldingPanel cardHoldingPanel;
  private SkatGraphicRepository bitmaps;
  private int suit = -1;
  private int value = -1;
  private boolean showBack = true;
}
