// $Id: SkatGraphicRepository.java,v 1.2 2006/09/04 03:40:42 orts_mburo Exp $

package ggs.ggsa._skat.gui;

import ggs.ggsa.main.Global;

import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.awt.Canvas;
import java.util.Observable;
import java.util.Observer;

/**
 * The SkatGraphicRepository that holds all images used in Skat games
 * 
 * @author Jan Sch&auml;fer <j@nschaefer.net>
 */
public class SkatGraphicRepository extends Observable implements Observer
{

  /** Creates a new instance of JSkatGraphicRepository */
  public SkatGraphicRepository() {

    cardFace = CardFaces.FRENCH;
    tracker = new MediaTracker(new Canvas());
    cards = new Image[4][8];

    loadCards(cardFace);
  }

  /**
   * Loads specific card face
   * 
   * @param cardFace
   */
  private void loadCards(CardFaces cardFace) {

    if (cardFace == CardFaces.FRENCH) {

      loadCards("french");
    } else if (cardFace == CardFaces.TOURNAMENT) {

      loadCards("tournament");
    } else {

      loadCards("french");
    }
		
    Global.dbgmsg("Bitmaps for cards loaded...");
  }

  /**
   * Load all card images
   * 
   * @param cardType
   *            The directory name for the card set to be loaded
   */
  public void loadCards(String cardType) {

    String suitString = new String();
    String valueString = new String();

    int i = 0;

    for (CardSuits suit : CardSuits.values()) {

      switch (suit) {

      case CLUBS:
	suitString = "clubs_";
	break;

      case SPADES:
	suitString = "spades_";
	break;

      case HEARTS:
	suitString = "hearts_";
	break;

      case DIAMONDS:
	suitString = "diamonds_";
	break;
      }

      int j = 0;

      for (CardValues value : CardValues.values()) {

	switch (value) {

	case SEVEN:
	  valueString = "07";
	  break;

	case EIGHT:
	  valueString = "08";
	  break;

	case NINE:
	  valueString = "09";
	  break;

	case TEN:
	  valueString = "10";
	  break;

	case JACK:
	  valueString = "jack";
	  break;

	case QUEEN:
	  valueString = "queen";
	  break;

	case KING:
	  valueString = "king";
	  break;

	case ACE:
	  valueString = "ace";
	  break;
	}

	cards[i][j] = Toolkit.getDefaultToolkit().getImage(ClassLoader.getSystemResource("data/images/cards/" + cardType + "/" + suitString + valueString + ".gif"));
	tracker.addImage(cards[i][j], 2);

	j++;
      }
			
      i++;
    }

    cardBack = Toolkit.getDefaultToolkit().getImage(ClassLoader.getSystemResource("data/images/cards/" + cardType + "/back.gif")); 
    tracker.addImage(cardBack, 2);

    try {
      tracker.waitForID(2);
    }
    catch (InterruptedException e) { }
  }

  /**
   * Gets the card image
   * 
   * @param suit
   *            The suit of the card
   * @param value
   *            The value of the card
   * @return The card image
   */
  public Image getCardImage(int suit, int value)
  {
    if (suit >= 0 && value >= 0) {

      try {
	return cards[suit][value];
      }

      catch (IndexOutOfBoundsException exc) {
	return null;
      }

    } else {

      return cardBack;
    }
  }

  /**
   * Gets the image for the skat table
   * 
   * @return The image for the skat table
   */
  public Image getSkatTableImage()
  {
    return skatTable;
  }

  /**
   * Implementation of the Observer pattern
   * 
   * @param observ
   *            The observed object
   * @param obj
   *            The changed context
   */
  public void update(Observable observ, Object obj)
  {
    System.out.println("UPDATE " + observ + " " + obj + " has changed...");
  }

  private CardFaces cardFace;
  private MediaTracker tracker;
  private Image skatTable;
  private Image cards[][];
  private Image cardBack;
  private enum CardFaces { FRENCH, TOURNAMENT }
  private enum CardSuits { CLUBS, SPADES, HEARTS, DIAMONDS }
  private enum CardValues { SEVEN, EIGHT, NINE, TEN, JACK, QUEEN, KING, ACE }
}
