// $Id: SkatPlayArea.java,v 1.3 2006/09/04 03:40:42 orts_mburo Exp $

package ggs.ggsa._skat.gui;

import ggs.ggsa.main.Global;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.CardLayout;
import javax.swing.JPanel;

import java.util.Observer;
import java.util.Observable;

/**
 * The SkatPlayArea that holds all CardHoldingPanels and the BiddingPanel
 * 
 * @author Jan Sch&auml;fer <j@nschaefer.net>
 */
public class SkatPlayArea extends JPanel implements Observer
{

  /**
   * Creates a new instance of SkatPlayArea
   */
  public SkatPlayArea()
  {
    initComponents();
  }

  private void initComponents()
  {
    setBackground(new Color(0, 96, 0));
    setLayout(new GridLayout(3, 1));
    setPreferredSize(new Dimension(800, 500));

    bitmaps = new SkatGraphicRepository();
		
    opponentPanel = new JPanel();
    opponentPanel.setLayout(new GridLayout(1, 2));
    opponentPanel.setOpaque(false);
    opponentPanel.add(new CardHoldingPanel(bitmaps, 0,
					   CardHoldingPanel.PanelTypes.NETWORK_PLAYER_PANEL, 10));
    opponentPanel.add(new CardHoldingPanel(bitmaps, 1,
					   CardHoldingPanel.PanelTypes.NETWORK_PLAYER_PANEL, 10));
    add(opponentPanel);

    skatTrickHoldingPanel = new JPanel();
    skatTrickHoldingPanel.setOpaque(false);
    skatTrickHoldingPanel.setLayout(new CardLayout());
    trickPanel = new JPanel();
    trickPanel.setOpaque(false);
    trickPanel.add(new CardHoldingPanel(bitmaps, -1,
					CardHoldingPanel.PanelTypes.TRICK_PANEL, 3));
    skatTrickHoldingPanel.add(trickPanel, "trick");
    skatPanel = new JPanel();
    skatPanel.setOpaque(false);
    skatPanel.add(new CardHoldingPanel(bitmaps, -1,
				       CardHoldingPanel.PanelTypes.SKAT_PANEL, 4));
    skatTrickHoldingPanel.add(skatPanel, "skat");
    bidPanel = new BiddingPanel();
    skatTrickHoldingPanel.add(bidPanel, "bidding");
    add(skatTrickHoldingPanel);

    playerPanel = new JPanel();
    playerPanel.setLayout(new GridLayout(1, 1));
    playerPanel.setOpaque(false);
    playerPanel.add(new CardHoldingPanel(bitmaps, 2,
					 CardHoldingPanel.PanelTypes.PLAYER_PANEL, 10));
    add(playerPanel);
  }

  /**
   * Overwrites the paintComponent method of JPanel draws the wooden table in
   * the background
   * 
   * @param g
   *            The Graphics environment
   */
  protected void paintComponent(Graphics g)
  {
    super.paintComponent(g);
    g.drawImage(bitmaps.getSkatTableImage(), 0, 0, this);
  }

  /**
   * Implementation of the Observer pattern
   * 
   * @param observ
   *            The Observable that is observed
   * @param obj
   *            The Object that has changed in the Observable
   */
  public void update(Observable observ, Object obj)
  {
    Global.dbgmsg("UPDATE " + observ + ": " + obj + " has changed...");
  }

  /**
   * Flips all cards on the JSkatPlayArea
   */
  public void flipCards() {

    getCardHoldingPanel(PanelTypes.PLAYER_ZERO).flipCards();
    getCardHoldingPanel(PanelTypes.PLAYER_ONE).flipCards();
  }

  /**
   * Sets all player data
   * 
   * @param playerZero
   *            Name of player of upper left hand
   * @param playerOne
   *            Name of player of upper right hand
   * @param playerZero
   *            Name of player of lower hand
   */
  private void setPlayerData(String playerZero, String playerOne,
			     String playerTwo)
  {
    getCardHoldingPanel(PanelTypes.PLAYER_ZERO).setPlayerName(playerZero);
    getCardHoldingPanel(PanelTypes.PLAYER_ONE).setPlayerName(playerOne);
    getCardHoldingPanel(PanelTypes.PLAYER_TWO).setPlayerName(playerTwo);
  }

  /**
   * Returns the CardHoldingPanel for a player
   * 
   * @param playerID
   *            The player ID
   * @return The CardHoldingPanel
   */
  public CardHoldingPanel getCardHoldingPanel(PanelTypes panelID)
  {
    CardHoldingPanel panel;

    switch (panelID) {

    case PLAYER_ZERO:

      panel = (CardHoldingPanel) opponentPanel.getComponent(0);
      break;

    case PLAYER_ONE:

      panel = (CardHoldingPanel) opponentPanel.getComponent(1);
      break;

    case PLAYER_TWO:

      panel = (CardHoldingPanel) playerPanel.getComponent(0);
      break;

    case SKAT:

      // skat panel
      panel = (CardHoldingPanel) skatPanel.getComponent(0);
      break;

    case TRICK:

      // trick panel
      panel = getTrickPanel();
      break;

    default:
      panel = (CardHoldingPanel) new JPanel();
    }

    return panel;
  }

  public BiddingPanel getBiddingPanel()
  {
    return bidPanel;
  }

  public CardHoldingPanel getTrickPanel()
  {
    return (CardHoldingPanel) trickPanel.getComponent(0);
  }

  private SkatGraphicRepository bitmaps;
  private JPanel opponentPanel;
  private JPanel skatTrickHoldingPanel;
  private JPanel skatPanel;
  private JPanel trickPanel;
  private BiddingPanel bidPanel;
  private JPanel playerPanel;
  private enum PanelTypes { PLAYER_ZERO, PLAYER_ONE, PLAYER_TWO, TRICK, BIDDING, SKAT }
}
