/* $Id: BoardGame.java,v 1.8 2006/08/28 06:57:30 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/


package ggs.ggsa.boardgamesvc;

import java.awt.*;
import javax.swing.*;
import ggs.ggsa.main.*;
import ggs.shared.*;

public abstract class BoardGame
{
  public static final byte BLACK = 0;
  public static final byte WHITE = 1;  

  // mouse actions
  public static final String PRESSED  = "PRESSED";
  public static final String RELEASED = "RELEASED";
  public static final String CLICKED  = "CLICKED";
  
  protected EventHandler handler;
  protected boolean      anti;
  protected JComponent   component; // gfx component to draw on
  
  public void copyOf(BoardGame x)
  {
    handler   = x.handler;
    anti      = x.anti;
    component = x.component;
  }

  // default board game service options

  static public String BGS = "BGS."; // board game service default option prefix
  
  public void set_default_options()
  {
    Global.options.put2(BGS+"ipw", new Integer(150)); // info panel width
    Global.options.put2(BGS+"tph", new Integer(180)); // talk panel height
    Global.options.put2(BGS+"ge1", new Geometry(550,570,205,70)); // 3 board window geoms
    Global.options.put2(BGS+"ge2", new Geometry(703,715,205,70));
    Global.options.put2(BGS+"ge3", new Geometry(408,416,205,70)); // also synchro geometry
    Global.options.put2(BGS+"rge", new Geometry(320, 250, 413, 334)); // request popup geometry
    Global.options.put2(BGS+"gc",  new Integer(0*65536 + 0*256 + 0)); // grid color
  }
  
  public void set_handler(EventHandler handler) { this.handler = handler; }
  public void set_anti(boolean f) { this.anti = f; }

  public abstract boolean has_pass();
  public abstract boolean has_synchro();  
  public abstract boolean has_komi();
  public abstract boolean has_rand();
  public abstract boolean has_rand_type();
  public abstract boolean has_anti();  
  public abstract boolean has_pref_color();    

  public abstract int     board_type_num();
  public abstract int     default_board_type();
  public abstract int     board_type_index(int type);
  public abstract String  board_type_string(int index);
  public abstract int     board_type(int index);

  public abstract String  start_pos(String ggf);
  
  public abstract boolean paint(JComponent component,
				Graphics g, int iw, int ih,
				boolean show_moves,
				boolean show_last_move,				
				String  last_move,
				boolean turned);

  public abstract void    mouse_action(String type, int gfx_x, int gfx_y);
  public abstract void    input(String s);
  
  public abstract String  info_string();
  public abstract int     to_move();

  public abstract String  make_move(String move);

  public abstract boolean turned_if_black();
  public abstract boolean never_turn();

  public abstract boolean advance(int delta);
  public abstract int     get_index();
  
  public abstract BoardGame new_BoardGame();
  public abstract String get_base_name();
  public abstract String get_name();
}
