/* $Id: BoardGameServiceClient.java,v 1.7 2006/09/04 17:25:54 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.boardgamesvc;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.text.*;
import javax.swing.*;

import ggs.ggsa.main.*;
import ggs.shared.*;
 
public abstract class BoardGameServiceClient extends ServiceClient
{
  private int end_id = 0;
  private DateFormat df = new SimpleDateFormat("HH:mm:ss");

  public static final String LOCAL_GAME     = "local";

  // server messages

  public static final String ADD_PREFIX     = "+";
  public static final String REM_PREFIX     = "-";
  public static final String GAME_END_PREFIX = "end";
  public static final String JOIN_PREFIX    = "join";
  public static final String UPDATE_PREFIX  = "update";
  public static final String RATING_UPDATE_PREFIX  = "rating_update";
  public static final String LOOK_PREFIX    = "look";
  public static final String TIMEOUT_PREFIX = "timeout";
  public static final String MATCH_PREFIX   = "match";
  public static final String FATAL_TIMEOUT_PREFIX = "fatal-timeout";
  public static final String COMMENT_PREFIX = "comment";
  public static final String SCORE_COMMENT_PREFIX = "score";
  public static final String WATCH_PREFIX   = "watch";

  // local messages

  public static final String LOAD_PREFIX = "load:";

  private EventFactory ef = null;
  private Vector<GameWindow> windows;

  protected ReqPopUp my_req;
  protected ReqPopUp your_req;

  private JMenu menu;
  private TabPanel tab_panel;

  private String handle_lock = new String();

  private Vector<String> running_game_strings;
  private Vector<String> finished_game_strings;
  private Vector<String> game_list; // shown in tab_panel
  
  private int FINISHED_NUM_MAX = 6;

  private BoardGame bg;

  public BoardGameServiceClient(BoardGame bg)
  {
    this.bg = bg;
  }
  
  class ServiceListMouseListener extends MouseAdapter 
  {
    ServiceListMouseListener(JList list)
    {
      this.list = list;
    }

    public void mouseClicked(MouseEvent e) 
    {
      int index = list.locationToIndex(e.getPoint());
      String selection = (String) list.getModel().getElementAt(index);

      // parse match id

      StringTokenizer st = new StringTokenizer(selection, " ");

      if (!st.nextToken().equals("+"))
	return;
      
      send("t " + getLogin() + " watch + " + st.nextToken());
    }

    private JList list;
  }
  
  public void init(Sender sender, Options opt)
  {
    super.init(sender, opt);
    ef = new EventFactory(getLogin());
    ef.add_handler(this);

    GGS_Game gg = new GGS_Game(bg.new_BoardGame());  // for game type

    my_req   = new ReqPopUp(getLogin(), this, gg, true);
    your_req = new ReqPopUp(getLogin(), this, gg, false);

    // tab panel
    
    tab_panel = new TabPanel();
    JList list = tab_panel.get_list();
    list.removeAll();
    list.invalidate();
    list.addMouseListener(new ServiceListMouseListener(list));
      
    //  Menu

    menu = new JMenu(getName());

    JMenuItem mi = new JMenuItem("Observers");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  sendWithFocus("t " + getLogin() + " watch ?");
	}});

    mi = new JMenuItem("History");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  sendWithFocus("t " + getLogin() + " H");
	}});

    mi = new JMenuItem("Top 8");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  sendWithFocus("t " + getLogin() + " top 8 24");
	}});

    mi = new JMenuItem("Top 8r");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  sendWithFocus("t " + getLogin() + " top 8r 24");
	}});

    mi = new JMenuItem("Who 8");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  sendWithFocus("t " + getLogin() + " who 8 a");
	}});

    mi = new JMenuItem("Who 8r");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  sendWithFocus("t " + getLogin() + " who 8r a");
	}});

    // Match Request

    mi = new JMenuItem("Match Request");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  my_req.make_visible();
	}});

    // set open to 0

    mi = new JMenuItem("Open 0");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  send("t " + getLogin() + " open 0");
	}});

    // set open to 1

    mi = new JMenuItem("Open 1");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  send("t " + getLogin() + " open 1");
	}});

    // set unrated

    mi = new JMenuItem("Unrated");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  send("t " + getLogin() + " rated -");
	}});

    // set rated

    mi = new JMenuItem("Rated");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  send("t " + getLogin() + " rated +");
	}});

    // service help

    mi = new JMenuItem("Help");
    menu.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  sendWithFocus("t " + getLogin() + " help");
	}});
  }

  public void arrived()
  {
    super.arrived();

    windows = new Vector<GameWindow>();
    running_game_strings = new Vector<String>();
    finished_game_strings = new Vector<String>();

    String login = getLogin();
    send("t " + login + " vt100 -");
    send("t " + login + " client +");
    send("t " + login + " open 1");
    send("t " + login + " notify + *");
    send("t " + login + " notify +");
    send("t " + login + " watch +");
    send("t " + login + " match");

    String cmds = getOptions().getString(getBaseName() + ".cm");

    if (cmds == null) {

      Misc.msg(getBaseName() + ".cm not defined!");
      Global.shut_down();

    } else {

      if (cmds.length() > 0) {
	StringTokenizer st = new StringTokenizer(cmds, ";");
	while (st.hasMoreTokens()) {
	  send("t " + getLogin() + " " + st.nextToken());
	}
      }
    }
  }

  public void departed()
  {
    super.departed();

    for (int i=0; i < windows.size(); i++) {
      GameWindow ow = windows.get(i);
      if (ow != null) { ow.shut_down(); windows.setElementAt(null, i); }
    }
    windows = null;
  }

  public JMenu getMenu() { return menu; }

  public JPanel getTabPanel() { return tab_panel; }
  
  private GameWindow getWindow(int index)
  {
    return windows.get(index);
  }

  private void addWindow(GGS_Game gg)
  {
    int index = newIndex();
    GameWindowProp prop = new GameWindowProp(getBaseName(), index, gg);
    GameWindow ow = new GameWindow(this, prop);
    ow.setVisible(true);
    windows.setElementAt(ow, index);
  }

  private void removeWindow(String game_id)
  {
    int index = windowIndex(game_id);

    if (index < 0) {
      Misc.msg("remove_window: window not found id=" + game_id);
      return;
    }

    // leave game if not already ended

    GameWindow ow = windows.get(index);

    if (ow.prop.gg.id().startsWith(".")) {

      if (ow.prop.gg.black_name().equals(getOptions().getString("lo")) ||
	  ow.prop.gg.white_name().equals(getOptions().getString("lo"))) {

	// own game => send break

	send("t " + getLogin() + " break " + ow.prop.gg.id());

      } else {

	send("t " + getLogin() + " watch - " + ow.prop.gg.id());

      }
    }

    ow.shut_down();
    windows.set(index, null);
  }

  private int newIndex()
  {
    for (int i=0; i < windows.size(); i++) {
      GameWindow ow = windows.get(i);
      if (ow == null)
	return i;
    }

    // no empty slots => create new one

    windows.add(null);
    return windows.size()-1;
  }

  private int windowIndex(String game_id)
  {
    //    Misc.msg(windows.size() + " windows");

    for (int i=0; i < windows.size(); i++) {
      GameWindow ow = windows.get(i);
      if (ow == null)
	continue;

      // Misc.msg(ow.prop.gg.game_id + " " + game_id);

      if (ow.prop.gg.id().equals(game_id))
	return i;
    }
    return -1;
  }

  private int endedGameWindowIndex(String game_id)
  {
    //    Misc.msg(windows.size() + " windows");

    for (int i=0; i < windows.size(); i++) {
      GameWindow ow = windows.get(i);
      if (ow == null)
	continue;

      // FIXME: should return latest matching game!
      if (ow.prop.gg.id().indexOf(game_id) >= 0)
	return i;
    }
    return -1;
  }


  //SwingUtilities.invokeLater(new Runnable() {    
  //    public void run() {
	  
  
  public boolean handle_event(EventMsg msg0)
  {
    final EventMsg msg = msg0;
    
    //    Misc.msg("Client.handle_event() called: sender=" +
    //      msg.get_sender());
    
    if (msg.get_sender().equals(Global.core.GGS_EVENT)) {
      
      EventArgEnum evenum = msg.get_enum();

      if (!evenum.has_more_args())
	return true; // empty message
      
      String s1 = evenum.next_arg();
      
      if (!s1.startsWith(getLogin()+":"))
	return false;
      
      StringTokenizer st = new StringTokenizer(s1, " ");
      String t = st.nextToken();  // skip LOGIN:

      if (!st.hasMoreTokens()) {
	Misc.msg("empty message from service");
	return false;
      }

      t = st.nextToken();

      if (t.equals(JOIN_PREFIX)) {

	// new game, parse game id and game record (ggf)
	//
	// format:  join game_id game-record(ggf)

	GGS_Game gg = new GGS_Game(bg.new_BoardGame());
	gg.set_id(st.nextToken());

	// concatenate all strings

	String all = "";

	while (st.hasMoreTokens()) all += st.nextToken() + " ";
	while (evenum.has_more_args())
	  all += (String)evenum.next_arg() + " ";

	// parse ggf game record

	String e = gg.from_ggf(all);
	if (e != null) { Misc.msg(">>>" + e); return false; }

	int ind = windowIndex(gg.id());

	if (ind >= 0) {

	  // second join => replace game

	  windows.get(ind).new_ggs_game(gg);

	} else {

	  // open new amazons window

	  Misc.msg(">>> NEW WINDOW");
	  addWindow(gg);
	}

	return true;

      } else if (t.equals(LOAD_PREFIX)) {

	// local game, parse game record (ggf)
	//
	// format:  join game_id game-record(ggf)

	GGS_Game gg = new GGS_Game(bg.new_BoardGame());
	gg.set_id(LOCAL_GAME + " " + nextEndId());

	// concatenate all strings

	String all = "";

	while (st.hasMoreTokens()) all += st.nextToken() + " ";
	while (evenum.has_more_args())
	  all += (String)evenum.next_arg() + " ";

	// parse ggf game record

	String e = gg.from_ggf(all);

	if (e != null) { Misc.msg(">>>" + e); return false; }

	int ind = windowIndex(gg.id());

	if (ind >= 0) {

	  // subsequent join => replace game

	  windows.get(ind).new_ggs_game(gg);

	} else {

	  // open new window

	  Misc.msg(">>> NEW WINDOW");
	  addWindow(gg);
	}

	return true;

      } else if (t.equals(MATCH_PREFIX)) {

	// match list
	//
	// format:  match i/n
	// .id ...
	// .id ...

	running_game_strings = new Vector<String>();

	while (evenum.has_more_args()) {
	  String s = ((String)evenum.next_arg()).trim();
	  running_game_strings.add(s);
	  refreshServiceList();
	}

	return true;

      } else if (t.equals(LOOK_PREFIX)) {

	// new game, parse game record (ggf)
	//
	// format:  look game-num(1|2) record1(ggf) [record2]

	if (!evenum.has_more_args()) {
	  Misc.msg(">>> look message corrupt");
	  return true;
	}

	st = new StringTokenizer((String)evenum.next_arg(), " ");

	if (!st.hasMoreTokens()) {
	  Misc.msg(">>> look message corrupt");
	  return false;
	}

	String n = st.nextToken(); // number of games ("1" or "2")

	// concatenate all remaining strings

	String all = "";

	while (st.hasMoreTokens()) all += st.nextToken() + " ";
	while (evenum.has_more_args())
	  all += (String)evenum.next_arg() + " ";

	if (n.equals("1")) {

	  // regular game

	  // parse ggf game record

	  GGS_Game gg = new GGS_Game(bg.new_BoardGame());
	  gg.set_id("replay " + nextEndId());

	  String e = gg.from_ggf(all);
	  if (e != null) { Misc.msg(">>>" + e); return false; }

	  int ind = windowIndex(gg.id());

	  if (ind >= 0) {

	    // second join => replace game

	    windows.get(ind).new_ggs_game(gg);

	  } else {

	    // open new game window

	    Misc.msg(">>> NEW WINDOW");
	    addWindow(gg);
	  }

	  ind = windowIndex(gg.id());
	  windows.get(ind).append("Result= " + gg.get_result());

	} else {

	  // synchro match

	  // split into two GGF strings

	  String all1 = "", all2 = "";

	  int ind = all.indexOf(";)");
	  if (ind < 0) { Misc.msg(">>> look message corrupt");  return false; }

	  all1 = all.substring(0, ind+2);
	  all2 = all.substring(ind+2);

	  // parse ggf records

	  GGS_Game gg1 = new GGS_Game(bg.new_BoardGame());
	  gg1.set_id("replay " + nextEndId() + ".0");
	  String e = gg1.from_ggf(all1);
	  if (e != null) { Misc.msg(">>>" + e); return false; }

	  GGS_Game gg2 = new GGS_Game(bg.new_BoardGame());
	  gg2.set_id("replay " + nextEndId() + ".1");
	  e = gg2.from_ggf(all2);
	  if (e != null) { Misc.msg(">>>" + e); return false; }

	  // create window .0

	  ind = windowIndex(gg1.id());

	  if (ind >= 0) {

	    // subsequent join => replace game

	    windows.get(ind).new_ggs_game(gg1);

	  } else {

	    // open new game window

	    Misc.msg(">>> NEW WINDOW");
	    addWindow(gg1);
	  }

	  ind = windowIndex(gg1.id());
	  windows.get(ind).append("Result= " + gg1.get_result());

	  // create window .1

	  ind = windowIndex(gg2.id());

	  if (ind >= 0) {

	    // subsequent join => replace game

	    windows.get(ind).new_ggs_game(gg2);

	  } else {

	    // open new game window

	    Misc.msg(">>> NEW WINDOW");
	    addWindow(gg2);
	  }

	  ind = windowIndex(gg2.id());
	  windows.get(ind).append("Result= " + gg2.get_result());
	}

	return true;

      } else if (t.equals(UPDATE_PREFIX)) {

	// new move, parse game id and move info
	//
	// format: update game_id last_move_info(move/eval/time)

	int index = windowIndex(st.nextToken());

	if (index < 0) {
	  Misc.msg(">>> game window not found"); return false;
	}

	if (!st.hasMoreTokens()) {
	  Misc.msg(">>> missing move info"); return false;
	}

	t = st.nextToken();
	windows.get(index).handle_event(new EventMsg(GameWindow.FS_MOVE_ID, t));
	return true;

      } else if (t.equals(COMMENT_PREFIX)) {

	// server score suggestion
	//
	// format: comment id score B|W value remarks ...

	int index = windowIndex(st.nextToken());
	if (index < 0) {
	  // was: return Global.core.handle_event(msg);
	  return false;
	}

	evenum = msg.get_enum();
	GameWindow gw = windows.get(index);

	while (evenum.has_more_args()) gw.append(evenum.next_arg());

	String comment_type = st.nextToken();

	if (comment_type.equals(SCORE_COMMENT_PREFIX)) {

	  String login = getOptions().getString("lo");
	  String color = st.nextToken();
	  int value = 0;

	  if (!color.equals("B") && !color.equals("W"))
	    Misc.msg("received illegal color in score message");

	  try { value = Integer.parseInt(st.nextToken()); }
	  catch (NumberFormatException e) {
	    Misc.msg("received illegal value in score message");
	  }

	  boolean for_me = false;

	  if (gw.prop.gg.black_name().equals(login)) {
	    for_me = true;
	    if (color.equals("W")) value = -value;
	  } else if (gw.prop.gg.white_name().equals(login)) {
	    for_me = true;
	    if (color.equals("B")) value = -value;
	  }

	  if (for_me) {
	    gw.set_input(gw.SCORE_INPUT_PREFIX + value);
	  }
	}

	return true;

      } else if (t.equals(TIMEOUT_PREFIX)) {

	// player timeout, parse game id
	//
	// format: timeout id() blabla

	int index = windowIndex(st.nextToken());
	if (index < 0) {
	  // was: return Global.core.handle_event(msg);
	  return false;
	}

	evenum = msg.get_enum();

	while (evenum.has_more_args())
	  windows.get(index).append(evenum.next_arg());

	return true;

      } else if (t.equals(FATAL_TIMEOUT_PREFIX)) {

	// fatal player timeout, parse game id
	//
	// format: fatal-timeout game_id blabla

	int index = windowIndex(st.nextToken());
	if (index < 0) { return Global.core.handle_event(msg); }

	evenum = msg.get_enum();

	while (evenum.has_more_args())
	  windows.get(index).append(evenum.next_arg());

	return true;

      } else if (t.equals(ADD_PREFIX)) {

	t = st.nextToken();

	if (t.startsWith(".")) {

	  // match request
	  // /os:  +  .3 1720.0 foo  clock1 8r20 U 1592.7 mic [clock2] match-id

	  Request req = new Request("???");

	  req.game_id = t;

	  t = st.nextToken();
	  req.rating = t;

	  t = st.nextToken();
	  req.player = t;

	  if (!t.equals(getOptions().getString("lo"))) {

	    req.your_clock.fromString(st.nextToken());
	    req.my_clock = req.your_clock.createCopy();

	    req.game_type = st.nextToken();
	    req.mode = st.nextToken().charAt(0);

	    req.stored_match_id = "";

	    t = st.nextToken();  // my rating
	    t = st.nextToken();  // my name

	    if (st.hasMoreTokens()) {

	      t = st.nextToken();

	      if (!t.startsWith(".")) {

		// 2nd clock

		if (!req.my_clock.fromString(t)) {
		  throw new Error("request: 2nd clock parse error");
		}

		if (st.hasMoreTokens()) {

		  // stored match id
		  req.stored_match_id = st.nextToken();
		}

	      } else {

		// stored match id
		req.stored_match_id = t;

	      }
	    }

	    if (!your_req.fromRequest(req)) {

	      send("t " + getLogin() + " decline " + req.game_id);

	    } else your_req.make_visible();
	  }

	} else if (s1.indexOf(ADD_PREFIX + " match") >= 0) {

	  // match started: add to running_game_strings
	  //
	  // format: + match match_id blabla

	  String gs = s1.substring((getLogin()+": " + ADD_PREFIX + " match").length()+1);
	  running_game_strings.add(gs);

	  refreshServiceList();
	  return true;

	} else {

	  // other ADD commands

	  String name = t;

	  if (!st.hasMoreTokens()) {
	    Misc.msg("illegal message from service");
	    return false;
	  }

	  t = st.nextToken();

	  if (t.equals(WATCH_PREFIX)) {

	    // observer arrived
	    // /os: + name watch .match_id
	    // send arrival message to all game windows

	    String id = st.nextToken();
	    boolean captured = false;
	    int index = windowIndex(id);
	    if (index >= 0) {
	      windows.get(index).arrived(name);
	      captured = true;
	    }

	    index = windowIndex(id + ".0");
	    if (index >= 0) {
	      windows.get(index).arrived(name);
	      captured = true;
	    }

	    index = windowIndex(id + ".1");
	    if (index >= 0) {
	      windows.get(index).arrived(name);
	      captured = true;
	    }

	    if (!captured) { return Global.core.handle_event(msg); }

	    return true;
	  }
	}

      } else if (t.equals(RATING_UPDATE_PREFIX)) {

	// rating update
	// format: rating_update match_id\n update1 \n update2

	String id = st.nextToken();

	// collect entire message

	String msg2 = "rating update:";
	while (evenum.has_more_args()) { msg2 += "\n" + evenum.next_arg(); }

	int index = windowIndex(id);
	if (index >= 0)
	  windows.get(index).append(msg2);

	// synchro games already ended
	index = endedGameWindowIndex(id + ".0");
	if (index >= 0)
	  windows.get(index).append(msg2);
	index = endedGameWindowIndex(id + ".1");
	if (index >= 0)
	  windows.get(index).append(msg2);
	return true;

      } else if (t.equals(GAME_END_PREFIX)) {

	// synchro game has ended
	// format: end match_id

	String id = st.nextToken();
	int index = windowIndex(id);

	// not for us?
	if (index < 0) {
	  return Global.core.handle_event(msg);
	}

	windows.get(index).ended(msg);
	return true;

      } else if (t.equals(REM_PREFIX)) {

	if (s1.indexOf(REM_PREFIX + " match") >= 0) {

	  // match has ended, send messages to all match games
	  // remove line from running_game_strings
	  // add line to finished_game_strings

	  // format: - match match_id blabla

	  String id = st.nextToken();
	  id = st.nextToken();
	  String s = id + " ";

	  for (String q : running_game_strings) {
	    if (q.startsWith(s)) {
	      running_game_strings.remove(q);
	      break;
	    }
	  }

	  String gs = s1.substring((getLogin()+": " + REM_PREFIX + " match").length()+1);

	  finished_game_strings.add(gs);
	  if (finished_game_strings.size() > FINISHED_NUM_MAX)
	    finished_game_strings.remove(0);

	  refreshServiceList();

	  boolean captured = false;
	  int index = windowIndex(id);
	  if (index >= 0) {
	    windows.get(index).ended(msg); captured = true;
	  }
	  index = endedGameWindowIndex(id + ".0");
	  if (index >= 0) {
	    windows.get(index).ended(msg);
	    captured = true;
	  }
	  index = endedGameWindowIndex(id + ".1");
	  if (index >= 0) {
	    windows.get(index).ended(msg);
	    captured = true;
	  }

	  if (captured)
	    return true;

	  return Global.core.handle_event(msg);
	}

	// other "-" commands

	t = st.nextToken();

	String name = t;

	if (!st.hasMoreTokens()) {
	  Misc.msg("corrupt message from service");
	  return false;
	}

	t = st.nextToken();

	if (t.equals(WATCH_PREFIX)) {

	  // observer arrived
	  // /os: - name watch .match_id
	  // send depart message to all game windows

	  String id = st.nextToken();
	  boolean captured = false;
	  int index = windowIndex(id);
	  if (index >= 0) {
	    windows.get(index).departed(name);
	    captured = true;
	  }

	  index = windowIndex(id + ".0");
	  if (index >= 0) {
	    windows.get(index).departed(name);
	    captured = true;
	  }

	  index = windowIndex(id + ".1");
	  if (index >= 0) {
	    windows.get(index).departed(name);
	    captured = true;
	  }

	  if (!captured) {
	    return Global.core.handle_event(msg);
	  }

	  return true;
	}

      } else if (t.startsWith(".")) {

	// kibitz or whisper
	//  /os: .1 1476 mic: bla

	String match_id = t;

	// send message to window
	// delete "/os: game_id" in first string

	evenum = msg.get_enum();
	String s = evenum.next_arg();
	int k = s.indexOf(' ');
	k = s.indexOf(' ', k+1);  // k is index of second space
	t = s.substring(k+1);

	String msg2 = df.format(new Date()) + " " + t;

	while (evenum.has_more_args()) { msg2 += "\n" + evenum.next_arg(); }

	int index = windowIndex(match_id);

	if (index >= 0) {
	  windows.get(index).append(msg2);
	  return true;
	}

	index = windowIndex(match_id + ".0");

	if (index >= 0) {
	  windows.get(index).append(msg2);
	}

	index = windowIndex(match_id + ".1");

	if (index >= 0) {
	  windows.get(index).append(msg2);
	}

	return true;

      } else if (t.equals("error") || t.equals("score") ||
		 t.equals("abort") || t.equals("undo")) {

	// error message
	//  /os: error .1 illegal move

	String first = t;
	t = st.nextToken();
	int index = windowIndex(t);

	if (index < 0) {
	  Misc.msg("error: game not found:" + t);
	  return false;
	}

	// send message to window
	// delete "/os: error game_id" in first string

	evenum = msg.get_enum();
	String s = evenum.next_arg();

	int k = s.indexOf(' ');
	k = s.indexOf(' ', k+1);  // k is index of second space
	k = s.indexOf(' ', k+1);  // k is index of third space
	t = s.substring(k+1);

	t = first + ": " + t;

	windows.get(index).append(t);

	while (evenum.has_more_args()) {
	  windows.get(index).append(evenum.next_arg());
	}

	return true;
      }

    } else if (msg.get_sender().equals(GameWindow.QUIT_ID)) {

      // delete window from list and dispose of it

      removeWindow(msg.get_enum().next_arg());
      return true;
    }

    // can't handle it here, let core take care of it
    return false;
  }
			       
  
  public void refreshServiceList()
  {
    game_list = new Vector<String>();
    
    for (int i=running_game_strings.size()-1; i >= 0; --i)
      game_list.add("+ " + running_game_strings.get(i));
    
    for (int i=0; i < finished_game_strings.size(); ++i)
      game_list.add("- " + finished_game_strings.get(i));
    
    JList list = tab_panel.get_list();
    list.removeAll();
    
    DefaultListModel model = new DefaultListModel();
    for (int i=0; i < game_list.size(); ++i)
      model.addElement(game_list.get(i));
    
    list.setModel(model);
    tab_panel.repaint();
  }

  
  public void shutDown()
  {
    if (windows != null) {
      for (int i=0; i < windows.size(); i++) {
	GameWindow ow = windows.get(i);
	if (ow != null) { ow.shut_down(); windows.setElementAt(null, i); }
      }
    }

    if (my_req   != null) { my_req.dispose(); my_req = null; }
    if (your_req != null) { your_req.dispose(); your_req = null; }
  }


  public int nextEndId() { return ++end_id; }

  public void clickedOnUser(String who)
  {
    my_req.make_visible(who);
  }

  public Vector<String> get_list()
  {
    Vector<String> l = new Vector<String>();

    for (String s : running_game_strings)
      l.add("+ " + s);
    
    for (String t : finished_game_strings)
      l.add("- " + t);

    return l;
  }

  private void selected(int index) {

    if (index >= running_game_strings.size())
      return;

    // parse match id

    StringTokenizer st =
      new StringTokenizer(running_game_strings.get(index), " ");

    send("t " + getLogin() + " watch + " + st.nextToken());
  }

  private void setContext() { Global.core.set_current_talk_context(getLogin()); }

  public String getLogin() { return "/" + getBaseName(); }
  private String getBaseName() { return bg.get_base_name(); }
  public String getName() { return bg.get_name(); }

  // ServiceClient interface
  public void send(String what)
  {
    Global.core.send_to_server(getLogin(), what);
  }

  public void sendWithFocus(String what)
  {
    Global.core.send_to_server_with_focus(getLogin(), what);
  }

}

