/* $Id: GGS_Game.java,v 1.3 2006/08/24 09:12:45 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.boardgamesvc;

import java.util.*;
import ggs.ggsa.main.*;

public class GGS_Game
{
  private String     game_id = "end"; // => game ended
  private PlayerInfo b_inf = new PlayerInfo();
  private PlayerInfo w_inf = new PlayerInfo();

  private Vector<MoveInfo> infos;
  private int index;  // of current position

  private String type;
  private BoardGame bg;
  private String start_pos_ggf;
  private int start_color_to_move;

  private String  date;
  private String  result;
  private boolean komi_valid;
  private float   komi;

  // game type

  private int     board_type = -1;
  private int     rand_type;     // >= 0 <=> is rand game
  private boolean synchro_game;
  private boolean komi_game;
  private boolean anti_game;
  private int     pref_color;    // color request of issuing player
  
  public GGS_Game(BoardGame bg)
  {
    this.bg = bg;
    
    komi_valid = false;
    komi = 0;
    b_inf.komi_move = null;
    w_inf.komi_move = null;
    //setup(new_Board(), new GameClock(), new GameClock());
  }

  // setup starting position and initialize clocks and moveinfo vector
  
  public void setup(GameClock b_clock, GameClock w_clock)
  {
    infos = new Vector<MoveInfo>();
    MoveInfo info = new MoveInfo();
    //info.bo = bo.createCopy();
    info.b_clock = b_clock.createCopy();
    info.w_clock = w_clock.createCopy();
    info.b_clock.reset();
    info.w_clock.reset();
    infos.addElement(info);
    index = 0;
  }

  public String komi_type_string() { return type; }

  public boolean advance(int delta) {
    bg.advance(delta);
    index += delta;
    boolean ok = index >= 0 && index < infos.size();
    if (index < 0) index = 0;
    if (index >= infos.size()) index = infos.size()-1;
    return ok;
  }

  public boolean at_end() { return index >= infos.size()-1; }

  public void set_end() { infos.setSize(index+1); }

  public int move_num() { return infos.size()-1; }

  public int get_index() { return index; }

  public MoveInfo current_info() {
    //Global.dbgmsg("GGS_index= " + index + " bg=" + bg.get_index());
    return (MoveInfo)infos.elementAt(index); 
  }

  public MoveInfo prev_info() {
    if (index > 0) 
      return (MoveInfo)infos.elementAt(index-1);
    else
      return null;
  }

  public MoveInfo prev_prev_info() {
    if (index > 1) 
      return (MoveInfo)infos.elementAt(index-2);
    else
      return null;
  }

  public String get_result() { return result; }
  
  public String make_move(String move_eval_time)
  {
    MoveInfo curr_info = current_info();

    GGS_Move gm = new GGS_Move();
    String err = gm.fromString(move_eval_time);
    if (err != null) return err;
      
    int was_to_move = bg.to_move();

    err = bg.make_move(gm.move);
    if (err != null) return err;

    curr_info.move = move_eval_time;

    MoveInfo info = new MoveInfo();
    info.b_clock = curr_info.b_clock.createCopy();
    info.w_clock = curr_info.w_clock.createCopy();

    if (was_to_move == PlayerColor.BLACK) {
      info.b_clock.stop((long)(gm.get_time() * USEC.uSec));
    } else {
      info.w_clock.stop((long)(gm.get_time() * USEC.uSec));
    }

    info.to_move = bg.to_move();
    
    ++index;

    if (index >= infos.size()) {
      infos.addElement(info);
    } else {
      infos.setElementAt(info, index);
      infos.setSize(index+1);
    }

    return null;
  }


  // ggf functions

  private static boolean skip_until(StringStream is, char u)
  {
    char c;
    do { c = is.get(); } while (is.ok() && c != u);
    return is.ok();
  }


  private static String read_until(StringStream is, char u)
  {
    char c;

    String s = "";
  
    do { 
      c = is.get();
      s += c; 
    } while (is.ok() && c != u);

    if (is.ok()) s = s.substring(0, s.length()-1);

    if (is.ok()) return s; else return null;
  }


  private static String next_token(StringStream is)
  {
    char c;
    String s = null;
    
    // skip white space and ';'

    do { c = is.get(); } while (is.ok() && (c == ' ' || c == '\n' || c == ';'));

    if (!is.ok()) return s;
  
    if (c == ')') {  // end of game
      return "" + c;
    }

    is.rewind(1);                         
    s = "";
    
    do { 

      c = is.get();

      if (Global.DBG_GGF) Global.dbgmsg("<" + c + ">");
      s += "" + Character.toUpperCase(c);
      if (s.length() > 3) return null;
    
    } while (is.ok() && c != '[');

    if (is.ok()) s = s.substring(0, s.length()-1);
    return s;
  }


  public String from_ggf(String str)
  {
    StringStream is = new StringStream(str);
    char c;
    GameClock b_clock = null;
    GameClock w_clock = null;

    type = null;
    infos = null;
    
    // search "(;"

    do {
      if (!skip_until(is, '(')) return "no ( found";
      c = is.get();
      if (c != ';') is.rewind(1);
    } while (c != ';');

    for (;;) {

      String s;
      float f;

      if (!is.ok()) return "no more tokens";
      s = next_token(is);
      if (s == null) return "no more tokens";

      if (Global.DBG_GGF) Global.dbgmsg("token: " + s);

      if      (s.equals(")")) break; // end of game
      else if (s.equals("GM") ||     // ZAP
	       s.equals("US") || 
	       s.equals("CP") || 
	       s.equals("COPYRIGHT") ||
	       s.equals("GN") || 
	       s.equals("C")  || 
	       s.equals("BL") || 
	       s.equals("WL") || 
	       s.equals("NB") || 
	       s.equals("NW") || 
	       s.equals("EV") || 
	       s.equals("LT")) { skip_until(is, ']'); } 

      else if (s.equals("PB")) {           // black name

	b_inf.name = read_until(is, ']');
	if (b_inf.name == null) return "] not found after PB";

      } else if (s.equals("PW")) {         // white name

	w_inf.name = read_until(is, ']');
	if (w_inf.name == null) return "] not found after PW";

      } else if (s.equals("RB")) {         // black rating

	if ((s=read_until(is, ']')) == null) return "missing ]";

	try { b_inf.rating = Float.valueOf(s).floatValue(); }
	catch (NumberFormatException e) { return "GGF: corrupt RB"; }

      } else if (s.equals("RW")) {         // white rating

	if ((s=read_until(is, ']')) == null) return "missing ]";

	try { w_inf.rating = Float.valueOf(s).floatValue(); }
	catch (NumberFormatException e) { return "GGF: corrupt RW"; }
      
      } else if (s.equals("PC")) {         // place

	if (read_until(is, ']') == null) {
	  return "missing ]";
	}

	// ... place = s;
      
      } else if (s.equals("DT")) {         // date

	if ((date=read_until(is, ']')) == null) {
	  return "missing ]";
	}

      } else if (s.equals("TY")) {         // type

	if ((s=read_until(is, ']')) == null) return "missing ]";

	type = s;
	type_from_string(s);
	bg.set_anti(anti_game);
	
      } else if (s.equals("TI")) {         // both times

	if ((s=read_until(is, ']')) == null) return "missing ]";
	b_clock = new GameClock();
	if (!b_clock.fromString(s)) return "ggf: time corrupt";
	w_clock = b_clock;

      } else if (s.equals("TB")) {         // b time

	if ((s=read_until(is, ']')) == null) return "missing ]";
	b_clock = new GameClock();
	if (!b_clock.fromString(s)) return "ggf: time corrupt";

      } else if (s.equals("TW")) {         // w time

	if ((s=read_until(is, ']')) == null) return "missing ]";
	w_clock = new GameClock();
	if (!w_clock.fromString(s)) return "ggf: time corrupt";

      } else if (s.equals("KM")) {         // komi

	if ((s=read_until(is, ']')) == null) return "missing ]";

	try { komi = Float.valueOf(s).floatValue(); }
	catch (NumberFormatException e) { return "syntax error: KM"; }
      
	komi_valid = true;
	
      } else if (s.equals("KB")) {         // black komi move

	if ((s=read_until(is, ']')) == null) return "missing ]";
	b_inf.komi_move = s;

      } else if (s.equals("KW")) {         // white komi move

	if ((s=read_until(is, ']')) == null) return "missing ]";
	w_inf.komi_move = s;

      } else if (s.equals("RE")) {         // result

	if ((s=read_until(is, ']')) == null) {
	  return "missing ]";
	}
	result = s;
      
      } else if (s.equals("BO")) {         // board

	if ((s=read_until(is, ']')) == null) return "missing ]";

	if (infos != null) return "only one BO[] allowed";

	// Global.dbgmsg("BO: <" + s + ">");

	String err = bg.start_pos(s);
	if (err != null) return err;

	start_pos_ggf = s.replace(' ', '\n');
	start_color_to_move = bg.to_move();
	
	if (b_clock == null || w_clock == null || type == null) {
	  return "missing times/game-type before BO[]";
	}

	setup(b_clock, w_clock);
	((MoveInfo)infos.elementAt(0)).to_move = bg.to_move();
  
      } else if (s.equals("B") || s.equals("W")) {         // move

	if ((s=read_until(is, ']')) == null) return "missing ]";

	if (infos == null) return "missing BO[]";

	String err = make_move(s);
	if (err != null) return err;

      } else {

	return "unknown GGF command: " + s;
      
      }
    }

    if (infos == null) return "missing BO[]";

    if (komi_valid) {

      // subtract time for not played komi move
      
      GGS_Move gm = new GGS_Move();

      if (start_color_to_move != PlayerColor.BLACK) {
	String err = gm.fromString(b_inf.komi_move);
	if (err != null) return err;
	current_info().b_clock.stop((long)(gm.get_time() * USEC.uSec));
      }
	
      if (start_color_to_move != PlayerColor.WHITE) {
	String err = gm.fromString(w_inf.komi_move);
	if (err != null) return err;
	current_info().w_clock.stop((long)(gm.get_time() * USEC.uSec));
      }
    }

    return null;
  }

  public String to_ggf(boolean one_line)
  {
    int indb = 0;
    int indw;
    String s = "";
    String my_eol = "";

    if (!one_line) my_eol += "\n";
    
    s += "(;GM[" + bg.get_name() + "]PC[GGS/" + bg.get_base_name() + "]DT["
      + date + "]" + my_eol;
    s += "PB[" + b_inf.name + "]";
    s += "PW[" + w_inf.name + "]" + my_eol;
    s += "RB[" + b_inf.rating + "]";
    s += "RW[" + w_inf.rating + "]" + my_eol;

    String s1 = ((MoveInfo)infos.elementAt(0)).b_clock.toString(true);
    String s2 = ((MoveInfo)infos.elementAt(0)).w_clock.toString(true);

    if (!s1.equals(s2)) {
      s += "TB[" + s1 + "]";
      s += "TW[" + s2 + "]" + my_eol;
    } else {
      s += "TI[" + s1 + "]" + my_eol;
    }
  
    s += "TY[" + type.toString() + "]";

    if (b_inf.komi_move != null) {
      s += "KB[" + b_inf.komi_move.toString() + "]";
    }
    if (w_inf.komi_move != null) {
      s += "KW[" + w_inf.komi_move.toString() + "]";
    }
    if (komi_valid) {
      s += "KM[" + komi + "]";
    }

    s += "RE[" + result + "]" + my_eol;

    // starting position

    s += "BO[" + start_pos_ggf + "]" + my_eol;
  
    int n = 0;

    for (int i=0; i < infos.size()-1; ++i) {
      if (((MoveInfo)infos.elementAt(i)).to_move == PlayerColor.BLACK)
	s += "B";
      else
	s += "W";
      s += "[" + ((MoveInfo)infos.elementAt(i)).move + "]";
      ++n;
      if (n >= 2) { n = 0; if (i < infos.size()-2) s += my_eol; }
    }
    if (n != 0) s += my_eol;
    
    s += ";)";

    return s;
  }

  public boolean is_komi_game() { return komi_game; }
  public boolean is_anti_game() { return anti_game; }
  public boolean is_rand_game() { return rand_type >= 0; }
  public boolean is_synchro_game() { return synchro_game; }  
  public int     get_pref_color() { return pref_color; }
  public int     get_rand_type() { return rand_type; }

  public int     get_board_type() {
    if (board_type >= 0) return board_type;
    return default_board_type();
  }
  
  public boolean has_komi() { return bg.has_komi(); }
  public boolean has_anti() { return bg.has_anti(); }
  public boolean has_rand() { return bg.has_rand(); }
  public boolean has_rand_type() { return bg.has_rand_type(); }  
  public boolean has_pref_color() { return bg.has_pref_color(); }  
  public boolean has_synchro() { return bg.has_synchro(); }
  public boolean has_pass() { return bg.has_pass(); }  

  public int     board_type_num() { return bg.board_type_num(); }
  public int     default_board_type() { return bg.default_board_type(); }
  public int     board_type_index(int type) { return bg.board_type_index(type); }
  public String  board_type_string(int index) { return bg.board_type_string(index); }
  public int     board_type(int index) { return bg.board_type(index); }

  public String  info_string() { return bg.info_string(); }
  public void    input(String s) { bg.input(s); }
  public boolean turned_if_black() { return bg.turned_if_black(); }
  public boolean never_turn() { return bg.never_turn(); }
  public String  black_name() { return b_inf.name; }
  public String  white_name() { return w_inf.name; }  
  public float   black_rating() { return b_inf.rating; }
  public float   white_rating() { return w_inf.rating; }  
  public String  id() { return game_id; }
  public void    set_id(String s) { game_id = s; }  
  
  public int to_move() { return bg.to_move(); }
  public void mouse_action(String t, int gfx_x, int gfx_y) {
    bg.mouse_action(t, gfx_x, gfx_y);
  }
  public boolean komi_valid() { return this.komi_valid; }
  public float   get_komi() { return komi; }
  public BoardGame get_board_game() { return bg; }  

  public String type_to_string() { return type_to_string(false, 0); }
  
  public String type_to_string(boolean with_komi, float komi)
  {
    String s = "";
    if (bg.has_synchro() && is_synchro_game()) s = "s";    
    s += board_type;

    if (bg.has_komi() && is_komi_game()) {
      s += "k";
      if (with_komi) {
        Format f = new Format("%+.2f");
        s += f.form(komi);
      }
    }
    
    if (bg.has_rand() && is_rand_game()) s += "r";
    if (bg.has_rand_type() && is_rand_game()) s += rand_type;
    if (bg.has_pref_color() && pref_color == PlayerColor.BLACK) s += "b";
    if (bg.has_pref_color() && pref_color == PlayerColor.WHITE) s += "w";  
    if (bg.has_anti() && is_anti_game()) s += "a";
    return s;
  }


  public String type_from_string(String s)
  {
    StringStream is = new StringStream(s);

    //    Global.dbgmsg("fromString: " + s);
    
    rand_type = -1;
    anti_game = false;
    synchro_game = false;
    komi_game = false;
    pref_color = PlayerColor.UNDEF;

    char c;

    if (bg.has_synchro()) {

      c = is.get();
      if (!is.ok()) return "illegal game type";

      if (Character.toUpperCase(c) == 'S') synchro_game = true;
      else is.rewind(1);
    }

    board_type = is.get_int();
    if (!is.ok()) return "illegal board type";

    //    Global.dbgmsg("ok=" + is.ok() + " eos=" + is.eos());

    while (is.ok() && !is.eos()) {

      c = is.get();
      //Global.dbgmsg("read: " + c);
      
      switch (Character.toUpperCase(c)) {

      case 'K':

        if (!bg.has_komi()) return "no komi games";
        if (pref_color != PlayerColor.UNDEF) {
          return "no preferred color in komi games";
        }
        komi_game = true;
        break;

      case 'R':

        if (!bg.has_rand()) return "no random games";
        if (pref_color != PlayerColor.UNDEF) {
          return "no preferred color in rand games";
        }

        if (bg.has_rand_type()) {
          
          rand_type = is.get_int();

          if (!is.ok()) return "rand-type missing";
          if (rand_type < 0) return "rand-type < 0";

        } else
          
          rand_type = 1;

        break;

      case 'A':

        if (!bg.has_anti()) return "no anti games";
        anti_game = true;
        break;

      case 'B':
      case 'W':

        if (!bg.has_pref_color()) return "no prefered colors";
        
        if (is_komi_game() || is_rand_game()) {
          return "no preferred color in komi or rand games";
        }
        if (pref_color != PlayerColor.UNDEF) {
          return "only one color preference allowed";
        }

        if (Character.toUpperCase(c) == 'B')
          pref_color = PlayerColor.BLACK;
        else
          pref_color = PlayerColor.WHITE;
        break;
      
      default:
        break;
      }
    }

    if (!is.eos()) { return "illegal game type"; }

    return null;
  }


  public boolean komi_move() {
    return is_komi_game() && move_num() == 0;
  }

  public boolean involved(String user) {
    return black_name().equals(user) || white_name().equals(user);
  }
  
  public boolean to_be_turned(String user)
  {
    if (never_turn()) return false;
    
    int bottom_color;
    boolean inv = involved(user);
    boolean km  = komi_move();

    if (km) {

      bottom_color = to_move();

    } else if (inv) {

      bottom_color = black_name().equals(user) ? PlayerColor.BLACK : PlayerColor.WHITE;

    } else return false;  // !km && !inv => standard orientation

    return (bottom_color == PlayerColor.BLACK) == turned_if_black();
  }
}

