/* $Id: OthelloLikeBoardGame.java,v 1.2 2006/08/22 05:53:38 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/


package ggs.ggsa.boardgamesvc;

import java.awt.*;
import ggs.ggsa.main.*;

public abstract class OthelloLikeBoardGame extends MailBoxBoardGame
{
  public static final int GFX_BLACK_DISC = 4;
  public static final int GFX_WHITE_DISC = 8;
  public static final int GFX_BLACK_MOVE = 16;
  public static final int GFX_WHITE_MOVE = 32;
  public static final int GFX_DISC_MASK  = GFX_BLACK_DISC + GFX_WHITE_DISC;

  public int border_width() { return 1; }

  public byte char2cont(char c)
  {
    switch (c) {
    case '*': return BLACK;
    case 'O': return WHITE;
    case '-': return EMPTY;
    default:
      throw new Error("illegal char in OthelloLikeBoardGame.char2cont");
    }
  }

  public char cont2char(byte co)
  {
    switch (co) {
    case BLACK: return '*';
    case WHITE: return 'O';
    case EMPTY: return '-';
    default:
      throw new Error("illegal cont in OthelloLikeBoardGame.cont2char");
    }
  }

  protected void paint_gfx_cont(Graphics g, int x, int y, int parity, int gfx_cont) {

    // Global.dbgmsg(x + " " + y + " " + parity);
    if ((gfx_cont & GFX_BORDER) != 0) return;

    if (anti) {
      
      if (parity == 0) g.setColor(col_aempty0); else g.setColor(col_aempty1);

    } else {

      if (parity == 0) g.setColor(col_empty0); else g.setColor(col_empty1);

    }
    
    g.fillRect(x+1,y+1,gfx_sq_w-1, gfx_sq_h-1);
    
    if ((gfx_cont & GFX_DISC_MASK) != 0) {

      // disc

      Color d_col;
      Color b_col;

      if ((gfx_cont & GFX_BLACK_DISC) != 0) {

        d_col = col_black;
        b_col = col_white;

      } else {

        b_col = col_black;
        d_col = col_white;

      }

      int disc_diam = (int)Math.round(gfx_sq_w * 0.85);
      disc_diam &= ~1;
      
      int xl = x + (gfx_sq_w - disc_diam)/2;
      int yt = y + (gfx_sq_h - disc_diam)/2;

      g.setColor(d_col);
      Bresenham.fillCircle(g, xl, yt, disc_diam);

      // System.out.println(xl + " " + yt + " " + disc_diam + " ");
      
      g.setColor(b_col);
      Bresenham.drawCircle(g, xl, yt, disc_diam);

      if ((gfx_cont & GFX_LAST_MOVE) != 0) {

        // last move: draw inner square

        int dot_diam = disc_diam/4;
	dot_diam &= ~1;
	
        xl = x + (gfx_sq_w - dot_diam)/2;
        yt = y + (gfx_sq_h - dot_diam)/2;
        g.drawRect(xl, yt, dot_diam, dot_diam);
      }
      
    } else {

      if ((gfx_cont & GFX_BLACK_MOVE) != 0 ||
          (gfx_cont & GFX_WHITE_MOVE) != 0) {

        // legal move: draw X

        int cx = x + gfx_sq_w/2;
        int cy = y + gfx_sq_h/2;
        int l = gfx_sq_w/12;

        if ((gfx_cont & GFX_BLACK_MOVE) != 0) 
          g.setColor(col_black);
        else 
          g.setColor(col_white);
        
        g.drawLine(cx-l,   cy-l, cx+l,   cy+l);
        g.drawLine(cx-l-1, cy-l, cx+l-1, cy+l);
        g.drawLine(cx-l,   cy+l, cx+l,   cy-l);
        g.drawLine(cx-l-1, cy+l, cx+l-1, cy-l);
      }
    }
  }

  public boolean pieces_on_grid() { return false; }
  protected boolean only_squares() { return true; }

  protected String read_extra_info(MailBoxGameState st, StringStream is) {
    return null;
  }
  
  protected String extra_info_to_string(MailBoxGameState st) {
    return "";
  }
  
  public boolean paint_all(Graphics g, boolean repaint,
			   int gfx_sq[], int prev_gfx_sq[],
			   boolean xflip, boolean yflip) {
    return false;
  }  

}
