/* $Id: Core.java,v 1.26 2006/08/31 06:26:18 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.main;

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.lang.*;
import java.text.*;
import ggs.shared.*;

public class Core implements EventHandler,Sender
{
  public final static boolean REPORT_EVENTS = false;

  public final static String GGS_EVENT  = "GGS";
  public final static String GUI_EVENT  = "GUI";

  public final static String GUI_QUIT   = "quit";
  public final static String GUI_CLOSE  = "close";
  public final static String GUI_RESIZE = "resize";
  public final static String GUI_INPUT  = "input ";  // + line

  public final static int GUI_MIN_Y      = 0;

  public final static String TO_SERVER = "toserver ";  // + line

  public final static String CENTRAL_TALK_CONTEXT = " GGS";

  private String   host;
  private int      port;
  private String   login;
  private String   passwd;
  private Geometry geom;

  private int      server_port;
  private boolean  server_mode;

  public Font  font;
  public Font  list_font;
  public Font  clock_font;
  public Font  button_font;

  private DateFormat df = new SimpleDateFormat("HH:mm:ss");

  private String newline = System.getProperty("line.separator");

  private Integer handle_lock = new Integer(0);

  private JMenu last_menu = null;
  private ServiceClient default_service = null;
  private String first_service_login = "";

  // services
  private Vector<ServiceClient> services = new Vector<ServiceClient>();
  
  private Vector<TalkContext> talk_contexts = new Vector<TalkContext>();
  TalkContext current_talk_context;

  void UI_init() 
  {
    int medium = Global.options.getInteger("fs").intValue();
    int large = (medium*6)/5;
    
    Font font_mono_medium = new Font("Lucida Sans Typewriter",
				     Font.PLAIN + Font.BOLD,
				     medium
				     );

    Font font_prop_medium = new Font("Lucida",
				     Font.PLAIN + Font.BOLD,
				     medium
				     );

    Font font_mono_large = new Font("Lucida Sans Typewriter",
				    Font.PLAIN + Font.BOLD,
				    large
				    );

    Font font_prop_large = new Font("Lucida",
				    Font.PLAIN + Font.BOLD,
				    large
				    );

    UIManager.put("Button.font",              font_prop_medium);
    UIManager.put("CheckBox.font",            font_prop_medium);
    UIManager.put("CheckBoxMenuItem.font",    font_prop_medium);
    UIManager.put("ColorChooser.font",        font_prop_medium);
    UIManager.put("ComboBox.font",            font_prop_medium);
    UIManager.put("EditorPane.font",          font_prop_medium);
    UIManager.put("Label.font",               font_prop_medium);
    UIManager.put("List.font",                font_prop_medium);
    UIManager.put("Menu.font",                font_prop_medium);
    UIManager.put("MenuBar.font",             font_prop_medium);
    UIManager.put("MenuItem.font",            font_prop_medium);
    UIManager.put("OptionPane.font",          font_prop_medium);
    UIManager.put("Panel.font",               font_prop_medium);
    UIManager.put("PasswordField.font",       font_prop_medium);
    UIManager.put("PopupMenu.font",           font_prop_medium);
    UIManager.put("ProgressBar.font",         font_prop_medium);
    UIManager.put("RadioButton.font",         font_prop_medium);
    UIManager.put("RadioButtonMenuItem.font", font_prop_medium);
    UIManager.put("ScrollPane.font",          font_prop_medium);
    UIManager.put("TabbedPane.font",          font_prop_medium);
    UIManager.put("Table.font",               font_prop_medium);
    UIManager.put("TableHeader.font",         font_prop_medium);
    UIManager.put("TextArea.font",            font_mono_medium);
    UIManager.put("TextField.font",           font_mono_medium);
    UIManager.put("TextPane.font",            font_prop_medium);
    UIManager.put("TitledBorder.font",        font_prop_medium);
    UIManager.put("ToggleButton.font",        font_prop_medium);
    UIManager.put("ToolBar.font",             font_prop_medium);
    UIManager.put("ToolTip.font",             font_prop_medium);
    UIManager.put("Tree.font",                font_prop_medium);
    UIManager.put("Viewport.font",            font_prop_medium);

    EmacsKeyBindings.load();    
  }

  public Core()
  {
  }
  
  public void run(String[] args)
  {
    Global.core = this; // prevents null-pointer exceptions

    // core default options:

    Global.options.put("ho", "opal.cs.ualberta.ca"); // host
    Global.options.put("po", new Integer(5000));     // port
    Global.options.put("lo", "");                    // no login -> keyboard input
    Global.options.put("pa", "");                    // no password -> keyboard input
    Global.options.put("cm", "");                    // initial commands
    Global.options.put("ge",  new Geometry(800,800,0,0)); // main window geometry
    Global.options.put("sv", "Othello:Skat:Amazons:Checkers:Chess:Go:Hex:DotsAndBoxes:Domineering:Ataxx");
    // service list

    Global.options.put("bf", "/data/audio/drip.au");  // beep file
    
    Global.options.put("ulw", new Integer(90));  // user list width 
    Global.options.put("tph", new Integer(400)); // talk panel height

    Global.options.put("s",   new Integer(-1));      // server-mode socket

    // for game widgets and previous main window, these will go away
    Global.options.put("fs",  new Integer(12)); // font size
    Global.options.put("lfs", new Integer(12)); // list font size
    Global.options.put("cfs", new Integer(14)); // clock font size
    Global.options.put("bfs", new Integer(14)); // button font size    

    // new
    // Global.options.put("mhs", new Integer(105));// main horizontal split (list width)
    // Global.options.put("ths", new Integer(70)); // talk horizontal split (list widt)

//     // not honoured yet
//     Global.options.put("bg",  Integer.decode("#bfbfbf"));  // window background
//     Global.options.put("fg",  Integer.decode("#0"     ));  // window foreground
//     Global.options.put("ibg", Integer.decode("#ffffff"));  // input background
//     Global.options.put("ifg", Integer.decode("#0"     ));  // input foreground
//     Global.options.put("tbg", Integer.decode("#bfbfbf"));  // text background
//     Global.options.put("tfg", Integer.decode("#0"     ));  // text foreground

    Global.options.put("dbg",  new Integer(0));      // debug output

    UI_init();
    
    for (String s : args)
      Global.options.put_line(s, true);

    // load service classes

    StringTokenizer st = new StringTokenizer(Global.options.getString("sv"), ":", false);
    while (st.hasMoreTokens()) {

      Class   cl  = null;
      ServiceClient svc = null;
      String  cn  = st.nextToken();

      cn = "ggs.ggsa._" + cn.toLowerCase() + "." + cn;

      Global.dbgmsg("loading class: " + cn + "...");

      try { cl = Class.forName(cn); }
      catch (ClassNotFoundException e) {
        Global.errmsg("class not found");
      }

      try { svc = (ServiceClient)cl.newInstance(); }
      catch (IllegalAccessException e) {
        Global.errmsg("illegal access");
      }
      catch (InstantiationException e) {
        Global.errmsg("class bad instantiation");
      }

      Global.dbgmsg("OK");

      if (services.size() == 0) {
	first_service_login = svc.getLogin();
      }

      svc.setDefaultOptions();
      services.addElement(svc);
    }


    //  Global.options.write();
    
    // try setting service variables (with warnings)

    for (String s : args)
      Global.options.put_line(s, true);

    host   = Global.options.getString("ho");
    port   = Global.options.getInteger("po").intValue();
    login  = Global.options.getString("lo");
    passwd = Global.options.getString("pa");
    geom   = Global.options.getGeometry("ge");

    Global.main_window = new MainWindow(geom);    

    // for game widgets
    
    font = new Font("monospaced",
                    Font.PLAIN,
                    Global.options.getInteger("fs").intValue());

    list_font = new Font("monospaced",
                         Font.PLAIN,
                         Global.options.getInteger("lfs").intValue());

    clock_font = new Font("monospaced",
                          Font.PLAIN,
                          Global.options.getInteger("cfs").intValue());

    button_font = new Font("monospaced",
                           Font.PLAIN+Font.BOLD,
                           Global.options.getInteger("bfs").intValue());

    server_port = Global.options.getInteger("s").intValue();
    server_mode = (server_port >= 0);

    if (server_port < 0) {
      Global.main_window.setTitle("GGSA " + Global.VERSION + " " + login + "@" +
                                  host + ":" + port);
    } else {
      Global.main_window.setTitle("GGSA " + Global.VERSION +
                                  " server@" + server_port);
    }

    Global.dbgmsg("xy="+geom.get_x()+" "+geom.get_y());

    init_talk_contexts();

    Global.sc = new SocketComm();

    try { Global.sc.connect(Global.main_window, host, port, login, passwd); }
    catch(UnknownHostException f) {
      Global.main_window.append("connect problems: unknown host " + host);
      Global.main_window.append(" shutting down in 5 seconds ...");
      Toolkit.getDefaultToolkit().beep();
      Global.sleep(5000);
      // Global.ggsa.destroy();
    }
    catch(IOException e) {
      Global.main_window.append("connect problems: io error " + host + ":" + port);
      Global.main_window.append(" shutting down in 5 seconds ...");
      Toolkit.getDefaultToolkit().beep();
      Global.sleep(5000);
      // Global.ggsa.destroy();
    }

    login = Global.options.getString("lo");

    // login established

    Global.main_window.setTitle("GGSA " + Global.VERSION + " " +
                                login + "@" + host + ":" + port);
    
    Global.ggs = new GGSEventThread(GGS_EVENT, Global.sc);
    Global.ggs.ef.set_default_handler(this);

    for (ServiceClient s : services) {
      Misc.msg("init " + s.getName());
      Global.ggs.ef.add_handler(s);
      s.init(this, Global.options);
      Global.main_window.init_service(s);
    }

    Global.ggs.start();

    send_to_server("vt100 -");
    send_to_server("notify + *");
    send_to_server("notify +");

    String cmds = Global.options.getString("cm");

    if (cmds.length() > 0) {
      st = new StringTokenizer(cmds,";");
      while (st.hasMoreTokens())
	send_to_server(st.nextToken());
    }

  }

  public void send_to_server(String s)
  {
    send_to_server(current_talk_context.name, s);
  }

  public void send_to_server(String context_name, String s)
  {
    send(s);
    output_text(context_name, s);
  }

  public void send_to_server_with_focus(String context_name, String s)
  {
    try { Global.sc.write_line(s); }
    catch (IOException e) {
      Global.errmsg("socket write failed 2");
    }
    output_text(context_name, s);
    set_current_talk_context(context_name);
  }

  private boolean handle_central_msg(EventMsg msg)
  {
    EventArgEnum en = msg.get_enum();

    if (!en.has_more_args())
      return false;

    String first = en.next_arg();

    if (first.length() < 4 || first.charAt(0) != ':') return false;
    if (first.charAt(2) != '+' && first.charAt(2) != '-') return false;
    if (first.charAt(3) != ' ') return false;

    // arrival/departure messages

    Vector<String> users  = new Vector<String>();
    Vector<String> addrem = new Vector<String>();

    en = msg.get_enum();

    while (en.has_more_args()) {

      String line = en.next_arg();
      if (line.length() >= 3 && first.charAt(0) == ':') {

        StringTokenizer st = new StringTokenizer(line, " ", false);
        st.nextToken(); // ':'

        if (st.hasMoreTokens()) {
          String s = st.nextToken(); // '+' '-'
          String u = st.nextToken();

          if (st.hasMoreTokens()) st.nextToken(); // level -- Igor

          if (!u.startsWith(".") && !u.startsWith("_") && !st.hasMoreTokens()) {
            users.add(u);
            addrem.add(s);

            // connect/disconnect message to talk_contexts if they exist

            if (talk_context_index(u) >= 0) {
              if (s.equals("+"))
                output_text(u, ">>> RECONNECTED", false);
              else
                output_text(u, ">>> DISCONNECTED", false);
            }
          }
        }
      }
    }

    // handle service arrival/departure

    boolean service_change = false;

    for (int i=0; i < users.size(); ++i) {
      String u = users.get(i);
      if (u.startsWith("/")) {
	
	Global.dbgmsg("service " + u);
        ServiceClient svc = find_service(u);
        if (svc != null) {

          if (addrem.get(i).equals("+")) {
            if (!svc.getConnected()) {
              service_change = true;
              svc.arrived();
	      Global.main_window.enable_service(svc);
	      
	      if (default_service == null ||
                  u.equals(first_service_login))
                set_default_service(u);
            }

	  } else {

            if (svc.getConnected()) {

              service_change = true;
              svc.departed();
	      Global.main_window.disable_service(svc);

              if (svc == default_service) {

                // search for next default service
                int j;
                ServiceClient next_default_service;

                for (j=0; j < services.size(); ++j) {
                  if (services.get(j).getConnected())
		    break;
                }

                if (j < services.size())
                  set_default_service(services.elementAt(j).getLogin());
                else
                  set_default_service("");
              }
            }
          }
        }
      }
    }

    Global.main_window.add_rem_users(login, users, addrem);
    return true;
  }


  public boolean handle_event(EventMsg msg0)
  {
    final EventMsg msg = msg0;
    String from = msg.get_sender();
    EventArgEnum en = msg.get_enum();

    // Global.dbgmsg("Core handle event thread: " + Thread.currentThread().toString() + " : " + from);

    //Misc.msg("Event: " + from);
      
    while (Global.main_window == null) {
      Global.dbgmsg("main_window = 0");
      Global.sleep(100);
    }
	  
    if (REPORT_EVENTS) {
      Global.dbgmsg("core received event: " + from);
    }

    if (from.equals(TO_SERVER)) {

      send_to_server(en.next_arg());
      return true;

    } else if (from.equals(GUI_EVENT)) {

      String what = en.next_arg();

      if (what.equals(GUI_CLOSE) || what.equals(GUI_QUIT)) {
	Global.shut_down();
      }

      if (what.startsWith(GUI_INPUT)) {

	Global.main_window.set_input("");

	String s = what.substring(GUI_INPUT.length());

	if (s.length() > 0) {

	  if (s.charAt(0) != '#') {

	    if (current_talk_context.input_prefix.equals(""))
	      send_to_server(s);
	    else
	      send_to_server(current_talk_context.input_prefix + " " + s);

	  } else {

	    Global.ggs.ef.send(s.substring(1));
	  }
	}

	return true;

      } else if (what.equals(GUI_RESIZE)) {

	//           Dimension dim = Global.main_window.getSize();

	//           if (dim.height < GUI_MIN_Y) {
	//             dim.height = GUI_MIN_Y;
	//             Global.main_window.setSize(dim);
	//           }

	return true;
      }

    } else if (from.equals(MainWindow.USER_ID)) {

      String what = en.next_arg();
      String who = en.next_arg();

      if (what.equals(MainWindow.SINGLE_ID)) {

	set_current_talk_context(who);

      } else if (what.equals(MainWindow.DOUBLE_ID)) {

	if (default_service != null) {
	  default_service.clickedOnUser(who);
	}
      }

      return true;

    } else if (from.equals(MainWindow.TALK_ID)) {

      String what = en.next_arg();
      String who  = en.next_arg();

      if (what.equals(MainWindow.SINGLE_ID)) {
	set_current_talk_context(who);
      } else if (what.equals(MainWindow.DOUBLE_ID)) {
	delete_talk_context(who);
      }
      return true;

      //      } else if (from.equals(MainWindow.SERVICE_ID)) {

      //  // fixme: to be moved to service class
	
      //         String what = en.next_arg();
      //         int sel_index = -1;

      //         try { sel_index = Integer.parseInt(en.next_arg()); }
      //         catch (NumberFormatException e) {}

      //         if (what.equals(MainWindow.SINGLE_ID)) {

      //           // find client and relative line number

      //           int ln = 0;
      //           int i = 0;

      //           while (sel_index > ln + (sl_line_numbers.get(i)).intValue()) {
      //             ln += (sl_line_numbers.get(i)).intValue();
      //             ++i;
      //           }

      //           if (ln != sel_index) { // ignore "/service :" line

      //             if (what.equals(MainWindow.SINGLE_ID))
      //               sl_services.get(i).selected(sel_index - ln - 1);
      //           }
      //         }

      //         return;

    } else if (from.equals(GGS_EVENT)) {

      if (handle_central_msg(msg))
        return true;
      
      // not handled

      EventArgEnum e = msg.get_enum();
      if (!e.has_more_args())
	return false;

      String s = e.next_arg();
      String c = talk_context_name(s);

      for (String t : msg.get_args()) {
	output_text(c, t, true);
      }

      return true;
    }
    
    // not handled
    
    output_text(CENTRAL_TALK_CONTEXT, "??? not handled:");
    
    for (String s : msg.get_args()) {
      output_text(CENTRAL_TALK_CONTEXT, "??? " + s);
    }
    return false;
  }

  public void output_text(String cname, String text)
  {
    try {
      if( cname.startsWith(CENTRAL_TALK_CONTEXT) || cname.startsWith("/") )
        output_text(cname, text, true);
      else {

        String file_name = new String();
        file_name += "logs/";
        StringTokenizer token = new StringTokenizer(text, " ");
        String formatted_text = new String();

        String first = token.nextToken();
        if( first.compareTo("t")==0 ) { // t blob ...
          file_name += token.nextToken(); // "blob"
          // remove first two tokens from text
          // find position of first non-whitespace character
          int index = 0;
          while( text.charAt(index++) != ' ' )
            ;
          // find first whitespace character
          while( text.charAt(index++) != ' ' )
            ;
          formatted_text = Global.options.getString("lo") + ": " + text.substring(index);
        }
        else if( first.startsWith(".") ) { // .svensk ...
          file_name += first;
          int index = 0;
          while( text.charAt(index++) != ' ' )
            ;
          formatted_text = text.substring(index);
        }
        else {
          formatted_text = text;
          file_name += first.substring(0,first.length()-1);
        }
        formatted_text = df.format(new Date()) + " " + formatted_text;
        output_text(cname, formatted_text, true);

        FileWriter file = new FileWriter(file_name, true);
        PrintWriter out = new PrintWriter(file);
        out.println(formatted_text);
        out.close();
      }
    } catch( IOException exception ) {
      System.out.println( exception.getMessage() );
    }
  }

  public void output_text(String cname, String text, boolean update_arrived)
  {
    int i = talk_context_index(cname);
    TalkContext tc;

    if (i < 0) {

      // not found: create new talk context in background

      tc = new TalkContext(cname, cname, "t " + cname, true);
      talk_contexts.add(tc);

    } else {

      tc = talk_contexts.get(i);

    }

    String s = "\n"+text;
    tc.text.append(s);

    if (update_arrived) {
      if (tc != current_talk_context) {
        if (tc.message_arrived == false) {
          tc.message_arrived = true;
          Global.main_window.update_talk_list(talk_contexts, current_talk_context);	  
        }
      }

      if (tc == current_talk_context) {
        Global.main_window.append(s);	
        current_talk_context.message_arrived = false;
      }
    }
  }


  private String talk_context_name(EventMsg msg)
  {
    EventArgEnum en = msg.get_enum();

    if (!en.has_more_args())
      return CENTRAL_TALK_CONTEXT;

    return talk_context_name(en.next_arg());
  }
    
  private String talk_context_name(String from)     
  {
    int i = from.indexOf(":");
    if (i <= 0) 
      return CENTRAL_TALK_CONTEXT;

    from = from.substring(0, i);
    i = from.indexOf(" ");
    if (i >= 0) 
      from = from.substring(0, i);

    return from;
  }

  private String msg_text(EventMsg msg) 
  {
    EventArgEnum en = msg.get_enum();
    StringBuffer sb = new StringBuffer();
    while (en.has_more_args()) {
      sb.append(en.next_arg());
      if (en.has_more_args())
	sb.append("\n");
    }

    return new String(sb);
  }

  private ServiceClient find_service(String login)
  {
    for (ServiceClient s : services) {
      if (s.getLogin().equals(login))
	return s;
    }
    return null;
  }

  private ServiceClient find_service_by_name(String name)
  {
    for (ServiceClient s : services) {
      if (s.getName().equals(name))
	return s;
    }
    return null;
  }

  public void set_default_service(String login) 
  {
    ServiceClient service = find_service(login);
    if (!service.getConnected())
      return;
    
    // Global.dbgmsg("SET BY LOGIN " + login + " " + service.getLogin());
    
    Global.main_window.set_default_service(service);
    default_service = service;
  }

  public void set_default_service_by_name(String name) 
  {
    ServiceClient service = find_service_by_name(name);
    if (!service.getConnected())
      return;

    // Global.dbgmsg("SET BY NAME " + name + " " + service.get_name());
    Global.main_window.set_default_service(service);
    default_service = service;
  }

  // fixme: to be moved to service class
  
  //   private void add_service_list(ServiceClient service, Vector<String> sl)
  //   {
  //     Vector<String> list = service.get_list();
  //     sl.add(client.game_name+ ": "+ list.size());
  //     for (String s : list)
  //       sl.add("  " + s);
  //     sl_services.add(service);
  //     sl_line_numbers.add(new Integer(list.size()+1));
  //   }

  //   public void service_list_update(ServiceClient cl)
  //   {
  //     Vector<String> service_list = new Vector<String>();

  //     sl_services = new Vector<Client>();
  //     sl_line_numbers = new Vector<Integer>();

  //     // default client's list first

  //     if (default_service_client != null)
  //       add_service_list(default_service_client, service_list);

  //     for (ServiceClient c : Global.services) {
  //       if (c != default_service_client && c.connected())
  //         add_service_list(c, service_list);
  //     }

  //     Global.main_window.new_service_list(service_list);
  //   }

  private void init_talk_contexts()
  {
    talk_contexts = new Vector<TalkContext>();
    talk_contexts.add(new TalkContext(".chat", ".chat", "t .chat", false));
    try {
      final String file_name = "channels.txt";
      FileReader file = new FileReader(file_name);
      BufferedReader in = new BufferedReader(file); 
      String line;
      while( (line=in.readLine()) != null) {
	String text[] = line.split("\\s");
	for( int i=0; i< text.length; i++ ) {
	  String val = text[i];
	  if (val.length() > 0) {
	    talk_contexts.add(new TalkContext(val, val, "t " + val, false));
	    System.out.println("Added " + val);
	  }
	}
      }
      file.close();
    } 
    
    catch( IOException exception ) {
      System.out.println( exception.getMessage() );
    }
    talk_contexts.add(new TalkContext(CENTRAL_TALK_CONTEXT, "", "", false));
    set_current_talk_context(CENTRAL_TALK_CONTEXT);
  }

  private int talk_context_index(String name)
  {
    for (int i=0; i < talk_contexts.size(); ++i) {
      if (name.equals(talk_contexts.get(i).name))
	return i;
    }

    return -1;
  }

  public void set_current_talk_context(String name)
  {
    // Global.dbgmsg("Core set current context: " + Thread.currentThread().toString());
    
    if (current_talk_context != null && current_talk_context.name.equals(name))
      return;

    int i = talk_context_index(name);
    TalkContext tc;

    if (i < 0) {

      // create new context

      tc = new TalkContext(name, name, "t " + name, true);
      talk_contexts.add(tc);
    } else {
      tc = (TalkContext) talk_contexts.get(i);
    }
    current_talk_context = tc;
    current_talk_context.message_arrived = false;

    Global.main_window.set_text(new String(current_talk_context.text));
    Global.main_window.update_talk_list(talk_contexts, current_talk_context);
  }

  public void delete_talk_context(String name) 
  {
    int i = talk_context_index(name);
    if (i < 0) 
      return;

    TalkContext tc = talk_contexts.get(i);
    if (!tc.deletable) 
      return;

    talk_contexts.remove(i);

    // find new context: search for one with unread message
    
    int j;
    for (j=0; j < talk_contexts.size(); ++j)
      if (talk_contexts.get(j).message_arrived) 
	break;

    if (j >= talk_contexts.size()) {
      
      // none exists, try same position
      j = i;
      if (j >= talk_contexts.size())
	j--;
    }

    set_current_talk_context(talk_contexts.get(j).name);
  }


  public void shut_down()
  {
    for (int i=0; i < services.size(); i++) {
      ServiceClient svc = services.elementAt(i);
      if (svc != null) {
	Global.dbgmsg("  ... " + svc.getLogin());
	svc.shutDown();
	services.setElementAt(null, i);
      }
    }
  }

  // RandProvider interface
  public Random getRand() { return null; }

  // Sender interface
  public void send(String msg)
  {
    try {
      Global.sc.write_line(msg);
    }
    catch (IOException e) {
      Global.errmsg("socket write failed");
    }
  }

}
