/* $Id: HoriLayout.java,v 1.1.1.1 2006/08/21 04:26:01 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.ggsa.main;

import java.lang.Error;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class HoriLayout implements LayoutManager
{
  public final static int V_PARENT = 1;  // height from parent
  public final static int H_PARENT = 2;  // width from parent

  // if !V_PARENT
  public final static int V_TOP    = 4;
  public final static int V_BOTTOM = 8;
  public final static int V_CENTER = 0;  // default

  // if H_PARENT
  public final static int H_LEFT   = 16;
  public final static int H_RIGHT  = 32;
  public final static int H_SPREAD = 0;  // default

  public final static int NO_LEFT_PHANTOM  = 64;
  public final static int NO_RIGHT_PHANTOM = 128;

  public boolean is_V_PARENT()  { return (align & V_PARENT) != 0; }
  public boolean is_H_PARENT()  { return (align & H_PARENT) != 0; }

  public boolean is_V_TOP()     { return (align & V_TOP) != 0;    }
  public boolean is_V_BOTTOM()  { return (align & V_BOTTOM) != 0; }
  public boolean is_V_CENTER()  { return !is_V_TOP() && !is_V_BOTTOM(); }

  public boolean is_H_LEFT()    { return (align & H_LEFT) != 0;   }
  public boolean is_H_RIGHT()   { return (align & H_RIGHT) != 0;  }
  public boolean is_H_SPREAD()  { return !is_H_LEFT() && !is_H_RIGHT(); }

  public boolean is_LEFT_PHANTOM()  { return (align & NO_LEFT_PHANTOM) == 0; }
  public boolean is_RIGHT_PHANTOM() { return (align & NO_RIGHT_PHANTOM) == 0; }

  private final static int P_WEIGHT_DEF = 1;

  private final static int PREFERRED    = 0;
  private final static int MINIMUM      = 1;
  private final static int MAXIMUM      = 2;


  private int  align;
  private int  h_gap;
  private int  p_weight;
  private Hashtable<Component,HoriLayoutConstraints> components =
    new Hashtable<Component,HoriLayoutConstraints>();

  public HoriLayout(int align, int h_gap) { this(align, h_gap, P_WEIGHT_DEF); }

  public HoriLayout(int align, int h_gap, int p_weight)
  {
    this.align = align;
    this.h_gap = h_gap;
    this.p_weight = p_weight;
  }    

  public void addLayoutComponent(String name, Component component) 
  {
  }
	
  public void removeLayoutComponent(Component component) 
  {
    components.remove(component);
  }
	
  public void setConstraints(Component comp, HoriLayoutConstraints cons) 
  {
    if (!is_H_PARENT() && cons.is_ADJ())
      throw new Error("adj. component in !H_PARENT-conainer");

    try { components.put(comp, (HoriLayoutConstraints)cons.clone() ); }
    catch( CloneNotSupportedException e ) { }
  }

  public HoriLayoutConstraints getConstraints(Component comp) 
  {
    HoriLayoutConstraints cons = components.get(comp);

    if (cons != null) {

      try { return (HoriLayoutConstraints) cons.clone(); }
      catch(CloneNotSupportedException e ) { return null; }
                
    } else return null;
  }


  public Dimension minimumLayoutSize(Container parent) 
  { 
    return calcLayoutSize(parent, MINIMUM); 
  }

  public Dimension preferredLayoutSize(Container parent) 
  { 
    return calcLayoutSize(parent, PREFERRED);
  }

  private Dimension calcLayoutSize(Container parent, int size_type) 
  {
    int max_y = 0, sum_x = 0;
    Dimension comp_size = null;
    Dimension cont_size = parent.getSize();

    /*
System.out.print("calcsize " + parent.getSize().width + " "
		   + parent.getSize().height + " " + size_type + " ");
   */

    for (int counter = 0; counter < parent.getComponentCount(); counter++) {

      Component comp = parent.getComponent(counter);
      Dimension dim = comp.getSize();

      if (is_V_PARENT()) {
	dim.height = cont_size.height - parent.getInsets().top - parent.getInsets().bottom;
      }

      if (is_H_PARENT()) {
	dim.width = cont_size.width - parent.getInsets().right - parent.getInsets().left;
      }
      
      comp.setSize(dim);
      
      if (size_type == PREFERRED) { 

	comp_size = comp.getPreferredSize(); 
	
      } else if (size_type == MINIMUM) { 
	
	comp_size = comp.getMinimumSize(); 
	
      } else { 
	
	comp_size = comp.getMaximumSize(); 
      }
      
      if (comp_size.height > max_y) max_y = comp_size.height;
      
      sum_x += sum_x + comp_size.width + h_gap;
    }
    
    sum_x -= h_gap;
    
    sum_x += parent.getInsets().left + parent.getInsets().right;
    max_y += parent.getInsets().top + parent.getInsets().bottom;

    if (is_V_PARENT()) max_y = parent.getSize().height;
    if (is_H_PARENT()) sum_x = parent.getSize().width;

    //    System.out.println(" -> " + sum_x + " " + max_y);

    return new Dimension(sum_x, max_y);
  }


  public void layoutContainer(Container parent) 
  {
    //    System.out.println("hori layout " + 
    //		       parent.getSize().width + " " + parent.getSize().height);

    Dimension dim = calcLayoutSize(parent, PREFERRED);

    int cw = parent.getSize().width - parent.getInsets().left - parent.getInsets().right;
    int ch = parent.getSize().height - parent.getInsets().top - parent.getInsets().bottom;

    if (is_H_LEFT() || is_H_RIGHT()) {

      //      System.out.println("H_LEFT | H_RIGHT");

      int cx, cy;

      if (is_H_RIGHT()) {
	int total_width = 0;

	for(int counter = 0; counter < parent.getComponentCount(); counter++) {
	  total_width += parent.getComponent(counter).getPreferredSize().width + h_gap;
	}
	
	total_width -= h_gap;
	cx = dim.width-parent.getInsets().right-total_width;

      } else {

	cx = parent.getInsets().left;

      }

      cy = parent.getInsets().top;

      for(int counter = 0; counter < parent.getComponentCount(); counter++) {
	
	Component comp      = parent.getComponent(counter);
	Dimension comp_size = comp.getPreferredSize();
	Point     comp_loc  = comp.getLocation();
	
	int y, width, height;
	
	if (is_V_PARENT()) {
	  
	  height = ch;
	  y = cy;
	  
	} else {
	  
	  height = comp_size.height;
	  
	  if      (is_V_CENTER()) y = cy + (ch - height)/2;
	  else if (is_V_TOP())    y = cy;
	  else                    y = cy + height - (ch+1)/2;

	}

	/*
	System.out.println("comp " + (counter+1) + " " + 
			   cx + " " + y + " " + comp_size.width + " " + height);
	System.out.println("ch= " + ch + " " + " height=" + height);
	*/
	comp.setLocation(cx, y);
	comp.setSize(comp_size.width, height);
	// comp.doLayout();
	
	cx += comp_size.width + h_gap;
      }

    } else {

      // H_SPREAD

      //      System.out.println("H_SPREAD");

      int fixed_width = 0;
      int w_sum = 0;

      for (int counter = 0; counter < parent.getComponentCount(); counter++) {
	Component comp = parent.getComponent(counter);
	HoriLayoutConstraints cons = getConstraints(comp);

	if (cons == null) throw new Error("no hori_constraint found");

	if (cons.is_FIX()) fixed_width += comp.getPreferredSize().width;
	else w_sum += cons.get_ADJ_weight();

	w_sum += cons.get_glue_weight();
	fixed_width += h_gap;
      }
      
      fixed_width -= h_gap;

      if (is_LEFT_PHANTOM())  w_sum += p_weight;
      if (is_RIGHT_PHANTOM()) w_sum += p_weight;

      int adj_width = cw - fixed_width;
      
      int cx = parent.getInsets().left;
      int cy = parent.getInsets().top;

      if (is_LEFT_PHANTOM()) 
	cx += Math.round((p_weight/(double)w_sum)*adj_width);

      for(int counter = 0; counter < parent.getComponentCount(); counter++) {
	
	Component comp = parent.getComponent(counter);
	Dimension comp_size = comp.getPreferredSize();
	Point     comp_loc  = comp.getLocation();
	
	int y, width, height;
	
	if (is_V_PARENT()) {
	  
	  height = ch;
	  y = cy;
	  
	} else {
	  
	  height = comp_size.height;
	  
	  if      (is_V_CENTER()) y = cy + (ch - height)/2;
	  else if (is_V_TOP())    y = cy;
	  else                    y = cy + height - (ch+1)/2;
	  
	}

	/*
System.out.println(is_V_CENTER());
System.out.println(y);
System.out.println(cy);
System.out.println(ch);
System.out.println(height);
*/

	HoriLayoutConstraints cons = getConstraints(comp);
	if (cons == null) throw new Error("no hori_constraint found");

	if (cons.is_ADJ()) {
	  width = (int)Math.round((cons.get_ADJ_weight()/(double)w_sum)*adj_width);
	} else {
	  width = comp_size.width;
	}

	comp.setLocation(cx, y);
	comp.setSize(width, height);

	/*
	System.out.println("comp " + (counter+1) + " " + 
			   cx + " " + y + " " + width + " " + height);
	*/
	cx += width + h_gap;

	int glue = cons.get_glue_weight();

	//System.out.println("glue=" + glue);

	if (glue > 0) cx += Math.round((glue/(double)w_sum)*adj_width);
      }
    }

    //System.out.println("done.\n");


  }

}
