/*
 * $Id: MainWindow.java,v 1.20 2006/09/04 17:25:54 orts_mburo Exp $
 * 
 * (c) Michael Buro, licensed under GPL2
 */

package ggs.ggsa.main;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import ggs.shared.*;

public class MainWindow extends javax.swing.JFrame implements ActionListener,WindowListener
{
  public static final int CHANGE_LIST_MAX_LEN = 100;
  public static final String USER_ID    = "USER";
  public static final String SERVICE_ID = "SERVICE";
  public static final String TALK_ID    = "TALK";

  public static final String SINGLE_ID  = "SELECTED";  
  public static final String DOUBLE_ID  = "DOUBLE";
  
  private Set<String> user_set = new TreeSet<String>(new GreaterUser());
  private LinkedList<String> history = new LinkedList<String>();
  private JMenu service_menu; // list of services

  private JMenu prev_service_cmd_menu;
  
  // functions that convert a list selection into a user name
  
  interface StringToUserName
  {
    public String convert(String s);
  }
  
  class IdentityConvert implements StringToUserName 
  {
    public String convert(String s) { return s; }
  }

  class RemoveFirstConvert implements StringToUserName 
  {
    public String convert(String s) { 
      // remove first character
      return s.substring(1);
    }
  }

  // send event message on mouse click to user lists
  
  class ClientListMouseListener extends MouseAdapter 
  {
    ClientListMouseListener(JList list, String msg_id, StringToUserName stn) 
    {
      this.list = list;
      this.msg_id = msg_id;
      this.stn = stn;
    }

    public void mouseClicked(MouseEvent e) 
    {
      int index = list.locationToIndex(e.getPoint());
      if (index < 0) return;
      
      String selection = (String) list.getModel().getElementAt(index);
	
      if (e.getClickCount() == 1) {
	Global.core.handle_event(new EventMsg(msg_id, MainWindow.SINGLE_ID, stn.convert(selection)));
      } else if (e.getClickCount() == 2) {
	Global.core.handle_event(new EventMsg(msg_id, MainWindow.DOUBLE_ID, stn.convert(selection)));
      }
    }
  
    private JList list;
    private String msg_id;
    private StringToUserName stn;
  }
  
  /** Creates new form MainWindow */
  public MainWindow(Geometry geom) 
  {
    initComponents();
    setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);

    addWindowListener(this);
    
    service_tabs.addChangeListener(new ChangeListener() {

        public void stateChanged(ChangeEvent evt) {
	  JTabbedPane pane = (JTabbedPane)evt.getSource();
	  int sel = pane.getSelectedIndex();
	  if (sel < 0)
	    return;
	  Global.core.set_default_service_by_name(pane.getTitleAt(sel));
        }
      });
    
    // remove items
    DefaultListModel dm = new DefaultListModel();
    history_list.setModel(dm);
    user_list.setModel(dm);
    // buddy_list.setModel(dm);
    talk_list.setModel(dm);
    
    input.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {

	  String cmd = e.getActionCommand();

	  if (Global.core != null) {

	    // there might be several newlines in the string
	    // due to copy&paste, don't push in that case

	    if (input.getText().indexOf('\n') < 0)
	      input.push();
	    set_input("");

	    Global.core.handle_event(new EventMsg(Core.GUI_EVENT, Core.GUI_INPUT+cmd));
	    input.requestFocus();
	  }
	}}
			    );

    user_list.addMouseListener(new ClientListMouseListener(user_list,
							   USER_ID,
							   new IdentityConvert()));
    talk_list.addMouseListener(new ClientListMouseListener(talk_list,
							   TALK_ID,
							   new RemoveFirstConvert()));
    // buddy_list.addMouseListener(
    //   new ClientListMouseListener(buddy_list, USER_ID, new IdentityConvert())
    // );
    history_list.addMouseListener(new ClientListMouseListener(history_list,
							      USER_ID,
							      new RemoveFirstConvert()));

    // Window Menu

    JMenu window = menu_bar.add(new JMenu("GGS"));

    // User List

    JMenuItem mi = new JMenuItem("User List");
    window.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  Global.core.set_current_talk_context(Global.core.CENTRAL_TALK_CONTEXT);
	  Global.core.send_to_server("who");
	}});

    // Login History

    mi = new JMenuItem("History");
    window.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  Global.core.set_current_talk_context(Global.core.CENTRAL_TALK_CONTEXT);
	  Global.core.send_to_server("H");
	}});

    // About

    mi = new JMenuItem("About");
    window.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  String s = Core.CENTRAL_TALK_CONTEXT;
	  Global.core.set_current_talk_context(s);
	  Global.core.output_text(s, "");
	  Global.core.output_text(s,"GGSA - An applet for GGS (Version " + Global.VERSION + ")");
	  Global.core.output_text(s,"URL: http://www.cs.ualberta.ca/~mburo/ggsa");
	  Global.core.output_text(s,"");
	  Global.core.output_text(s,"(c) Michael Buro, mburo@cs.ualberta.ca");
	  Global.core.output_text(s,"(c) Daniel Lidstrom, daniel_lidstrom@chello.se");
	  Global.core.output_text(s,"");
	  Global.core.output_text(s,"This program is free software; you can redistribute it and/or modify");
	  Global.core.output_text(s,"it under the terms of the GNU General Public License as published by");
	  Global.core.output_text(s,"the Free Software Foundation; either version 2 of the License, or");
	  Global.core.output_text(s,"(at your option) any later version.");
	  Global.core.output_text(s,"");
	  Global.core.output_text(s,"This program is distributed in the hope that it will be useful,");
	  Global.core.output_text(s,"but WITHOUT ANY WARRANTY; without even the implied warranty of");
	  Global.core.output_text(s,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the");
	  Global.core.output_text(s,"GNU General Public License for more details.");
	  Global.core.output_text(s,"");
	  Global.core.output_text(s,"You should have received a copy of the GNU General Public License");
	  Global.core.output_text(s,"along with this program; if not, write to the Free Software");
	  Global.core.output_text(s,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA");
	}});

    // Help

    mi = new JMenuItem("News");
    window.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  Global.core.set_current_talk_context(Global.core.CENTRAL_TALK_CONTEXT);
	  Global.core.send_to_server("help news");
	}});

    mi = new JMenuItem("Sys.Aliases");
    window.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  Global.core.set_current_talk_context(Global.core.CENTRAL_TALK_CONTEXT);
	  Global.core.send_to_server("salias");
	}});

    window.addSeparator();

    // Quit

    mi = new JMenuItem("Quit");
    window.add(mi);
    mi.setEnabled(true);
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  Global.shut_down();
	  // was: Global.ggsa.destroy();
	}});

    // Service Menu

    service_menu = menu_bar.add(new JMenu("Service"));

    setTitle("GGSA");

    if (geom.get_x() > 0) 
      setLocation(geom.get_x(), geom.get_y());

    setSize(geom.get_w(), geom.get_h());
    validateTree();
    
    // setBackground(new Color(Global.options.getInteger("bg").intValue()));
    // setForeground(new Color(Global.options.getInteger("fg").intValue()));

    // set min/preferred sizes of lists

    //     Global.set_dim(talk_list,       Global.options.getInteger("ulw").intValue(), -1);
    //     Global.set_dim(user_list_panel, Global.options.getInteger("ulw").intValue(), -1);
    //     Global.set_dim(talk_panel,      -1, Global.options.getInteger("tph").intValue());

    //     main_hsplit.resetToPreferredSizes();
    //     talk_hsplit.resetToPreferredSizes();
    //     vsplit.resetToPreferredSizes();

    //     talk_hsplit.doLayout();
    //     user_list_panel.doLayout();
    //     talk_panel.doLayout();

    // compensate for window decoration width
    int mhs = main_hsplit.getWidth() - Global.options.getInteger("ulw").intValue() - 20;
    main_hsplit.setDividerLocation(mhs);
    validateTree();
    // main_hsplit.doLayout();

    vsplit.setDividerLocation(geom.get_h() - Global.options.getInteger("tph").intValue());
    validateTree();
    // vsplit.doLayout();

    talk_hsplit.setDividerLocation(Global.options.getInteger("ulw").intValue());
    // talk_hsplit.doLayout();
    validateTree();

    setVisible(true);
    
    repaint();
  }

  // window listener functions
  public void windowClosing(WindowEvent event) {
    Global.shut_down();
  }
  public void windowOpened(WindowEvent event ) { }
  public void windowIconified(WindowEvent event ) { }
  public void windowDeiconified(WindowEvent event ) { }
  public void windowClosed(WindowEvent event ) { }
  public void windowActivated(WindowEvent event ) { }
  public void windowDeactivated(WindowEvent event ) { }
  public void actionPerformed(ActionEvent e) { }
  
  
  public void init_service(ServiceClient svc)
  {
    // add menu item
    
    final ServiceClient service = svc;
    JMenuItem mi = new JMenuItem(svc.getName());
    mi.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  Global.core.set_default_service(service.getLogin());
	}});
    
    mi.setEnabled(false);
    service_menu.add(mi);

    // add service tab
    
    JPanel panel = svc.getTabPanel();
    panel.setName(svc.getName());

    // insert tab at its position (sorted)

    int i;
    
    for (i=0; i < service_tabs.getTabCount(); ++i) {
      if (((JPanel)service_tabs.getComponentAt(i)).getName().compareTo(svc.getName()) > 0)
	break;
    }
    
    service_tabs.add(panel, i);

    invalidate();
    validateTree();    

    repaint();
  }


  public void enable_service(ServiceClient svc)
  {
    int i;
    for (i=0; i < service_menu.getItemCount(); ++i) {
      if (service_menu.getItem(i).getText().equals(svc.getName()))
	break;
    }

    if (i >= service_menu.getItemCount()) {
      Global.dbgmsg("enable_service: unknown service");
      return;
    }

    service_menu.getItem(i).setEnabled(true);

    validateTree();    
    repaint();
  }


  public void disable_service(ServiceClient svc)
  {
    // disable menu item

    int i;
    for (i=0; i < service_menu.getItemCount(); ++i) {
      if (service_menu.getItem(i).getText().equals(svc.getName())) {
	break;
      }
    }

    if (i >= service_menu.getItemCount()) {
      Global.dbgmsg("disable_service: unknown service");
      return;
    }

    service_menu.getItem(i).setEnabled(false);

    validateTree();    
    repaint();
  }
  

  public void set_default_service(ServiceClient svc)
  {
    // remove previous service menu

    if (prev_service_cmd_menu != null)
      menu_bar.remove(prev_service_cmd_menu);

    if (svc != null) {

      // select service tab

      for (int i=0; i < service_tabs.getTabCount(); ++i) {
	if (((JPanel)service_tabs.getComponentAt(i)).getName().equals(svc.getName())) {
	  service_tabs.setSelectedIndex(i);
	  break;
	}
      }

      // set talk context

      Global.core.set_current_talk_context(svc.getLogin());
      
      prev_service_cmd_menu = svc.getMenu();
      prev_service_cmd_menu.setText(svc.getName());
      menu_bar.add(prev_service_cmd_menu);

    } else {

      // disable all tabs

      Component[] co = service_tabs.getComponents();

      for (Component c : co)
	c.setEnabled(false);
      
      prev_service_cmd_menu = null;
      return;

    }

    menu_bar.revalidate();  // otherwise, changes don't show
    menu_bar.repaint();
    validateTree();
    repaint();
  }
  
  /** This method is called from within the constructor to
   * initialize the form.
   * WARNING: Do NOT modify this code. The content of this method is
   * always regenerated by the Form Editor.
   */
  // <editor-fold defaultstate="collapsed" desc=" Generated Code ">//GEN-BEGIN:initComponents
  private void initComponents() {
    main_hsplit = new javax.swing.JSplitPane();
    vsplit = new javax.swing.JSplitPane();
    service_tabs = new javax.swing.JTabbedPane();
    talk_panel = new javax.swing.JPanel();
    talk_hsplit = new javax.swing.JSplitPane();
    talkPanel2 = new javax.swing.JPanel();
    input = new ggs.ggsa.main.JHistTextField();
    talk_label = new javax.swing.JLabel();
    jScrollPane3 = new javax.swing.JScrollPane();
    text = new javax.swing.JTextArea();
    tc_list_panel = new javax.swing.JPanel();
    jScrollPane4 = new javax.swing.JScrollPane();
    talk_list = new javax.swing.JList();
    lists_panel = new javax.swing.JPanel();
    jSplitPane2 = new javax.swing.JSplitPane();
    user_list_panel = new javax.swing.JPanel();
    user_list_scroll_pane = new javax.swing.JScrollPane();
    user_list = new javax.swing.JList();
    history_list_panel = new javax.swing.JPanel();
    jScrollPane2 = new javax.swing.JScrollPane();
    history_list = new javax.swing.JList();
    menu_bar = new javax.swing.JMenuBar();

    setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
    main_hsplit.setDividerLocation(550);
    main_hsplit.setDividerSize(6);
    main_hsplit.setResizeWeight(1.0);
    vsplit.setDividerLocation(300);
    vsplit.setDividerSize(6);
    vsplit.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
    vsplit.setResizeWeight(0.5);
    vsplit.setTopComponent(service_tabs);

    talk_panel.setLayout(new java.awt.BorderLayout());

    talk_hsplit.setDividerLocation(70);
    talk_hsplit.setDividerSize(6);

    talk_label.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
    talk_label.setText("GGS");

    text.setColumns(20);
    text.setEditable(false);
    text.setLineWrap(true);
    text.setRows(5);
    jScrollPane3.setViewportView(text);

    org.jdesktop.layout.GroupLayout talkPanel2Layout = new org.jdesktop.layout.GroupLayout(talkPanel2);
    talkPanel2.setLayout(talkPanel2Layout);
    talkPanel2Layout.setHorizontalGroup(
      talkPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
      .add(org.jdesktop.layout.GroupLayout.TRAILING, talkPanel2Layout.createSequentialGroup()
        .addContainerGap()
        .add(talk_label)
        .add(15, 15, 15)
        .add(input, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 418, Short.MAX_VALUE))
      .add(org.jdesktop.layout.GroupLayout.TRAILING, jScrollPane3, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 470, Short.MAX_VALUE)
    );
    talkPanel2Layout.setVerticalGroup(
      talkPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
      .add(org.jdesktop.layout.GroupLayout.TRAILING, talkPanel2Layout.createSequentialGroup()
        .add(jScrollPane3, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 265, Short.MAX_VALUE)
        .addPreferredGap(org.jdesktop.layout.LayoutStyle.RELATED)
        .add(talkPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.BASELINE)
          .add(input, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE, 28, org.jdesktop.layout.GroupLayout.PREFERRED_SIZE)
          .add(talk_label)))
    );
    talk_hsplit.setRightComponent(talkPanel2);

    tc_list_panel.setLayout(new java.awt.BorderLayout());

    tc_list_panel.setBorder(javax.swing.BorderFactory.createTitledBorder("Contexts"));
    talk_list.setModel(new javax.swing.AbstractListModel() {
      String[] strings = { "Item 1", "Item 2", "Item 3", "Item 4", "Item 5" };
      public int getSize() { return strings.length; }
      public Object getElementAt(int i) { return strings[i]; }
    });
    talk_list.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
    jScrollPane4.setViewportView(talk_list);

    tc_list_panel.add(jScrollPane4, java.awt.BorderLayout.CENTER);

    talk_hsplit.setLeftComponent(tc_list_panel);

    talk_panel.add(talk_hsplit, java.awt.BorderLayout.CENTER);

    vsplit.setRightComponent(talk_panel);

    main_hsplit.setLeftComponent(vsplit);

    jSplitPane2.setDividerLocation(150);
    jSplitPane2.setDividerSize(6);
    jSplitPane2.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
    jSplitPane2.setResizeWeight(0.05);
    user_list_panel.setLayout(new java.awt.BorderLayout());

    user_list_panel.setBorder(javax.swing.BorderFactory.createTitledBorder("Users"));
    user_list.setModel(new javax.swing.AbstractListModel() {
      String[] strings = { "Item 1", "Item 2", "Item 3", "Item 4", "Item 5" };
      public int getSize() { return strings.length; }
      public Object getElementAt(int i) { return strings[i]; }
    });
    user_list_scroll_pane.setViewportView(user_list);

    user_list_panel.add(user_list_scroll_pane, java.awt.BorderLayout.CENTER);

    jSplitPane2.setRightComponent(user_list_panel);

    history_list_panel.setLayout(new java.awt.BorderLayout());

    history_list_panel.setBorder(javax.swing.BorderFactory.createTitledBorder("History"));
    history_list.setModel(new javax.swing.AbstractListModel() {
      String[] strings = { "Item 1", "Item 2", "Item 3", "Item 4", "Item 5" };
      public int getSize() { return strings.length; }
      public Object getElementAt(int i) { return strings[i]; }
    });
    jScrollPane2.setViewportView(history_list);

    history_list_panel.add(jScrollPane2, java.awt.BorderLayout.CENTER);

    jSplitPane2.setLeftComponent(history_list_panel);

    org.jdesktop.layout.GroupLayout lists_panelLayout = new org.jdesktop.layout.GroupLayout(lists_panel);
    lists_panel.setLayout(lists_panelLayout);
    lists_panelLayout.setHorizontalGroup(
      lists_panelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
      .add(jSplitPane2, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 96, Short.MAX_VALUE)
    );
    lists_panelLayout.setVerticalGroup(
      lists_panelLayout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
      .add(org.jdesktop.layout.GroupLayout.TRAILING, jSplitPane2, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 608, Short.MAX_VALUE)
    );
    main_hsplit.setRightComponent(lists_panel);

    setJMenuBar(menu_bar);

    org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
    getContentPane().setLayout(layout);
    layout.setHorizontalGroup(
      layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
      .add(main_hsplit, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 653, Short.MAX_VALUE)
    );
    layout.setVerticalGroup(
      layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
      .add(main_hsplit, org.jdesktop.layout.GroupLayout.DEFAULT_SIZE, 610, Short.MAX_VALUE)
    );
    pack();
  }// </editor-fold>//GEN-END:initComponents
  

  private boolean shutting_down = false;

  private int compare_names(String first, String second)
  {
    if (first == second) return 0;

    String s1 = first.toUpperCase();
    String s2 = second.toUpperCase();

    int r = s1.compareTo(s2);

    if (r == 0) return 0;

    boolean c1 = s1.length() > 0 && s1.charAt(0) == '/';
    boolean c2 = s2.length() > 0 && s2.charAt(0) == '/';

    if (c1 || c2) {
      if (c1 && !c2) return +1;
      if (c2 && !c1) return -1;
    }

    return r;
  }

  public final class GreaterUser implements Comparator<String>
  {
    public int compare(String first, String second)
    {
      return compare_names(first, second);
    }
  }

  public final class GreaterTC implements Comparator<TalkContext>
  {
    public int compare(TalkContext s1, TalkContext s2)
    {
      if (s1 == s2)
	return 0;

      int r = s1.name.toUpperCase().compareTo(s2.name.toUpperCase());

      if (r == 0) return 0;

      boolean d1 = s1.name.startsWith("/");
      boolean d2 = s2.name.startsWith("/");

      if (d1 ^ d2) 
	if (d2) 
	  return -1;
	else 
	  return +1;

      boolean c1 = !s1.message_arrived;
      boolean c2 = !s2.message_arrived;

      if (c1 || c2) {
	if (c1 && !c2) return +1;
	if (c2 && !c1) return -1;
      }

      return r;
    }
  }
  
  public void set_text(String t) 
  {
    text.setText(t);
    text.setCaretPosition(text.getDocument().getLength());
  }

  public void append(String s)
  {
    text.append(s);
    text.setCaretPosition(text.getDocument().getLength());
  }

  public void set_input(String s)
  {
    input.setText(s);
    input.setCaretPosition(input.getDocument().getLength());
    input.requestFocus();
  }

  /** client list changes */
  public void add_rem_users(String login, Vector<String> users, Vector<String> addrem) 
  {
    boolean changed = false;

    for (int i=0; i < users.size(); ++i) {

      if (addrem.get(i).equals("+"))
	changed |= user_set.add(users.get(i));
      else
	changed |= user_set.remove(users.get(i));
    }

    if (user_set.add(login)) {
      users.add(login);
      addrem.add("+");
    }

    // change list update

    int n = CHANGE_LIST_MAX_LEN;
    int k = users.size();
    int m = history.size();
      
    for (int i=0; i < k; ++i) {
      history.addFirst((addrem.get(i))+users.get(i));
      m++;
      if (m > n) {
	history.removeLast();
	m--;
      }
    }

    DefaultListModel history_model = new DefaultListModel();
    for (String s : history)
      history_model.addElement(s);

    history_list.setVisible(false);
    history_list.setModel(history_model);
    history_list.setVisible(true);

    if (changed) {

      // user list update
      user_list_panel.setBorder(javax.swing.BorderFactory.createTitledBorder("Users ("+ user_set.size()+")"));
      
      // set client list with sorted list of clients
      DefaultListModel user_model = new DefaultListModel();
      for (String s : user_set)
	user_model.addElement(s);

      user_list.setModel(user_model);
    }
  }

  public void update_talk_list(Vector<TalkContext> tcs, TalkContext current_tc) 
  {
    boolean input_focus = input.has_focus();

    Collections.sort(tcs, new GreaterTC());

    int line = 0;
    int index = 0;

    // create talk tabs
    
    talk_list.removeAll();

    int i = 0, selected = -1;
    DefaultListModel talk_model = new DefaultListModel();

    for (TalkContext tc : tcs) {
      String name = tc.name;
      if (tc != current_tc && tc.message_arrived)
	name = "*" + tc.name; // -> use RemoveOneConvert
      else
	name = " " + tc.name;
      if (tc == current_tc)
	selected = i;
      i++;
      talk_model.addElement(name);
    }

    talk_list.setModel(talk_model);
    talk_list.setSelectedIndex(selected);
    
    talk_label.setText(current_tc.name);
    
    if (input_focus) 
      input.requestFocus();

    validate();
    repaint();    
  }

  public void shut_down()
  {
    shutting_down = true;
    dispose();
  }

  // Variables declaration - do not modify//GEN-BEGIN:variables
  private javax.swing.JList history_list;
  private javax.swing.JPanel history_list_panel;
  private ggs.ggsa.main.JHistTextField input;
  private javax.swing.JScrollPane jScrollPane2;
  private javax.swing.JScrollPane jScrollPane3;
  private javax.swing.JScrollPane jScrollPane4;
  private javax.swing.JSplitPane jSplitPane2;
  private javax.swing.JPanel lists_panel;
  private javax.swing.JSplitPane main_hsplit;
  private javax.swing.JMenuBar menu_bar;
  private javax.swing.JTabbedPane service_tabs;
  private javax.swing.JPanel talkPanel2;
  private javax.swing.JSplitPane talk_hsplit;
  private javax.swing.JLabel talk_label;
  private javax.swing.JList talk_list;
  private javax.swing.JPanel talk_panel;
  private javax.swing.JPanel tc_list_panel;
  private javax.swing.JTextArea text;
  private javax.swing.JList user_list;
  private javax.swing.JPanel user_list_panel;
  private javax.swing.JScrollPane user_list_scroll_pane;
  private javax.swing.JSplitPane vsplit;
  // End of variables declaration//GEN-END:variables
}
