/** $Id */

// (c) Michael Buro, licensed under GPL2

// stretches/shrinks chess piece images

package ggs.ggsa.main;

import java.awt.*;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.MediaTracker;

public class ShrinkStretch extends Frame
{
  String[] pieces = { "p", "n", "b", "r", "q", "k" };
  // int[] sizes = { 21, 25, 29, 33, 37, 40, 45, 49, 54,
  //		  58, 64, 72, 80, 87, 95, 108, 116, 129 };
  int[] sizes = { 21, 25, 29, 33, 37, 45, 58, 80, 108, 129 };

  class ChessImages {
    public Image b_images[] = new Image[sizes.length];
    public Image w_images[] = new Image[sizes.length];

    private Image b_last;
    private Image w_last;
    private int last_size = -1;
    private int pref_size = -1;
    
    private void refresh(int sz) {
      if (sz == last_size) return;

      last_size = sz;
      int closest = 0;
      int min_diff = Math.abs(sz - sizes[0]);
    
      for (int i=1; i < sizes.length; ++i) {
	int d = Math.abs(sz - sizes[i]);
	if (d < min_diff) { closest = i; min_diff = d; }
      }

      b_last = b_images[closest];
      w_last = w_images[closest];

      if (min_diff == 1 || min_diff == 2) pref_size = sizes[closest];
      else pref_size = sz;
    }

    public int pref_size(int size) {
      refresh(size);
      return this.pref_size;
    }
    
    public Image get_b(int size) {
      refresh(size);
      return b_last;
    }

    public Image get_w(int size) {
      refresh(size);
      return w_last;
    }
  }

  ChessImages[] ci = new ChessImages[pieces.length];
  
  public void init()
  {
    // Create a media tracker to wait for the image

    MediaTracker tracker = new MediaTracker(this);
    String path = "gif/";
    
    for (int i=0; i < pieces.length; ++i) {

      ci[i] = new ChessImages();
	
      for (int j=0; j < sizes.length; ++j) {

	String f = path + pieces[i] + "dd" + sizes[j] + ".gif";
	  
	ci[i].b_images[j] = Toolkit.getDefaultToolkit().getImage(f);

	if (ci[i].b_images[j] == null) {
	  System.out.println(f + " not found");
	  System.exit(20);
	}
	  
	f = path + pieces[i] + "ld" + sizes[j] + ".gif";
	
	ci[i].w_images[j] = Toolkit.getDefaultToolkit().getImage(f);

	if (ci[i].w_images[j] == null) {
	  System.out.println(f + " not found");
	  System.exit(20);
	}
	  
	// Tell the media tracker to watch the images
	tracker.addImage(ci[i].b_images[j], 0);
	tracker.addImage(ci[i].w_images[j], 0);	
      }

    }

    // Wait for the images to be loaded
    try {
      tracker.waitForAll();
    } catch (Exception ignore) { }

    System.out.println("FOO");

  }

  public void paint(Graphics g)
  {
    int ww = getSize().width;
    int wh = getSize().height;
    int sz = Math.min(ww/(pieces.length+1), wh/3);
    int ps = 0;
    
    for (int i=0; i < pieces.length; ++i) {

      // Draw the image in its normal size

      ps = ci[i].pref_size(sz);
      int d = (sz - ps)/2;
      
      g.drawImage(ci[i].get_w(sz), 10+i*(sz+2)+d, 40+d, ps, ps, this);
      g.drawImage(ci[i].get_b(sz), 10+i*(sz+2)+d, 40+sz+2+d, ps, ps, this);      

    }
    
    System.out.println("size= " + sz + " ps=" + ps);
  }


  static public void main(String args[]) {

    ShrinkStretch s = new ShrinkStretch();

    s.setSize(800,600);
    s.init();
    s.setVisible(true);

    while (true);
  }
}


