/**
   $Id: SkatMatch.java,v 1.11 2006/09/04 03:40:42 orts_mburo Exp $
   $Source: /usr/bodo1/cvs/cvssoftware/ggs/src/ggs/service/skat/SkatMatch.java,v $
   
   Abstract Game Base Class

   (c) Michael Buro, licensed under GPL2
*/

package ggs.service.skat;

import java.util.*;
import ggs.service.common.*;
import ggs.shared.*;

class SkatMatch extends Match
{
  public SkatMatch(Game game_bp) 
  {
    this.game_bp = game_bp;
  }
  
  /** 
   * Initialize match
   * 
   * @return false iff match type mt OK
   * @return also return min/max number of players
   */
  public boolean init(String mt, int[] player_nums)
  {
    // no need to consult game - only one match type right now
    
    player_nums[0] = player_nums[1] = 3;
    return false;
  }

  /** 
      Start match for given number of players (pn)
  */
  public void start(int pn, Random gen)
  {
    player_num = pn;
    viewNum = pn + 1;
    game = new SkatGame();
    game.initState("regular", player_num, gen);

    // create start state views
    
    startStateViews = new Vector<String>();

    for (int i=0; i < viewNum; ++i)
      startStateViews.add(game.getStartStateView(indexToPerspective(i)));

    // prepare move view and move vectors

    moveViews = new Vector<Vector<String>>();

    for (int i=0; i < viewNum; ++i)
      moveViews.add(new Vector<String>());

    moves = new Vector<String>();
  }
  
  

  /**
     execute player cmd given as string array
      
     @return "" if move is OK and match is finished, otherwise != null
     if move is illegal, otherwise null if move is legal. Also
     generate messages for players and observers.
  */
  public String exec(int player, String move, boolean inClient)
  {
    // for now: just a single game per match
    
    String res = game.makeMove(player, move, true, false);

    if (res != null)
      return res;

    // compute move views and append to lists
    for (int i=0; i < player_num; ++i) 
      moveViews.get(i).add(game.lastMoveView(i));
      
    moveViews.get(player_num).add(game.lastMoveView(StdTable.VIEW_PUB)); // public
      
    // remember all move data (for saving game)
    moves.add(game.lastMoveData());
    
    if (game.finished())
      return "";

    return null;
  }

  /** @return last move in view of perspective */
  public String lastMoveView(int perspective)
  {
    return game.lastMoveView(perspective);
  }

  /** @return view of match from perspective */
  public String view(int perspective)
  {
    // start state
    
    int pi = perspectiveToIndex(perspective);
    StringBuffer sb = new StringBuffer(startStateViews.get(pi));

    // move sequence
    
    Vector<String> mv = moveViews.get(pi);
    
    for (String m : mv) {
      if (m != null)
	sb.append(" " + m);
    }
    
    return sb.toString();
  }
  
  /** access to match type */
  public String getMatchType() 
  {
    return match_type;
  }

  // helpers
  
  /** @return index of perspective */
  int perspectiveToIndex(int perspective)
  {
    return perspective >= 0 ? perspective : player_num;
  }

  /** @return index of perspective */
  int indexToPerspective(int index)
  {
    return index < player_num ? index : -1;
  }

  // data
  
  private int player_num; // number of players
  private int viewNum; // number of views: #players + 1 (public)
  private Game game_bp;
  private String match_type = "game";
  private SkatGame game;
  private Vector<String> startStateViews;   // #players + 1
  private Vector<Vector<String>> moveViews; // #players + 1
  private Vector<String> moves; // complete move data
}
