/* $Id: EventFactory.java,v 1.1 2006/08/28 06:57:31 orts_mburo Exp $
   
   (c) Michael Buro, licensed under the GPL2
*/

package ggs.shared;

import java.lang.Error;
import java.util.*;

public class EventFactory
{
  private Vector<EventHandler> handlers = new Vector<EventHandler>();
  private EventHandler default_handler = null;
  private String id = null;

  public EventFactory(String id)
  {
    if (id == null) throw new Error("id == null");
    this.id = id;
  }

  public void add_handler(EventHandler h) {
    handlers.add(h);
  }

  public void set_default_handler(EventHandler h) {
    default_handler = h;
  }

  public void remove_handler(EventHandler h) {
    if (default_handler == h) default_handler = null;
    while (handlers.remove(h));
  }

  public boolean send() { 
    return send(new EventMsg(id));
  }

  public boolean send(String s1) { 
    return send(new EventMsg(id, s1));
  }

  public boolean send(String s1, String s2) { 
    return send(new EventMsg(id, s1, s2));
  }

  public boolean send(String s1, String s2, String s3) { 
    return send(new EventMsg(id, s1, s2, s3));
  }

  public boolean send(Vector<String> args) { 
    return send(new EventMsg(id, args));
  }

  private boolean send(EventMsg msg)
  {
    boolean caught = false;

    for (EventHandler h : handlers)
      caught |= h.handle_event(msg);

    if (!caught && default_handler != null) 
      caught |= default_handler.handle_event(msg);

    return caught;
  }

}

