/**
   $Id: Game.java,v 1.9 2006/09/04 03:40:42 orts_mburo Exp $
   $Source: /usr/bodo1/cvs/cvssoftware/ggs/src/ggs/shared/Game.java,v $
   
   Abstract Game Base Class

   (c) Michael Buro, licensed under GPL2     
*/

package ggs.shared;

import java.io.*;
import java.util.*;

public interface Game
{
  /**
     init state w.r.t. game type and number of players
     @param game_type  type of game
     @param player_num number of players
     @param gen        random number generator
     @return null if parameters correct, error message otherwise. Also
     generate initial messages to players and observers
   */
  public String initState(String game_type, int player_num, Random gen);

  /**
   * @param view >=0: player, -1: public
   * @return string state representation for perspective
   */
  public String getStartStateView(int view);

  /**
   * set state from string
   * @return null if OK, error message otherwise
   */
  public String setStartState(String s);
  
  /**
     @return bit set that contains 1 for each player to move
   */
  public long playersToMove();

  /**
   * make move
   * @param player who moves
   * @param move to be played
   * @param doit change state, if false only check validity
   * @param inClient true iff move executed in client
   * @return error message if move is illegal, otherwise execute move
   * and create player move views iff doit is true and return null
   */
  public String makeMove(int player, String move, boolean doit, boolean inClient);

  /**
   * @param user perspective (bit i true iff player i watched)
   * @return view of last move
   */
  public String lastMoveView(int perspective);

  /**
     @return true iff game is finished
   */
  public boolean finished();

  /**
     @return game result for a player
   */
  public int result(int player);

  /**
     @return maximum result in state for a player
   */
  public int maxResult(int player);

  // -------- "static" functions ------------
  
  /**
     @return null if game type is OK, error message otherwise
     @return min/max number of players
   */
  public String checkGameType(String gt, int[] player_nums);

  /** @return game name */
  public String getName();
  
  /** @return new game */
  public Game newGame(); 

  /** @return new table */
  public Table newTable();

  // clock modes?

  /** "static" test function */
  public void test();
}
