/**
   $Id: StdTable.java,v 1.19 2006/09/04 03:40:42 orts_mburo Exp $
 
   Standard table

   (c) Michael Buro, licensed under GPL2    
*/

package ggs.shared;

import java.util.*;


public class StdTable extends Table 
{
  // special perspectives (>=0 indicates player view)
  public static final int VIEW_PUB = -1;
  public static final int VIEW_ALL = -2;  
  
  public StdTable(Match match)
  {
    this.match = match;
  }

  /** @return true iff table type OK */
  public boolean init(Sender sender, Random rand, String creator, String type, String pw)
  {
    if (!super.init(sender, rand, creator, type, pw))
      return false;

    int[] player_nums = new int[2];
      
    if (match.init(type, player_nums)) {
      sendError(creator, "illegal match-type " + type);
      return false;
    }

    minPlayerNum = player_nums[0];
    maxPlayerNum = player_nums[1];
    
    owner = creator;
    
    join(creator); // creator joins the table automatically
    
    return true;
  }

  /** user joins the table */
  public boolean join(String user)
  {
    if (!userArrived(user))
      return false;
    
    sendToAllExcept(user, "join " + user);
    return true;
  }

  private String setToString(Set<String> s)
  {
    StringBuffer sb = new StringBuffer();
    
    Iterator<String> it = s.iterator();

    while (it.hasNext()) {
      sb.append(it.next() + " ");
    }

    return sb.toString();
  }
  
  /** @return string representation of table in view of user */
  public String view(String user)
  {
    // write table info

    StringBuffer sb = new StringBuffer();

    // players

    sb.append(setToString(players));
    sb.append(" . ");
    
    // views

    Set<String> ks = views.keySet();

    for (String s : ks) {
      sb.append(s + " ");
      sb.append(setToString(players));
      sb.append(" . "); // list terminator 
    }

    sb.append(" . "); // list terminator
      
    // other data
    
    sb.append(" " + minPlayerNum);
    sb.append(" " + maxPlayerNum);
    sb.append(" " + owner);
    sb.append(" " + (matchInProgress?1:0));

    if (matchInProgress) {
      
      // match history
      int view = playerIndex(views.get(user)); // -1 if not found
      sb.append(" " + match.view(view));
    }

    return sb.toString();
  }

  /** create table from view string */
  public String fromSReader(SReader sr)
  {
    // read table info

    // read match info

    //match.fromString(
    return "FOO";
  }
  
  /*
   * execute table cmd in client
   */
  public boolean execInClient(SReader first, Vector<String> rest)
  {
    return execList("service", first, rest, cliCmds);
  }
  
  /**
   * execute table cmd in service
   */
  public boolean execInService(String sender, SReader first, Vector<String> rest)
  {
    prevOwner = owner;
    boolean r = execList(sender, first, rest, svcCmds);

    if (r && prevOwner != owner) {
      // owner changed: inform table
      sendToAll("owner " + owner);
    }

    return r;
  }

  /**
   * execute msg from list
   * @return true iff handled
   */
  private boolean execList(String sender,
			   SReader first,
			   Vector<String> rest,
			   ExecData[] ed)
  {
    String cmd = first.nextWord();
    setInfoChanged(false);
    
    for (ExecData e : ed) {
      Misc.msg("exec test : " + e.cmd1);
      if (!e.cmd1.equals(cmd) && !e.cmd2.equals(cmd))
	continue;
      Misc.msg("exec ! " + e.cmd1);      
      e.exec.exec(sender, first, rest);
      return true;
    }
    return false;    
  }

  /**
   * Returns information about the table
   * @return The table information
   */
  public String getTableInfo()
  {
    StdTableInfo ti = new StdTableInfo();

    ti.tableId = getId();
    ti.players = new TreeSet<String>(getPlayers());
    ti.owner = getOwner();
    ti.matchType = getType();
    // ti.observers = new TreeSet<String>(getKibitzersAndObservers());
    ti.matchInProgress = matchInProgress;
    ti.minPlayerNum = getMinPlayerNum();
    ti.maxPlayerNum = getMaxPlayerNum();
    ti.hasPasswd = getPasswd() != null;
    return ti.toString();
  }

  public HashMap<String, String> getViews() { return views; }

  /** @return minimum number of players */
  public int getMinPlayerNum() { return minPlayerNum; }

  /** @return maximum number of players */
  public int getMaxPlayerNum() { return maxPlayerNum; }

  /**
   * send move view to to users at table
   */
  private void sendMoveViews(String player)
  {
    // loop through all table users

    Set<String> ks = views.keySet();

    for (String user : ks) {
      int view = playerIndex(views.get(user)); // -1 if not found
      send(user, "play " + player + " " + match.lastMoveView(view));
    }
  }
  
  /** send message to player by index */
  public void sendToPlayer(int player, String msg)
  {
    Misc.msg("message to player " + player + ":" + msg);
    send(playerList.get(player), msg);
  }

  /** send message to user */
  public void sendToUser(String user, String msg)
  {
    Misc.msg("message to user " + user + ":" + msg);
    send(user, msg);
  }

  /** send message to all */
  public void sendToAll(String msg)
  {
    Misc.msg("message to all: " + msg);

    Set<String> all = getAll();

    for (String u : all) {
      sendToUser(u, msg);
    }
  }

  /** send message to all except for one user */
  public void sendToAllExcept(String user, String msg)
  {
    Misc.msg("message to all except " + user + ": " + msg);

    Set<String> all = getAll();

    for (String u : all) {
      if (!u.equals(user))
	sendToUser(u, msg);
    }
  }

  /**
   * Removes disconnected client from table
   * 
   */
  public void disconnected(String client)
  {
    if (userLeft(client)) {
      // fixme: inform match that client has left
      // fixme: inform table clients that client has left
      sendToAllExcept(client, "leave " + client);
    }
  }

  /** @return compute player index */
  // fixme: use faster data structure
  private int playerIndex(String player)
  {
    for (int i = 0; i < playerList.size(); ++i) {
      if (playerList.get(i) == player)
	return i;
    }

    return -1;
  }
  
  /** command interface */

  abstract class ExecCmd
  {
    /** return true iff exec OK */
    abstract void exec(String sender, SReader first, Vector<String> rest);

    /** send error message to client */
    public void error(String client, String cmd, String usage, String msg)
    {
      sendError(client, cmd + " - " + msg + "\\ usage: " + cmd + " " + usage);
    }
  }

  /** element of command list (see below) */

  class ExecData
  {
    public ExecData(String cmd1, String cmd2, ExecCmd exec)
    {
      this.cmd1 = cmd1;
      this.cmd2 = cmd2; // abbrevation
      this.exec = exec;
    }

    public String cmd1, cmd2;
    public ExecCmd exec;
  }

  // -------------- command execution in service --------------

  
  /*
   * cmd: kibleave <player>
   * 
   * no longer kibitzing player
   */

  class SvcKibLeave extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      String player = first.nextWord();

      if (player == null) {
	error(client, "syntax");
	return;
      }

      if (!removeKibitzer(client, player)) {
	error(client, "not a kibitzer");
	return;
      }

      sendToAll("kibleave " + client + " " + player);
    }

    void error(String client, String msg)
    {
      sendError(client, "kibleave", "<player>", msg);
    }
  }

  /*
   * cmd: kibboot <kibitzer>
   * 
   * remove one of your kibitzers
   */

  class SvcKibBoot extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      String kibitzer = first.nextWord();

      if (kibitzer == null) {
	error(client, "syntax");
	return;
      }
      
      if (!removeKibitzer(kibitzer, client)) {
	error(client, "not a kibitzer");
	return;
      }

      sendToAll("kibboot " + client + " " + kibitzer);
    }

    void error(String client, String msg)
    {
      sendError(client, "kibboot", "<kibitzer>", msg);
    }
  }

  /*
   * cmd: kibreq <player>
   * 
   * request for being a kibitz to a player
   */

  class SvcKibReq extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      String player = first.nextWord();

      if (player == null) {
	error(client, "syntax");
	return;
      }

      if (!getPlayers().contains(player)) {
	error(client, "not a player");
	return;
      }

      sendToAllExcept(client, "kibreq " + client + " " + player);
      sendToUser(client, "kibreq " + client + " " + player + " " + kibid);
      addKibReq(client, player, kibid);
      kibid++;
    }

    void error(String client, String msg)
    {
      super.error(client, "table", "<table> kibreq <player>", msg);
    }
  }

  /*
   * cmd: sitin
   * 
   * become a player
   */

  class SvcSitIn extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      if (getPlayers().size() < getMaxPlayerNum()) {
	
	// there is an empty seat at the table
	String r = userSitsIn(client); 
	if (r == null) {
	  sendToAll("sitin " + client);
	} else
	  error(client, r);

      } else {

	// no empty seat available
	error(client, "no more available seats");
      }
    }
    
    void error(String client, String msg)
    {
      sendError(client, "sitin", "", msg);
    }
  }

  /*
   * cmd: sitout
   * 
   * become an observer
   */

  class SvcSitOut extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      String r =userSitsOut(client);

      if (r == null)
	sendToAll("sitout " + client);
      else
	error(client, r);
    }
    
    void error(String client, String msg)
    {
      sendError(client, "sitout", "", msg);
    }
  }

  /*
   * cmd: leave
   * 
   * leave table
   */

  class SvcLeave extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      if (userLeft(client)) {

	sendToAll("leave " + client);
	send(client, "leave " + client);

	// fixme: match needs to be notified here

      } else {

	error(client, "you are not at this table");
      }
    }

    void error(String client, String msg)
    {
      sendError(client, "leave", "", msg);
    }
  }

  /*
   * cmd: tell p|o|a <msg>
   * 
   * send messages to anyone or everyone or somewhere in the middle
   */
  class SvcTell extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      String mode = first.nextWord();
      if (mode == null) {
	sendError(client, "tell", "o|a|p <msg>", "missing parameter");
	return;
      }

      Misc.msg("tell not implemented");
    }
  }
    
  //    } else {
	
  // 	// first[#]
  // 	// #1: p | o | a
  // 	// #2: msg

  // 	// create a new vector and remove the first 2 items so we can rebuild the message

  // 	Vector<String> tmpVec = new Vector<String>(Arrays.asList(parms));
  // 	tmpVec.remove(0);
  // 	tmpVec.remove(0);
  // 	String tmpStr = new String("");
  // 	for(String s : tmpVec) {
  // 	  tmpStr = tmpStr.concat(" ").concat(s);
  // 	}
	
  // 	String msg = client + ": " + tmpStr;

  // 	if (first[1].equals("a")) { // send to all
  // 	  for (String c : get_players())
  // 	    send(c, msg);
	    
  // 	  for (String c : get_observers())
  // 	    send(c, msg);
	    
  // 	} else if(first[1].equals("p")) { // send to players
  // 	  for (String c : get_players())
  // 	    send(c, msg);
  // 	} else if(first[1].equals("o")) { // send to observers
  // 	  for (String c : get_observers())
  // 	    send(c, msg);
  // 	} else {
  // 	  sendError(client,
  // 		    "usage: tell <[p]layers|[o]bservers|[a]ll> <msg>");
  // 	}

  /*
   * cmd: play <move>
   * 
   * play move
   */

  class SvcPlay extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      String move = first.nextWord();
      
      if (move == null) {
	error(client, "move missing");
	return;
      }

      if (!getMatchInProgress()) {
	error(client, "match not started yet");
	return;
      }

      int player = playerIndex(client);
      
      if (player < 0) {
	error(client, "you are not playing this match");
	return;
      }
      
      String res = getMatch().exec(player, move, false);
      
      if (res == null) {
	// move ok
	sendMoveViews(client);
	return;
      }
      
      if (res.equals("")) {
	// move ok and match over
	sendMoveViews(client);
	setMatchInProgress(false);
	return;
      }
      
      // invalid move
      error(client, res);
    }

    void error(String client, String msg)
    {
      sendError(client, "play", "<move>", msg);
    }
  }

  
  /*
   * cmd: start
   * 
   * start game, can only be issued by table owner
   */

  class SvcStart extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      createPlayerList();
      
      if (!client.equals(owner)) {
	sendError(client, "start", "", "must be table owner");
	return;
      }

      if (matchInProgress) {
	sendError(client, "start", "", "match already in progress");
	return;
      }

      if (playerList.size() < getMinPlayerNum()) {
	sendError(client, "start", "",
		  "" + (getMinPlayerNum()-playerList.size()) + " more player(s) needed");
	return;
      }

      setInfoChanged(true);
      setMatchInProgress(true);
      match.start(playerList.size(), getRand());

      // send initial state

      Set<String> ks = views.keySet();
      
      for (String s : ks) {
	send(s, "state " + getId() + " " + view(s));
      }
    }
  }
  

  // -------------- command execution in client --------------

  /*
   * cmd: join <user>
   * 
   * user joined table
   */

  class CliJoin extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      userArrived(client);
    }
  }

  /*
   * cmd: leave <user>
   * 
   * user left table
   */

  class CliLeave extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      userLeft(client);
    }
  }

  /*
   * cmd: sitin <user>
   * 
   * user becomes player
   */

  class CliSitIn extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      userSitsIn(client);
    }
  }

  /*
   * cmd: sitout <user>
   * 
   * user no longer player
   */

  class CliSitOut extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      userSitsOut(client);
    }
  }

  /*
   * cmd: play <player> <move>
   * 
   * player played move
   */

  class CliPlay extends ExecCmd
  {
    public void exec(String client, SReader first, Vector<String> rest)
    {
      int player = playerIndex(client);
      
      if (player < 0) {
	Misc.err("CliPlay: " + client + " not playing this match");
	return;
      }

      String move = first.nextWord();
      
      if (move == null) {
	Misc.err("CliPlay: received null move");
	return;
      }
	
      String res = getMatch().exec(player, move, true);
      if (res != null) {
	Misc.err("CliPlay: illegal move: " + res);
	return;
      }
    }
  }

  // -------------- data + helpers ----------------

  public Match getMatch() { return match; }

  public boolean getMatchInProgress() { return matchInProgress; }

  public void setMatchInProgress(boolean f)
  {
    matchInProgress = f;
  }


  /**
   * Remove kibitzer from player
   * 
   * @param kibitzer
   *                Name of the kibitzer
   * @param player
   *                Name of observed
   * 
   * @return true iff removed
   */
  public boolean removeKibitzer(String kibitzer, String player)
  {
    boolean r = views.get(kibitzer).equals(player);
    if (r) {
      views.put(kibitzer, ""); // reset
      setChanged();
      // notifyObservers(kibitzers);
    }

    return r;
  }

  /**
   * new users become observers
   * 
   * @return true iff new
   */

  public boolean userArrived(String user)
  {
    String view = views.get(user);

    if (view != null)
      return false; // already there

    views.put(user, ""); // public

    setInfoChanged(true);
    notifyObservers();
    return true;
  }

  /*
   * remove user from all lists
   * 
   * @return true iff user left game
   */

  public boolean userLeft(String user)
  {
    if (views.remove(user) == null)
      return false;
      
    if (players.remove(user))
      createPlayerList();

    if (user.equals(owner)) {

      // owner left: pick owner

      if (players.size() > 0) {
	// random player
	Object[] a = players.toArray();
	owner = (String)a[getRand().nextInt(a.length)];
      } else if (views.size() > 0) {
	// random user at table
	Object[] a = views.keySet().toArray();
	owner = (String)a[getRand().nextInt(a.length)];
      } else {
	owner = null; // table is dead
      }
    }
    
    for (String s : views.keySet()) {
      if (views.get(s).equals(user)) {
	views.put(s, "");
      }
    }

    setInfoChanged(true);      
    notifyObservers();
    return true;
  }

  /**
   * user becomes player
   * 
   * @return null if OK, error message otherwise
   */
  public String userSitsIn(String user)
  {
    String view = views.get(user);

    if (view == null)
      return "not at table";

    if (!players.add(user))
      return "you are a player already";

    createPlayerList();
    
    // set user's view to self
    views.put(user, user);

    setInfoChanged(true);
    notifyObservers();
    return null;
  }

  /**
   * player becomes oberserver
   * 
   * @return null if ok, error message otherwise
   */
  public String userSitsOut(String user)
  {
    if (!players.remove(user))
      return "you are not a player";

    createPlayerList();
    
    views.put(user, ""); // public

    setInfoChanged(true);    
    notifyObservers();
    return null;
  }

  /** @return set of users at table */
  public Set<String> getAll() { return views.keySet(); }

  /** @return set of players at table */  
  public Set<String> getPlayers() { return players; }

  /** @return owner of table */
  public String getOwner() { return owner; }

  /** add kibitz request to list */
  public void addKibReq(String from, String to, int kibid)
  {
    kibReqs.add(new KibReq(from, to, kibid));
  }

  /**
   * remove kibitz request from list if it is present
   * @return true iff removed
   */ 
  public boolean acceptKibReq(String from, String to, int kibid)
  {
    for (int i = 0; i < kibReqs.size(); i++) {
      KibReq k = kibReqs.get(i); 

      if (k.to.equals(to) && k.from.equals(from) && k.id == kibid) {
	kibReqs.remove(i);
	return true;
      }
    }

    return false;
  }

  //   public int compareTo(Object o)
  //   {
  //     // TODO Auto-generated method stub
  //     return 0;
  //   }

  /** create player list from set, call whenever players changes */
  private void createPlayerList()
  {
    Iterator<String> it = players.iterator();
    playerList.clear();

    while (it.hasNext()) {
      playerList.add(it.next());
    }

  }
  
  // data

  // players   : participate in match
  // observers : see public match information
  // kibitzers : see private player information and public information

  // players can be kibitzers

  
  // vvvvvv data shared with client vvvvvv
  
  /** players at table */
  private Set<String> players = new TreeSet<String>();

  /** player list (maps int to player) */
  private Vector<String> playerList = new Vector<String>();

  /** view of each table participant, "" : public */
  private HashMap<String, String> views = new HashMap<String,String>();

  private int minPlayerNum, maxPlayerNum; // min-max number of players at table

  private String owner; // table owner; can start match, kick out people etc.
  private boolean matchInProgress; // is match in progress?
  private Match match;  // match being played

  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

  private String prevOwner; // previous owner
  
  /** list of kibitz requests */
  private Vector<KibReq> kibReqs = new Vector<KibReq>();

  /** running kibitz request id */
  private int kibid;

  /** commands to service table */
  
  private ExecData svcCmds[] =
  {
    new ExecData("play",     "p",  new SvcPlay()),   // play move
    //new ExecData("tell",     "t",  new SvcTell()),   // table tell
    new ExecData("start",    "r",  new SvcStart()),    // start match
    new ExecData("leave",    "l",  new SvcLeave()),    // leave table	
    new ExecData("sitin",    "si", new SvcSitIn()),    // become player	
    new ExecData("sitout",   "so", new SvcSitOut()),   // become observer      
    new ExecData("kibreq",   "kr", new SvcKibReq()),   // request to become kibitzer
    //new ExecData("kibacc",   "ka", new SvcKibAcc(s)),   // accepte/decline kib.request
    new ExecData("kibboot",  "kb", new SvcKibBoot()),  // kick out one of you kibitzers
    new ExecData("kibleave", "kl", new SvcKibLeave())  // stop kibitzing a player
  };


  /** messages to client table */
  
  private ExecData cliCmds[] =
  {
    //new ExecData("ready",  "r",  new CliReady()),  // player sent ready
    new ExecData("join",   "j",  new CliJoin()),     // user joined table
    new ExecData("leave",  "l",  new CliLeave()),    // user left table
    new ExecData("sitin",  "si", new CliSitIn()),    // user becomes player
    new ExecData("sitout", "so", new CliSitOut()),   // player becomes observer
    new ExecData("play",   "p",  new CliPlay()),     // move was played
    //new ExecData("kibreq", "kr", new CliKibReq()), // kibitz request
    //new ExecData("kibacc", "ka", new CliKibitz()), // kibitz accepted/declined
    //new ExecData("tell",   "t",  new CliTell()),   // table tell
  };

}
