/**
   $Id: Table.java,v 1.10 2006/09/03 02:48:39 orts_mburo Exp $
   $Source: /usr/bodo1/cvs/cvssoftware/ggs/src/ggs/shared/Table.java,v $
   
   abstract Table class

   (c) Michael Buro, licensed under GPL2   
*/

package ggs.shared;

import java.util.*;

public abstract class Table extends Observable
{
  /** user joins the table, @return true if joine */
  public abstract boolean join(String user);

  /** @return string representation of table history in view of user */
  public abstract String view(String user);

  /** create table from sreader */
  public abstract String fromSReader(SReader sr);  
  
  /**
   * execute cmd in service, call setInfoChanged(true) if TableInfo changed
   * @return true iff handled
   */
  public abstract boolean execInService(String sender,
					SReader first, Vector<String> lines);

  /**
   * execute cmd in client, call setInfoChanged(true) if TableInfo changed
   * @return true iff handled     
   */
  public abstract boolean execInClient(SReader first, Vector<String> lines);

  /** access to match */
  public abstract Match getMatch();

  /** @return information about table */
  public abstract String getTableInfo();

  /** initialize state */
  public boolean init(Sender sender, Random rand, String creator, String type, String pw)
  {
    this.sender  = sender;
    this.rand    = rand;
    this.creator = creator;
    this.type    = type;    
    this.passwd  = pw;
    return true;
  }

  /** @return table id */
  public int getId() { return id; }

  /** set id */
  public void setId(int id) { this.id = id; }

  /** @return sender */
  public Sender getSender() { return sender; }

  /** @return RNG */
  public Random getRand() { return rand; }

  /** @return passwd */
  public String getPasswd() { return passwd; }

  /** @return table type */
  public String getType() { return type; }

  /** send error message to client on a table */
  public void sendError(String client, String cmd, String usage, String msg)
  {
    sender.send(" t " + client +
		" error table " + id + " " + cmd + " : " + msg + "\\ usage: " + cmd + " " + usage);
  }

  /** send error message to client at a table */
  public void send(String client, String msg)
  {
    sender.send("t " + client + " table " + id + " " + msg);
  }

  /** send error message to client at a table */
  public void sendError(String client, String msg)
  {
    sender.send("t " + client + " table " + id + " error: " + msg);
  }

  /** send messages to players */
  //  public abstract void sendMsgs(Vector<String> msgs);
  
  /** send message to individual player */
  //public abstract void sendToPlayer(int player, String msg);

  /** remove client from table */
  public abstract void disconnected(String client);

  public boolean getInfoChanged() { return infoChanged; }
  protected void setInfoChanged(boolean f) { infoChanged = f; }
  
  /** data */
  private int id;
  private String passwd; // for joining game
  private Sender sender;
  private Random rand;
  private String creator; // who created this table
  private String type;
  private String tableInfo;
  private boolean infoChanged;
}
